#include <stdio.h>
#include <sys/param.h>
#include <sys/types.h>

#include <xview/xview.h>

#include <iluxport.h>

typedef struct rock {
  ilu_boolean input;
  int fd;
  ilu_IOHandler handler;
  ilu_private rock;
  struct rock * next;
} * Rock;

static Rock list_of_rocks = NULL;
static Rock unused_rocks = NULL;

static Notify_value func(Notify_client client, int fd) 
{
  (*(((Rock ) client)-> handler))(fd, ((Rock ) client)->rock);
  return NOTIFY_DONE;
}

static Rock GetRock(int fd, ilu_boolean input)
{
  Rock c;
  for (c = list_of_rocks; c != NULL; c = c->next) {
    if (fd == c->fd && input == c->input) {
      return c;
    }
  }
  return NULL;
}

static void DeleteRock(int fd, ilu_boolean input) 
{
  Rock c, p;
  for (p = NULL, c = list_of_rocks; c != NULL; p = c, c = c->next) {
    if (fd == c->fd && input == c->input) {
      if (p != NULL)
	p->next = c->next;
      else
	list_of_rocks = c->next;
      c->input = ilu_FALSE;
      c->handler = NULL;
      c->rock = NULL;
      c->next = unused_rocks;
      unused_rocks = c;
    }
  }
}

static Rock NewRock ()
{
  Rock r;

  if (unused_rocks == NULL)
    return (Rock) malloc(sizeof(* r));
  else
    {
      r = unused_rocks;
      unused_rocks = r->next;
      return (r);
    }
}

static Rock AddRock (int fd,ilu_IOHandler handler, ilu_private rock, ilu_boolean input)
{
  Rock r;
  DeleteRock(fd, input);
  r = NewRock();
  if (r == NULL) {
    fprintf(stderr, "malloc failed in AddRock\n");
    exit(1);
  }
  r->input = input;
  r->fd = fd;
  r->handler = handler;
  r->rock = rock;
  r->next = list_of_rocks;
  list_of_rocks = r;
  return r;
}

static ilu_boolean reg_inp(int fd, ilu_IOHandler handler, ilu_private rock)
{
  Rock r = AddRock(fd, handler, rock, ilu_TRUE);
  return ((notify_set_input_func((Notify_client) r, (Notify_func) func, fd) != NULL) ? ilu_TRUE : ilu_FALSE);
}


static ilu_boolean can_inp(int fd)
{
  ilu_boolean b;
  Rock r = GetRock(fd, ilu_TRUE);

  if (r != NULL)
    b = ((notify_set_input_func((Notify_client) r, NULL, fd) != NULL) ? ilu_TRUE : ilu_FALSE);
  else
    b = ilu_TRUE;
  DeleteRock(fd, ilu_TRUE);
  return b;
}

static ilu_boolean reg_out (int fd, ilu_IOHandler handler, ilu_private rock) 
{
  Rock r = AddRock(fd, handler, rock, ilu_FALSE);
  return ((notify_set_output_func((Notify_client) r, (Notify_func) func, fd) == NULL) ? ilu_FALSE : ilu_TRUE);
}

static ilu_boolean can_out(int fd)
{
  ilu_boolean b;
  Rock r = GetRock(fd, ilu_FALSE);

  if (r != NULL)
    b = ((notify_set_output_func((Notify_client) r, NULL, fd) != NULL) ? ilu_TRUE : ilu_FALSE);
  else
    b = ilu_TRUE;
  DeleteRock(fd, ilu_FALSE);
  return b;
}

static void (*ilu_timer_proc) (ilu_FineTime t) = NULL;
static ilu_FineTime ilu_timer_time;
static Notify_func ilu_notify_func;

void timer_func (Notify_client client, int which)
{
  if (client == (Notify_client) &ilu_timer_proc && ilu_timer_proc != NULL)
    {
      (*ilu_timer_proc)(ilu_timer_time);
      ilu_notify_func = NULL;
    }
}

static void set_alarm (ilu_FineTime t,
			/*Main Invariant holds;
			  L2 otherwise unconstrained*/
		       void (*proc) (ilu_FineTime t))
{
  struct itimerval ivalue;
  ilu_FineTime initial_timeout = ilu_FineTime_Sub(t, ilu_FineTime_Now());

  fprintf (stderr, "Warning:  ILU asked XView to set an alarm, and this function isn't implemented!\n");
  ilu_timer_proc = proc;
  ilu_timer_time = t;
  ivalue.it_value.tv_sec = initial_timeout.ft_s;
  ivalue.it_value.tv_usec = initial_timeout.ft_t * (1000000/ilu_FineTimeRate);
  ivalue.it_interval.tv_sec = 0;
  ivalue.it_interval.tv_usec = 0;
  ilu_notify_func = notify_set_itimer_func ((Notify_client) &ilu_timer_proc, (Notify_func) timer_func, ITIMER_REAL,
					    &ivalue, NULL);
}

static void can_alarm (void)
{
  notify_set_itimer_func ((Notify_client) &ilu_timer_proc, NOTIFY_FUNC_NULL, ITIMER_REAL, NULL, NULL);
  ilu_timer_proc = NULL;
}

void ilu_xview_Initialize (void)
{
  _ilu_Assert(ilu_AddRegisterersToDefault(reg_inp, can_inp, reg_out, can_out, set_alarm, can_alarm),
	      "ilu_xview_Initialize");
}
