#
# Copyright (c) 1991, 1992, 1993, 1994 Xerox Corporation.  All Rights Reserved.
#
# Unlimited use, reproduction, and distribution of this software is
# permitted.  Any copy of this software must include both the above
# copyright notice of Xerox Corporation and this paragraph.  Any
# distribution of this software must comply with all applicable United
# States export control laws.  This software is made available AS IS,
# and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
# INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
# PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
# THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
# CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
# XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
#
# $Id: rwho2PyServer.py,v 1.1 1994/10/14 01:44:00 severson Exp $
#

import os, socket, sys, time

import ilu
import rwho2__skel
import rwho2a__skel

def getCurrentTime():
    return time.asctime(time.localtime(time.time()))

def trim(line):
    return line[0:-1]

def isPrefixOf(s1, s2):
    lens1 = len(s1)
    return lens1 <= len(s2) and s1 == s2[0:lens1]

def getWhoList():
    f = os.popen("who", "r")
    lineList = f.readlines()
    f.close()
    return lineList

holdList = []

def hold(obj):
    global holdList
    holdList.append(obj)
    return obj


class user(rwho2__skel.user):
    def __init__(self, name):
	self.name = name

    def Name(self):
	return self.name

class super(rwho2__skel.super):
    def __init__(self, id):
	self.id = id

    def ID(self):
	return self.id

class machine(rwho2__skel.machine):
    def __init__(self):
	self.startTime = getCurrentTime()
	self.hostName = socket.gethostname()

    def Status(self):
	f = os.popen("uptime", "r")
	line = f.readline()
	f.close()
	return trim(line)

    def IsUserPresent(self, username):
	for line in getWhoList():
	    entry = trim(line)
	    if isPrefixOf(username + " ", entry):
		return ilu.TRUE
	return ilu.FALSE

    def StartTime(self):
	return self.startTime

    def Name(self):
	return self.hostName

    def ListUsers(self):
	time = getCurrentTime()
	users = []
	for line in getWhoList():
	    entry = trim(line)
	    if len(entry) >= 3 and entry[0:3] == 'su ':
		u = superuser(entry, "foo")
	    else:
		u = user(entry)
	    users.append(hold(u))
	return {"time": time, "users": users}

    def EchoString(self, s):
	return s

    def EchoInt(self, s):
	return s

class superuser(rwho2a__skel.superuser, user, super):
    def __init__(self, name, id):
	user.__init__(self, name)
	super.__init__(self, id)

def main(argv):
    if len(argv) > 1:
	serverID = argv[1]
    else:
	serverID = None
    if len(argv) > 2:
	transport = argv[2]
    else:
	transport = None
    if len(argv) > 3:
	protocol = argv[3]
    else:
	protocol = None

    server = ilu.CreateServer(serverID, transport, protocol)

    mach = machine()
    print mach.IluSBH()

    ilu.RunMainLoop()

main(sys.argv)
