/*
Copyright (c) 1991, 1992, 1993 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

$Id: rwho2Server.cc,v 1.11 1994/12/13 06:36:58 janssen Exp $
*/

#include "rwho2.H"	/* needed for def of rwho2 classes and types */
#include "rwho2a.H"
#include <stdio.h>	/* I/O defs (including popen and pclose) */
#include <string.h>
#include <sys/types.h>
#include <time.h>
#include <errno.h>
#include <sys/errno.h>

extern "C" {
  extern void gethostname(char *, int);
};

/* Unfortunately, "strdup" isn't part of the ANSI C standard. */
static char *
copy_string(const char *str)
{
  char *result = 0;

  if (str != 0)
    {
      result = (char *) malloc(strlen(str) + 1);
      if (result)
        strcpy(result, str);
    }
  return result;
}

class superImpl : public rwho2_T_super {

 public:

  superImpl();			// constructor
  virtual ~superImpl();		// destructor

 // methods

  static class superImpl *Create (char *name);
  virtual ilu_CString ID (rwho2Status *_status);

 // data slots

 private:
   char *id;
};

superImpl::superImpl()
{
  this->id = NULL;
}

superImpl::~superImpl()
{
  if (this->id != NULL)
    free(this->id);
}

class superImpl *superImpl::Create (char *name)
{
  class superImpl *n = new superImpl;
  n->id = copy_string(name);
  return (n);
}

ilu_CString superImpl::ID (rwho2Status *_status)
{
  _status->returnCode = rwho2Reply_Success;
  return (copy_string(this->id));
}

class superuserImpl : public rwho2a_T_superuser {

 public:

  superuserImpl();			// constructor
  virtual ~superuserImpl();		// destructor

 // methods

  static class superuserImpl *Create (char *name, char *id);
  virtual rwho2_T_username Name (rwho2Status *_status);
  virtual ilu_CString ID (rwho2Status *_status);

 // data slots

 private:
   char *name;
   char *id;
};

superuserImpl::superuserImpl()
{
  this->name = NULL;
  this->id = NULL;
}

superuserImpl::~superuserImpl()
{
  if (this->name != NULL)
    free(this->name);
  if (this->id != NULL)
    free(this->id);
}

class superuserImpl *superuserImpl::Create (char *name, char *id)
{
  class superuserImpl *n = new superuserImpl;
  n->name = copy_string(name);
  n->id = copy_string(id);
  return (n);
}

rwho2_T_username superuserImpl::Name (rwho2Status *_status)
{
  _status->returnCode = rwho2Reply_Success;
  return (copy_string(this->name));
}

ilu_CString superuserImpl::ID (rwho2Status *_status)
{
  _status->returnCode = rwho2Reply_Success;
  return (copy_string(this->id));
}

class userImpl : public rwho2_T_user {

 public:

  userImpl(char *name, char *status);		// constructor

 // methods

  static class userImpl *Create (char *name);
  virtual rwho2_T_username Name (rwho2Status *_status);

 // data slots

  char *status;
};

userImpl::userImpl(char *name, char *s) : iluObject(name)
{
  status = s;
}

class userImpl *userImpl::Create (char *status)
{
  class userImpl *n;
  char *p, *s;

  p = strpbrk(status, " \t");
  s = p + strspn(p, " \t");
  *p = (char) 0;
  n = new userImpl (status, s);
  return (n);
}

rwho2_T_username userImpl::Name (rwho2Status *_status)
{
  char buf[1000];

  _status->returnCode = NULL;
  sprintf (buf, "%s -- %s", this->ILUGetInstanceHandle(), this->status);
  return (copy_string(buf));
}


/* First, subclass the public version of rwho2_T_machine, and
   provide real implementations of each of the methods. */

class rwho2Impl : public rwho2_T_machine {

 public:

  rwho2Impl();			// constructor
  virtual ~rwho2Impl();		// destructor

 // methods

  virtual ilu_CString Status (rwho2Status *_status);
  virtual ilu_Boolean IsUserPresent (rwho2Status *_status, rwho2_T_username username);
  virtual ilu_CString StartTime (rwho2Status *_status);
  virtual ilu_CString Name (rwho2Status *_status);
  virtual rwho2_T_userRec * ListUsers (rwho2Status *_status);
  virtual ilu_CString EchoString (rwho2Status *_status, ilu_CString s);
  virtual ilu_Cardinal EchoInt (rwho2Status *_status, ilu_Cardinal s);

 // data slots

 private:

   char *startTime;
};


static char *GetTimeAsString()
{
  long clock;
  char *buf;

  clock = time(NULL);
  buf = copy_string(ctime(&clock));
  buf[strlen(buf) - 1] = '\0';
  return (buf);
}

rwho2Impl::rwho2Impl()
{
  this->startTime = GetTimeAsString();
}

rwho2Impl::~rwho2Impl()
{
  if (this->startTime != NULL)
    free(this->startTime);
}

ilu_CString rwho2Impl::Status (rwho2Status *_status)
{
  char buf[1000];
  FILE *f;

  f = popen ("uptime", "r");
  fgets(buf, sizeof(buf), f);
  pclose(f);
  buf[strlen(buf) - 1] = '\0';
  _status->returnCode = rwho2Reply_Success;
  return ((ilu_CString) copy_string(buf));
}

rwho2_T_userRec * rwho2Impl::ListUsers (rwho2Status *_status)
{
  FILE *f;
  char buf[1000];
  rwho2_T_userRec *users = (rwho2_T_userRec *) malloc (sizeof(rwho2_T_userRec));

  users->users = _rwho2_T_userSeq_sequence::Create(10, NULL);
  users->time = GetTimeAsString();
  f = popen ("who", "r");
  while (fgets (buf, sizeof(buf), f) != NULL && strlen(buf) > 1)
    {
      char *p;

      buf[strlen(buf) - 1] = '\0';
      if ((p = strchr(buf, ' ')) != NULL 
	  && (*--p == 'u') && ((*--p) == 's'))
	users->users->Append ((class rwho2_T_user *) superuserImpl::Create(buf, "foo"));
      else
	users->users->Append((class rwho2_T_user *) userImpl::Create(copy_string(buf)));
    }
  pclose(f);
  _status->returnCode = rwho2Reply_Success;
  return (users);  
}

ilu_Boolean rwho2Impl::IsUserPresent (rwho2Status *_status, rwho2_T_username username)
{
  FILE *f;
  char buf[1000];
  int i;
  ilu_Boolean s;

  f = popen ("who", "r");
  i = strlen(username);
  s = ilu_FALSE;
  while (fgets (buf, sizeof(buf), f) != NULL && strlen(buf) > 1)
    if (strncmp(username, buf, i) == 0)
      s = ilu_TRUE;
  pclose(f);
  _status->returnCode = rwho2Reply_Success;
  return (s);
}

ilu_CString rwho2Impl::StartTime (rwho2Status *_status)
{
  _status->returnCode = rwho2Reply_Success;
  return (copy_string(this->startTime));
}

ilu_CString rwho2Impl::Name (rwho2Status *_status)
{
  static char *Hostname = NULL;

  if (Hostname == NULL)
    {
      char buf[1000];
      gethostname(buf, sizeof(buf));
      Hostname = copy_string(buf);
    }

  _status->returnCode = rwho2Reply_Success;
  return ((ilu_CString) copy_string(Hostname));
}

ilu_CString rwho2Impl::EchoString (rwho2Status *_status, ilu_CString s)
{
  ilu_CString s2;

  _status->returnCode = rwho2Reply_Success;
  s2 = (ilu_CString) malloc(strlen(s) + 1);
  strcpy (s2, s);
  return (s2);
}

ilu_Cardinal rwho2Impl::EchoInt (rwho2Status *_status, ilu_Cardinal s)
{
  _status->returnCode = rwho2Reply_Success;
  return (s);
}


/***********************************************************************/
/*  Set up to do dynamic incarnation of users			       */
/***********************************************************************/

class rwho2ObjTable : public virtual iluObjectTable {

public:

  virtual iluObject * ObjectOfIH (ilu_CString ih);

};

static char *figureStatus (char *who)
{
  FILE *f;
  char buf[1000];
  char *s = NULL;
  int n = strlen(who);

  /* return NULL if who is not on system, status string otherwise */

  f = popen ("who", "r");
  while (fgets (buf, sizeof(buf), f) != NULL && strlen(buf) > 1)
    {
      buf[strlen(buf) - 1] = '\0';
      if ((strncmp(who, buf, n) == 0) && (buf[n] == ' '))
	s = copy_string(buf);
    }
  pclose(f);
  return(s);
}

static userImpl *nobody = NULL;

iluObject * rwho2ObjTable::ObjectOfIH (ilu_CString ih)
{
  iluObject *u = NULL;
  char *status;

  if (nobody == NULL)
    nobody = new userImpl ("nobody", "not present");

  if ((status = figureStatus (ih)) != NULL)
    u = userImpl::Create (status);

  if (u == NULL)
    fprintf (stderr, "Can't instantiate userImpl object for user <%s>.\n", ih);
  return (u);
}

/***********************************************************************/
/***********************************************************************/
/***********************************************************************/
/***********************************************************************/

/*		And here's the main line			       */

/***********************************************************************/
/***********************************************************************/
/***********************************************************************/

main (int ac, char **av)
{
  /* first, make a server object */
  class iluServer s((ac > 1) ? av[1] : NULL, new rwho2ObjTable);
  
  class rwho2Impl *i;

  /* Do this, instead of overriding i.ILUGetServer */
  
  ilu::SetDefaultServer(&s);
  
  /* now, add the only port */

  s.AddPort ( (ac > 2) ? av[2] : NULL, (ac > 3) ? av[3] : NULL, ilu_TRUE);

  /* create an instance of the rwho2Impl object */

  i = new rwho2Impl;

  /* better print out the SBH to pass as arg to rwhoClient... */

  printf ("%s\n", i->ILUStringBindingHandle());

  /* and let things run... */

  iluServer::Run();

  /* should never return, but just in case: */

  return(1);
}

