#include <stdlib.h>	/* for atoi */
#include <unistd.h>	/* for gethostname */
#include <stdio.h>

#include <sys/time.h>
#include <sys/times.h>

#include "timeit.h"

#define DOMAIN	"parc.xerox.com"

int main(int ac, char **av)
{
  char hostname[1000], objectid[1000];
  timeit_p handle;
  struct timeval timebegin, timeend;
  char *tests[] = { "cardinal (C unsigned long)", "real (C double)", "ilu.CString (10 chars)" };
  ilu_CString retstr;
  double utime, stime, cutime, cstime;
  ilu_cardinal i, repeat, j, testtype;
  struct tms beginning, ending;
  ILU_C_ENVIRONMENT e;

  timeit__Initialize( );

  if (ac < 3)
    {
      fprintf (stderr, "Usage:  client TEST { 1=cardinal, 2=real, 3=string } REPEATCOUNT [ HOSTNAME ]\n");
      exit(1);
    }
  testtype = atoi(av[1]);
  repeat = atoi(av[2]);
  if (testtype == 0 || testtype > 3 || repeat < 1)
    {
      fprintf (stderr, "error:  testtype of %ld, repeatcount of %ld specified.\n",
	       testtype, repeat);
      exit(1);
    }
  if (ac > 3)
    strcpy (hostname, av[3]);
  else
    gethostname(hostname, sizeof(hostname));
  sprintf (objectid, "0@TimingTest.%s.%s", hostname, DOMAIN);
  if ((handle = ILU_C_LookupObject (objectid, &_timeit_p__ILUClassRecord)) == NULL)
    {
      fprintf (stderr, "error:  Can't obtain object <%s>\n", objectid);
      exit(1);
    }

  gettimeofday (&timebegin, NULL);
  switch (testtype)
    {
    case 1:

      times(&beginning);
      for (i = 0;  i < repeat;  i++)
	{
	  if (j = timeit_p_ping1 (handle, &e, i), e.returnCode != NULL)
	    {
	      fprintf (stderr, "Failure %p in call %lu to ping1\n", e.returnCode, i+1);
	      exit(2);
	    }
	}
      times(&ending);
      break;

    case 2:

      times(&beginning);
      for (i = 0;  i < repeat;  i++)
	{
	  utime = i;
	  if (stime = timeit_p_ping2 (handle, &e, utime), e.returnCode != NULL)
	    {
	      fprintf (stderr, "Failure %p in call %lu to ping2\n", e.returnCode, i+1);
	      exit(2);
	    }
	}
      times(&ending);
      break;

      break;

    case 3:

      times(&beginning);
      for (i = 0;  i < repeat;  i++)
	{
	  if (retstr = timeit_p_ping3 (handle, &e, "some string"), e.returnCode != NULL)
	    {
	      fprintf (stderr, "Failure %p in call %lu to ping3\n", e.returnCode, i+1);
	      exit(2);
	    }
	}
      times(&ending);
      break;

      break;

    }
  gettimeofday (&timeend, NULL);

  utime = ((ending.tms_utime - beginning.tms_utime) * 1000) / (repeat * 60.0);
  stime = ((ending.tms_stime - beginning.tms_stime) * 1000) / (repeat * 60.0);
  cutime = ((ending.tms_cutime - beginning.tms_cutime) * 1000) / (repeat * 60.0);
  cstime = ((ending.tms_cstime - beginning.tms_cstime) * 1000) / (repeat * 60.0);

  printf ("test \"%s\", repeat %lu:\n", tests[testtype - 1], repeat);
  printf ("  elapsed time, ms per call:  %.3f      total:  %.2f sec\n",
	  ((((double) timeend.tv_sec * 1000000.0) + timeend.tv_usec) -
	   (((double) timebegin.tv_sec * 1000000.0) + timebegin.tv_usec))
	  / (((double) repeat) * 1000.0),
	  ((((double) timeend.tv_sec * 1000000.0) + timeend.tv_usec) -
	   (((double) timebegin.tv_sec * 1000000.0) + timebegin.tv_usec)) / 1000000.0);
  printf ("     user time:  %3lu - %3lu = %.3f millisec/call\n", ending.tms_utime, beginning.tms_utime, utime);
  printf ("   system time:  %3lu - %3lu = %.3f millisec/call\n", ending.tms_stime, beginning.tms_stime, stime);

  exit(0);
}
