/*
Copyright (c) 1991-1995 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/

#ifndef __ilu_H_
#define __ilu_H_ 1

#include <stdlib.h>
#include <string.h>

extern "C" {
#include <iluxport.h>
};

  typedef unsigned char ilu_Boolean;

  typedef ilu_boolean ilu_kernelBoolean;
#define ilu_kernelTRUE ilu_TRUE
#define ilu_kernelFALSE ilu_FALSE

  typedef long int ilu_Integer;
  typedef short int ilu_ShortInteger;
  typedef ilu_longinteger ilu_LongInteger;

  typedef unsigned long int ilu_Cardinal;
  typedef unsigned short int ilu_ShortCardinal;
  typedef ilu_longcardinal ilu_LongCardinal;

  typedef double ilu_Real;
  typedef float ilu_ShortReal;
  typedef ilu_longreal ilu_LongReal;

  typedef unsigned short ilu_Character;
  typedef unsigned char ilu_ShortCharacter;

  typedef char * ilu_CString;
  typedef char * ilu_T_CString;
  typedef ilu_Character *ilu_WString;
  typedef unsigned char ilu_Byte;

  typedef char *ilu_Exception;		/* address of exception description */
  typedef ilu_Object ilu_KernelObject;

struct _ilu_LongInteger_s {
  ilu_Integer high;
  unsigned long int low;
};

struct _ilu_LongCardinal_s {
  ilu_Cardinal high;
  ilu_Cardinal low;
};

class ilu {
public:
  static ilu_Exception	Success;
  static ilu_Exception	ProtocolError;

  static ilu_Boolean	ParseSBH		(ilu_CString sbh, ilu_CString *oid, ilu_CString *contact_info);
  static ilu_Boolean	ParseOID		(ilu_CString oid, ilu_CString *iid, ilu_CString *sid);
  static ilu_Class	FindClassFromTypeName	(char *name);
  static ilu_Class	FindClassFromID		(char *id);



//private:  (the rest is private to ILU)

  static ilu_Call	BeginCall		(ilu_KernelObject obj);
  static ilu_Boolean 	BeginException		(ilu_Call call, ilu_Cardinal eIndex, ilu_Cardinal argSize);
  static ilu_Boolean	BeginRequest		(ilu_Call call, ilu_Class pclass, ilu_Method method, ilu_Cardinal argSize);
  static ilu_Boolean	BeginReply		(ilu_Call call, ilu_Boolean exceptions, ilu_Cardinal argSize);
  static ilu_Boolean	FinishCall		(ilu_Call call);
  static ilu_Boolean	FinishException 	(ilu_Call call);
  static ilu_Boolean	FinishRequest		(ilu_Call call);
  static ilu_Boolean	FinishReply		(ilu_Call call);
  static ilu_Boolean	FinishParameters	(ilu_Call call, void *obj);
  static void		NoReply			(ilu_Call call);
  
  static ilu_Exception	ExceptionOfMethod	(ilu_Method method, ilu_Cardinal index);
  static void		EnterServer		(ilu_Server ks, ilu_Class c);
  static void		ExitServer		(ilu_Server ks, ilu_Class c);
  static void *		SBHToObject		(ilu_CString sbh, ilu_CString optional_typeid, ilu_Class putative_class);
  static ilu_CString	SBHOfObject		(ilu_KernelObject obj);
  static ilu_Class	GetObjectClass		(ilu_KernelObject obj);
  static ilu_Server	GetObjectServer		(ilu_KernelObject obj);
  static ilu_KernelObject	CreateTrueKernelObject	(char *oid, ilu_Server server, ilu_Class c, void * lspo);
  static class iluObject *	GetLanguageSpecificObject	(ilu_KernelObject obj);
  static void		SetLanguageSpecificObject	(ilu_KernelObject obj, class iluObject * lspo);
  static void		RegisterClass		(ilu_Class c);
  
  static void		EndSequence		(ilu_Call call);
  static void		EndUnion		(ilu_Call call);
  static void		EndArray		(ilu_Call call);
  static void		EndRecord		(ilu_Call call);
  
  static unsigned char	InputByte		(ilu_Call call, ilu_Byte *b);
  static ilu_Boolean	InputBoolean		(ilu_Call call, ilu_Boolean *b);
  static ilu_Boolean	InputOptional		(ilu_Call call, ilu_Boolean *b);
  static ilu_Cardinal	InputCardinal		(ilu_Call call, ilu_Cardinal *b);
  static ilu_Character	InputCharacter		(ilu_Call call, ilu_Character *c);
  static ilu_ShortCardinal	InputEnum	(ilu_Call call, ilu_ShortCardinal *c);
  static ilu_Integer	InputInteger		(ilu_Call call, ilu_Integer *c);
  static ilu_Real		InputReal	(ilu_Call call, ilu_Real *);
  static ilu_ShortCardinal	InputShortCardinal	(ilu_Call call, ilu_ShortCardinal *);
  static ilu_ShortInteger	InputShortInteger	(ilu_Call call, ilu_ShortInteger *);
  static ilu_ShortReal	InputShortReal	(ilu_Call call, ilu_ShortReal *);
  static ilu_LongCardinal	InputLongCardinal	(ilu_Call call, ilu_LongCardinal *);
  static ilu_LongInteger	InputLongInteger	(ilu_Call call, ilu_LongInteger *);
  static ilu_LongReal	InputLongReal	(ilu_Call call, ilu_LongReal *);
  static ilu_CString	InputString	(ilu_Call call, ilu_CString buf, ilu_Cardinal *len, ilu_Cardinal limit);
  static ilu_CString	InputStringVec	(ilu_Call call, ilu_CString buf, ilu_Cardinal len);
  static ilu_WString	InputWString	(ilu_Call call, ilu_WString buf, ilu_Cardinal *len, ilu_Cardinal limit);
  static ilu_WString	InputWStringVec	(ilu_Call call, ilu_WString buf, ilu_Cardinal len);
  static unsigned char*	InputBytes	(ilu_Call call, unsigned char *buf, ilu_Cardinal *len, ilu_Cardinal limit);
  static unsigned char*	InputOpaque	(ilu_Call call, unsigned char *buf, ilu_Cardinal len);
  
  static ilu_KernelObject	InputObjectID	(ilu_Call call, ilu_Boolean discriminator_p, ilu_Class putative_class);
  
  static ilu_Cardinal	InputSequence	(ilu_Call call, ilu_Cardinal *count, ilu_Cardinal limit);
  static ilu_ShortCardinal	InputUnion	(ilu_Call call, ilu_ShortCardinal *discriminator);
  static ilu_Boolean	InputArray	(ilu_Call call);
  static ilu_Boolean	InputRecord	(ilu_Call call);
  
  static ilu_Boolean	OutputByte	(ilu_Call call, unsigned char byte);
  static ilu_Boolean	OutputBoolean	(ilu_Call call, ilu_Boolean b);
  static ilu_Boolean	OutputOptional	(ilu_Call call, ilu_Boolean present);
  static ilu_Boolean	OutputCardinal	(ilu_Call call, ilu_Cardinal val);
  static ilu_Boolean	OutputCharacter	(ilu_Call call, ilu_Character val);
  static ilu_Boolean	OutputEnum	(ilu_Call call, ilu_ShortCardinal val);
  static ilu_Boolean	OutputInteger	(ilu_Call call, ilu_Integer val);
  static ilu_Boolean	OutputReal	(ilu_Call call, double val);
  static ilu_Boolean	OutputShortCardinal	(ilu_Call call, ilu_ShortCardinal val);
  static ilu_Boolean	OutputShortInteger	(ilu_Call call, ilu_ShortInteger val);
  static ilu_Boolean	OutputShortReal	(ilu_Call call, float val);
  static ilu_Boolean	OutputLongCardinal	(ilu_Call call, ilu_LongCardinal val);
  static ilu_Boolean	OutputLongInteger	(ilu_Call call, ilu_LongInteger val);
  static ilu_Boolean	OutputLongReal	(ilu_Call call, ilu_LongReal val);
  static ilu_Boolean	OutputString	(ilu_Call call, ilu_CString buf, ilu_Cardinal len, ilu_Cardinal limit);
  static ilu_Boolean	OutputStringVec	(ilu_Call call, ilu_CString buf, ilu_Cardinal len);
  static ilu_Boolean	OutputWString	(ilu_Call call, ilu_WString buf, ilu_Cardinal len, ilu_Cardinal limit, ilu_CString *buf2, ilu_Cardinal *buf2len);
  static ilu_Boolean	OutputWStringVec(ilu_Call call, ilu_WString buf, ilu_Cardinal len, ilu_CString *buf2, ilu_Cardinal *buf2len);
  static ilu_Boolean	OutputBytes	(ilu_Call call, unsigned char * bytes, ilu_Cardinal len, ilu_Cardinal limit);
  static ilu_Boolean	OutputOpaque	(ilu_Call call, unsigned char * buf, ilu_Cardinal len);
  
  static ilu_Boolean	OutputObjectID	(ilu_Call call, ilu_KernelObject obj, ilu_Boolean discriminator_p, ilu_Class c);
  
  static ilu_Boolean	OutputSequence	(ilu_Call call, ilu_Cardinal length, ilu_Cardinal limit);
  static ilu_Boolean	OutputUnion	(ilu_Call call, ilu_ShortCardinal discriminator);
  static ilu_Boolean	OutputArray	(ilu_Call call);
  static ilu_Boolean	OutputRecord	(ilu_Call call);
  
  static ilu_Cardinal	SizeOfByte	(ilu_Call call, unsigned char byte);
  static ilu_Cardinal	SizeOfBoolean	(ilu_Call call, ilu_Boolean b);
  static ilu_Cardinal	SizeOfOptional	(ilu_Call call, ilu_Boolean present);
  static ilu_Cardinal	SizeOfCardinal	(ilu_Call call, ilu_Cardinal val);
  static ilu_Cardinal	SizeOfCharacter	(ilu_Call call, ilu_Character val);
  static ilu_Cardinal	SizeOfEnum	(ilu_Call call, ilu_ShortCardinal val);
  static ilu_Cardinal	SizeOfInteger	(ilu_Call call, ilu_Integer val);
  static ilu_Cardinal	SizeOfReal	(ilu_Call call, double val);
  static ilu_Cardinal	SizeOfShortCardinal	(ilu_Call call, ilu_ShortCardinal val);
  static ilu_Cardinal	SizeOfShortInteger	(ilu_Call call, ilu_ShortInteger val);
  static ilu_Cardinal	SizeOfShortReal	(ilu_Call call, float val);
  static ilu_Cardinal	SizeOfLongCardinal	(ilu_Call call, ilu_LongCardinal val);
  static ilu_Cardinal	SizeOfLongInteger	(ilu_Call call, ilu_LongInteger val);
  static ilu_Cardinal	SizeOfLongReal	(ilu_Call call, ilu_LongReal val);
  static ilu_Cardinal	SizeOfString	(ilu_Call call, ilu_CString buf, ilu_Cardinal len, ilu_Cardinal limit);
  static ilu_Cardinal	SizeOfStringVec	(ilu_Call call, ilu_CString buf, ilu_Cardinal len);
  static ilu_Cardinal	SizeOfWString	(ilu_Call call, ilu_WString buf, ilu_Cardinal len, ilu_Cardinal limit, ilu_CString *buf2, ilu_Cardinal *buf2len);
  static ilu_Cardinal	SizeOfWStringVec(ilu_Call call, ilu_WString buf, ilu_Cardinal len, ilu_CString *buf2, ilu_Cardinal *buf2len);
  static ilu_Cardinal	SizeOfBytes	(ilu_Call call, unsigned char * bytes, ilu_Cardinal len, ilu_Cardinal limit);
  static ilu_Cardinal	SizeOfOpaque	(ilu_Call call, unsigned char * buf, ilu_Cardinal len);
  
  static ilu_Cardinal	SizeOfObjectID	(ilu_Call call, ilu_KernelObject obj, ilu_Boolean discriminator_p, ilu_Class c);
  
  static ilu_Cardinal	SizeOfSequence	(ilu_Call call, ilu_Cardinal length, ilu_Cardinal limit);
  static ilu_Cardinal	SizeOfUnion	(ilu_Call call, ilu_ShortCardinal discriminator);
  static ilu_Cardinal	SizeOfArray	(ilu_Call call);
  static ilu_Cardinal	SizeOfRecord	(ilu_Call call);
  
  static void		SetDefaultServer (class iluServer * server);
  static class iluServer * GetDefaultServer ();
  
  static ilu_Port	CreatePort	(ilu_Server server,
					 char *protocolType,
					 char *transportType);
  static void	SetServerDefaultPort	(ilu_Server s, ilu_Port p);

  static ilu_ProtocolException	WaitForReply	(ilu_Call call, ilu_Cardinal *successCode);
  static ilu_ReceiveRequestStatus ReceiveRequest (ilu_Connection conn, ilu_Call *pcall, ilu_Class *pclass, ilu_Method *method, ilu_Cardinal *serial_number);
  
  static ilu_Connection	HandleNewConnection	(ilu_Port port);
  static int		FileDescriptorOfConnection	(ilu_Connection conn);
  static int		FileDescriptorOfMooringOfPort	(ilu_Port port);
  static void		CloseConnection		(ilu_Connection conn);

  static void 		RunMainLoop		(int *stop);
  static void		ExitMainLoop		(int *stop);
  static ilu_Boolean 	RegisterInputHandler	(int fd, void (*handlerProc)(int fd, void *rock), void *rock);
  static ilu_Boolean	UnregisterInputHandler	(int fd);
  static ilu_Boolean 	RegisterOutputHandler	(int fd,void (*handlerProc)(int fd, void *rock),void *rock);
  static ilu_Boolean	UnregisterOutputHandler	(int fd);
};

class iluServer;

class iluObject {
public:
			iluObject();
			iluObject(ilu_Server);
			iluObject(ilu_Server, char *instance_handle);
			iluObject(char *instance_handle);
  virtual 		~iluObject();

  virtual ilu_CString	ILUStringBindingHandle	();
  virtual ilu_Boolean	ILUPublish		();
  virtual ilu_Boolean	ILUWithdraw		();
  static  void *	Lookup			(char *oid, ilu_Class pclass);
  char *		ILUClassName		(void);
  char *		ILUClassId		(void);

  // An implementer of true objects is responsible for either
  // overriding this method or calling SetDefaultServer.
  virtual iluServer *	ILUGetServer		(void);

  ilu_Class		ILUInstanceClassRecord;

  static  void		_ILU_RegisterAsGCCallback	(class iluObject *);  // ugly hack to get the RPCObject

//private:  (the rest is private to ILU)

  static  iluObject *	CreateFromRegistry	(ilu_Class c, ilu_KernelObject);
  static  void		RegisterClass		(ilu_Class c,
						 class iluObject * (*proc)(ilu_KernelObject));
  static  void *	InputObject		(ilu_Call call, ilu_Boolean discriminator_p, ilu_Class putative_class);
  static  ilu_Boolean	OutputObject		(ilu_Call call, iluObject *obj, ilu_Class putative_class);
  static  ilu_Cardinal	SizeOfObject		(ilu_Call call, iluObject *obj, ilu_Class putative_class);

  virtual char *	ILUGetInstanceHandle	();
  virtual ilu_KernelObject ILUGetRPCObject	();
  virtual void		ILUSetRPCObject		(ilu_KernelObject obj);
  
  virtual ilu_Server	ILUGetKernelServer	(void);

  virtual void *	ILUCastDown		(ilu_Class cast_to);
  ilu_KernelObject	ILUEnsureKernelObject	();
  ilu_Server		ILUEnsureKernelServer	();
  void *		ILUMostSpecificObject	() { return (this->ILUMostSpecificObj); };
  void			ILUSetMostSpecificObject (void *o) { this->ILUMostSpecificObj = o; };

private:

  ilu_KernelObject	ILURPCObject;
  ilu_Server		ILURPCServer;
  char *		ILUPublishProof;
  char *		ILUInstanceHandle;
  void *		ILUMostSpecificObj;

};

typedef void * ilu_Alarm;
/* an alarm is an active object which can be set to invoke a procedure
   with an argument "rock" at a specified time, asynchronously. */

class iluObjectTable {
 public:

  virtual ~iluObjectTable();

  virtual iluObject *	ObjectOfIH (ilu_CString ih) = 0;

};

class iluMainLoop {
 public:

  virtual void		Run (int *stop) = 0;
  virtual void		Exit (int *stop) = 0;
  virtual ilu_Boolean	RegisterInputHandler (int fd,
					void (*handlerProc)(int,void *),
					void *handlerArg) = 0;
  virtual ilu_Boolean	UnregisterInputHandler (int fd) = 0;
  virtual ilu_Boolean	RegisterOutputHandler (int fd,
					void (*handlerProc)(int,void *),
					void *handlerArg) = 0;
  virtual ilu_Boolean	UnregisterOutputHandler (int fd) = 0;
  virtual ilu_Alarm	CreateAlarm() = 0;
  virtual void		SetAlarm(ilu_Alarm alarm, ilu_FineTime t, void (*proc)(void *rock), void *rock) = 0;
  virtual void		ClearAlarm (ilu_Alarm alarm) = 0;
};

class iluServer {
 public:

			/* AC: changed to not have defaults, but CFront
			 * can't grok them.
			 */
			iluServer(char *serviceID /* = NULL */,
				  iluObjectTable *objtab /* = NULL */);
			  /* Initializes the server, with 0 ports.
			     You'll need to call AddPort to make the
			     server reachable.  If a NULL object table
			     is given, a new one with a default
			     implementation is used. Caller owns serviceID. */

  virtual ilu_Boolean	AddPort (char *protocolType,
				 char *transportType,
				 ilu_Boolean be_default);
			  /* Adds another port to an existing server.
			     If "be_default" or this is the first port
			     added to the server, the new port will become
			     the default port for this server. */

  virtual ilu_Server	KernelServer(void);

  static  void		Run(void);
			  /* This is the main, outer loop for a server
			     program.  It is what animates all the
			     true servers. It never returns. */

  static  ilu_Boolean	RegisterInputHandler (int fd,
					void (*handlerProc)(int,void *),
					void *handlerArg);
			  /* Call this to extend the dispatch in the
			     main loop implemented by Run. */

  static  ilu_Boolean	UnregisterInputHandler (int fd);
			  /* Call this to cancel an extension installed
			     by a call on RegisterInputHandler. */

  static  void		iluSetMainLoop(iluMainLoop *ml);
			  /* If your application needs to alter the main
			     loop used by ILU, call this.  Call it, if
			     at all, before using the previous three
			     members. */

  static  ilu_Boolean	SetFork(void (*fork)(void (*proc)(void *arg),
					void *arg));

private:

  ilu_Server ks;	  /* The kernel's rep. of this server. */

};

#endif				/* ifndef __ilu_H_ */
