/*
** Copyright (c) 1993, 1994 Xerox Corporation.  All Rights Reserved.
**
** Unlimited use, reproduction, and distribution of this software is
** permitted.  Any copy of this software must include both the above
** copyright notice of Xerox Corporation and this paragraph.  Any
** distribution of this software must comply with all applicable United
** States export control laws.  This software is made available AS IS,
** and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
** INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
** AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
** PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
** THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
** CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
** XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: ilu.c,v 1.44 1995/03/03 01:30:21 janssen Exp $ */
/* Last tweaked by Mike Spreitzer February 17, 1995 9:59 am PST */

#include <stdio.h>
#include <string.h>

#include "iluchdrs.h"
 
#define  TRUE  1
#define  FALSE 0
#define  AND   &&
#define  OR    ||
#define  NOT   !

#define OPTIONAL(x)	x
#define PASS(x)		x
#define RETAIN(x)	x
#define GLOBAL(x)	x
 
#define SEQUENCE_INCREMENT 5

#define MOST_SPECIFIC_ILU_CLASS(cobj)	((cobj)->type)[0]->c
#define ILU_CLASS_OF_C_CLASS(cclass)	(cclass)[0]->c

/* L1, L2, Main unconstrained */
static ilu_boolean ILU_C_Logging = ilu_FALSE;

static ilu_Server _ILU_C_InitializeDefaultServer(void);

void ilu_CString__Free (ilu_CString s)
{
  if (s != ILU_NIL)
    ilu_free(s);
}

void ilu_WString__Free (ilu_WString s)
{
  if (s != ILU_NIL)
    ilu_free(s);
}

struct typeRecord_s {
  ilu_Class iluclass;
  ILU_C_Type cclass;
  struct typeRecord_s *next;
};

typedef struct typeRecord_s *typeRecord;

static typeRecord TypeRegistry = ILU_NIL;

void _ILU_C_RegisterSurrogateCType (ilu_Class c, ILU_C_Type t)
{
  typeRecord r;

  r = (typeRecord) ilu_must_malloc(sizeof(*r));
  r->iluclass = c;
  r->cclass = t;
  r->next = TypeRegistry;
  TypeRegistry = r;
}

/* Inside(obj->server, obj->class) */
static ILU_C_Object *_ILU_C_CreateSurrogateFromRegistry (ilu_Class c, ilu_Object obj)
{
  typeRecord p;
  ILU_C_Object *lspo = ILU_NIL;

  for (p = TypeRegistry;  p != ILU_NIL;  p = p->next)
    if (c == p->iluclass)
      break;
  if (p != ILU_NIL)
    {
      lspo = ( ILU_C_Object * ) ilu_must_malloc( sizeof( ILU_C_Object ));
      lspo->server = ilu_ServerOfObject(obj);
      lspo->instanceId = ( ilu_Object ) obj;
      lspo->nmsHandle = ILU_NIL;
      lspo->data = 0;
      lspo->type = p->cclass;
      ilu_RegisterLanguageSpecificObject(obj, lspo);
    }
  else
    {
      fprintf (stderr, "ILU_C:  attempt to create surrogate instance of class \"%s\", but surrogate code is not loaded.\n", c->cl_name);
    }
  return (lspo);
}
 
ilu_CString ILU_C_ClassName (ILU_C_Object *o)
{
  ilu_Class c = ILU_C_ClassRecordOfInstance (o);
  if (o == ILU_NIL)
    return (ILU_NIL);
  else
    return (c->cl_name);
}

ilu_CString ILU_C_ClassID (ILU_C_Object *o)
{
  ilu_Class c = ILU_C_ClassRecordOfInstance (o);
  if (o == ILU_NIL)
    return (ILU_NIL);
  else
    return (c->cl_unique_id);
}

ilu_Class ILU_C_FindILUClassByTypeName (ilu_CString typename)
{
  return (ilu_FindClassFromName(typename));
}

ilu_Class ILU_C_FindILUClassByTypeID (ilu_CString typeID)
{
  return (ilu_FindClassFromID(typeID));
}

/* L1, L2, Main unconstrained */
ilu_Class ILU_C_ClassRecordOfInstance(ILU_C_Object *o)
{
  if (o == ILU_NIL OR o->type == ILU_NIL)
    return (ILU_NIL);
  return ((o->type)[0]->c);
}

/* L1, L2, Main unconstrained */
char *_ILU_C_GetNMSHandle (ILU_C_Object *o)
{
  if ( NOT o )
    return( ILU_NIL );
  return( o->nmsHandle );
}

/* L1, L2, Main unconstrained */
void _ILU_C_SetNMSHandle(ILU_C_Object *o, char *h)
{
  if ( NOT o )
    return;
  o->nmsHandle = h;
}

/* L1, L2, Main unconstrained */
void *_ILU_C_FindMethod (ILU_C_Object *h, ilu_Class cls, int ndx)
{
  ILU_C_Type c;
  _ILU_C_MethodBlock *d;
 
  c = h->type;
  while ((d = *c++) != ILU_NIL) {
    if (d->c == cls)
      {
	return(( void * ) d->methods[ ndx ]);
      }
  }
  return(( void * ) ILU_NIL );
}
 
/* Main Invariant holds */
ilu_CString ILU_C_SBHOfObject (ILU_C_Object *obj)
{
  ilu_CString sbh = ILU_NIL;

  ilu_Object kobj = _ILU_C_KernelObjOfObj (obj);
  if (kobj != ILU_NIL)
    {
      sbh = ilu_SBHOfObject (kobj);
      ilu_ExitServer (obj->server, MOST_SPECIFIC_ILU_CLASS(obj));
    }
  return (sbh);
}

/*Main Invariant holds*/
ILU_C_Object *ILU_C_SBHToObject (ilu_CString sbh, ilu_CString ilu_type_id, ilu_Class putative_type)
{
  ILU_C_Object * h = ILU_NIL;
  ilu_Object obj;

  obj = ilu_ObjectOfSBH(sbh, ilu_type_id, putative_type);
  if (obj == ILU_NIL)
    return(ILU_NIL);
  h = ilu_GetLanguageSpecificObject(obj);
  if (h == ILU_NIL)
    h = _ILU_C_CreateSurrogateFromRegistry (ilu_ClassOfObject(obj), obj);
  ilu_ExitServer (ilu_ServerOfObject(obj), putative_type);
  return(h);
}

/* Main Invariant holds */
void
ILU_C_PingObject(ILU_C_ENVIRONMENT * env, ILU_C_Object * obj)
{
  ilu_Object      kobj = _ILU_C_KernelObjOfObj(obj);
  ilu_Class       mst;
  ilu_Error       err;
  ilu_boolean     ok;
  if (kobj == ILU_NIL) {
    ILU_C_RAISE_SYSTEM(env, BAD_PARAM, 0, NO);
    return;
  }
  mst = ilu_ClassOfObject(kobj);
  err = ilu_DeltaHolds(kobj, 1);
  ilu_ExitServer(obj->server, mst);
  ILU_ERR_SWITCH(err) {
    ILU_SUCCESS_CASE;
    ILU_ERR_CASE(GcRegFailed, v) {
      ILU_C_RAISE_SYSTEM(env, COMM_FAILURE, 0, NO);
      return;
    }
    ILU_ERR_ELSE {
      ILU_C_RAISE_SYSTEM(env, INTERNAL, 0, NO);
      return;
    }
  } ILU_ERR_ENDSWITCH;
  ok = ilu_PingObject(kobj);
  ilu_EnterServer(obj->server, mst);
  err = ilu_DeltaHolds(kobj, -1);
  ilu_ExitServer(obj->server, mst);
  ILU_ERR_SWITCH(err) {
    ILU_SUCCESS_CASE;
    ILU_ERR_CASE(GcRegFailed, v) {
      ILU_C_RAISE_SYSTEM(env, COMM_FAILURE, 0, MAYBE);
      return;
    }
    ILU_ERR_ELSE {
      ILU_C_RAISE_SYSTEM(env, INTERNAL, 0, MAYBE);
      return;
    }
  } ILU_ERR_ENDSWITCH;
  if (ok)
    env->_major = ILU_C_NO_EXCEPTION;
  else
    ILU_C_RAISE_SYSTEM(env, COMM_FAILURE, 0, MAYBE);
  return;
}

/* Main Invariant holds */
void ILU_C_DestroyObject(ILU_C_OBJECT obj)
{
  ilu_Object      kobj = _ILU_C_KernelObjOfObj(obj);
  ilu_Class       mst;
  if (kobj == ILU_NIL)
    return;
  mst = ilu_ClassOfObject(kobj);
  ilu_RegisterLanguageSpecificObject(kobj, ILU_NIL);
  obj->instanceId = ILU_NIL;
  ilu_ExitServer(obj->server, mst);
  return;
}

typedef struct {
  ilu_cardinal    ncobjs, nviobjs;
}               DestructCounters;

/* Inside(kobj's server, ilu_rootClass) */
static int
DestroyIt(ilu_Object kobj, ilu_refany rock)
{
  DestructCounters *dc = (DestructCounters *) rock;
  ILU_C_OBJECT    cobj = ilu_GetLanguageSpecificObject(kobj);
  if (cobj != ILU_NIL) {
    dc->ncobjs++;
    ilu_RegisterLanguageSpecificObject(kobj, ILU_NIL);
    cobj->instanceId = ILU_NIL;
  }
  if (ilu_VeryInterested(kobj))
    dc->nviobjs++;
  return 0;
}

/* Main Invariant holds */
void 
ILU_C_CloseServer(ILU_C_ENVIRONMENT * env, ilu_Server s,
		  ilu_cardinal * ncobjs, ilu_cardinal * nviobjs,
		  ilu_cardinal * nconns)
{
  int             res;
  DestructCounters dc = {0, 0};
  res = ilu_BankAndScanServer(s, DestroyIt, &dc, nconns);
  if (ncobjs != ILU_NIL)
    *ncobjs = dc.ncobjs;
  if (nviobjs != ILU_NIL)
    *nviobjs = dc.nviobjs;
  env->_major = ILU_C_NO_EXCEPTION;
  return;
}

/* Main Invariant holds */
void
ILU_C_DestroyObjectAndServer(ILU_C_ENVIRONMENT * env,
			     ILU_C_OBJECT obj)
{
  ilu_Server      s = obj->server;
  ILU_C_DestroyObject(obj);
  ILU_C_CloseServer(env, s, ILU_NIL, ILU_NIL, ILU_NIL);
  return;
}

/* Main Invariant holds */
ilu_boolean
ILU_C_ValidateOrDestroyObjSvr(ILU_C_ENVIRONMENT * env,
			      ILU_C_OBJECT obj)
{
  ILU_C_PingObject(env, obj);
  switch (env->_major) {
  case ILU_C_NO_EXCEPTION:
    return ilu_TRUE;
  case ILU_C_SYSTEM_EXCEPTION:
    if (ILU_C_EXCEPTION_ID(env) != ILU_C_STDEX(COMM_FAILURE))
      return ilu_FALSE;
    break;
  default:			/* can't happen */
    ILU_C_RAISE_SYSTEM(env, INTERNAL, 0, MAYBE);
    return ilu_FALSE;
  }
  ILU_C_EXCEPTION_FREE(env);
  ILU_C_DestroyObjectAndServer(env, obj);
  return ilu_FALSE;
}

/* Main Invariant holds */
ilu_CString ILU_C_PublishObject (ILU_C_Object *obj)
{
  ilu_Object kobj = _ILU_C_KernelObjOfObj (obj);
  if (kobj != ILU_NIL)
    return(ilu_PublishObject (kobj));
  else
    return (ILU_NIL);
}

/* Main Invariant holds */
ilu_boolean ILU_C_WithdrawObject (ILU_C_Object *obj, ilu_CString proof)
{
  ilu_Object kobj = _ILU_C_KernelObjOfObj (obj);
  if (kobj != ILU_NIL)
    return (ilu_WithdrawObject (kobj, proof));
  else
    return (ilu_FALSE);
}

/* Main Invariant holds; L2 otherwise unrestricted */
OPTIONAL(GLOBAL(ILU_C_Object *))
ILU_C_LookupObject(ilu_CString oid,
		   ilu_Class static_type)
{
  int             change;
  return ILU_C_ReLookupObject(oid, static_type, &change);
}

/* Main Invariant holds; L2 otherwise unrestricted */
OPTIONAL(GLOBAL(ILU_C_Object *))
ILU_C_ReLookupObject(ilu_CString oid, ilu_Class static_type, int *change)
{
  ILU_C_Object   *h = ILU_NIL;
  ilu_Object      obj;

  *change = ilu_ReLookupObject(oid, static_type, &obj);
  if (obj == ILU_NIL)
    return (ILU_NIL);
  h = ilu_GetLanguageSpecificObject(obj);
  if (h == ILU_NIL)
    h = _ILU_C_CreateSurrogateFromRegistry(ilu_ClassOfObject(obj), obj);
  ilu_ExitServer(ilu_ServerOfObject(obj), static_type);
  return (h);
}

/* Main invariant holds */

ILU_C_Object *_ILU_C_GetServerSingleton (ilu_Call call)
{
  ILU_C_Object *h;
  ilu_Object kobj;

  if ((kobj = ilu_GetCallSingleton (call)) != ILU_NIL)
    {
      h = (ILU_C_Object *) ilu_GetLanguageSpecificObject (kobj);
      ilu_ExitServer (ilu_ServerOfObject(kobj), ilu_ClassOfObject(kobj));
      return (h);
    }
  else
    return (ILU_NIL);
}

/*before: not Inside (cobj->server, cobj->type->c) */
/*after:  return != ILU_NIL => Inside(cobj->server, cobj->type->c) */
ilu_Object _ILU_C_KernelObjOfObj (ILU_C_Object *cobj)
{
  ilu_EnterServer (cobj->server, MOST_SPECIFIC_ILU_CLASS(cobj));
  if (cobj->instanceId != ILU_NIL)
    return (cobj->instanceId);
  else
    {
      ilu_ExitServer (cobj->server, MOST_SPECIFIC_ILU_CLASS(cobj));
      return (ILU_NIL);
    }
}

/* locking unconstrained */
ilu_CString _ILU_C_Strdup (ilu_CString s)
{
  if (s == ILU_NIL)
    return (ILU_NIL);
  else
    {
      ilu_CString s2 = ilu_must_malloc(strlen(s) + 1);
      strcpy (s2, s);
      return(s2);
    }
}

/* Main invariant holds */
ILU_C_Object   *
_ILU_C_CreateTrueObject(ILU_C_Type class,
			OPTIONAL(ilu_CString) instance_handle,
			OPTIONAL(ilu_Server) server,
			void *client_data)
{
  ILU_C_Object   *lspo;
  ilu_Class       iluclass;
  ilu_CString     id;
  char            idbuf[10];
  static int      idcounter = 0;

  if (server == ILU_NIL AND(server = ilu_GetDefaultServer()) == ILU_NIL)
    server = _ILU_C_InitializeDefaultServer();

  iluclass = (*class)->c;

  lspo = (ILU_C_Object *) ilu_must_malloc(sizeof(ILU_C_Object));
  lspo->type = class;
  lspo->server = server;
  lspo->nmsHandle = ILU_NIL;
  lspo->data = client_data;

  if (instance_handle != ILU_NIL)
    id = _ILU_C_Strdup(instance_handle);
  else {
    sprintf(idbuf, "%u", ++idcounter);
    id = _ILU_C_Strdup(idbuf);
  }

  ilu_EnterServer(lspo->server, iluclass);
  lspo->instanceId = ilu_FindOrCreateTrueObject(id, lspo->server,
						iluclass, lspo);
  ilu_ExitServer(lspo->server, iluclass);
  if (lspo->instanceId == ILU_NIL) {
    ilu_free(lspo);
    fprintf(stderr,
	    "ILU_C:  Can't create kernel object for true object of type \"%s\".\n",
	    iluclass->cl_name);
    return (ILU_NIL);
  } else
    return (lspo);
}

/*before: L1 = {}; L2 >= {call's conn's callmu, iomu}
  after:  same */
ILU_C_Object *_ILU_C_InputObject (ilu_Call call, ilu_Class putative_class, ilu_boolean discriminator)
{
  ilu_Object obj = ILU_NIL;
  ILU_C_Object *o;
  ilu_Class c;
 
  ilu_InputObjectID (call, &obj, discriminator, putative_class);
  if (obj == ILU_NIL)
    return (ILU_NIL);
  /* now Inside(obj->server, obj->class) */
  if ((o = ( ILU_C_Object * ) ilu_GetLanguageSpecificObject( obj )) == ILU_NIL)
    {
      if ((c = ilu_ClassOfObject(obj)) != ILU_NIL)
	o = _ILU_C_CreateSurrogateFromRegistry (c, obj);
    }
  ilu_ExitServer (ilu_ServerOfObject(obj), putative_class);
  return (o);
}
 
/*Main Invariant holds*/
ilu_boolean _ILU_C_OutputObject (ilu_Call call, ILU_C_Object *obj,
				 ilu_Class putative_class, ilu_boolean discriminator)
{
  if ( obj == ILU_NIL && putative_class->cl_singleton == ILU_NIL )
    return( ilu_FALSE );
  if (obj != ILU_NIL)
    ilu_EnterServer (obj->server, MOST_SPECIFIC_ILU_CLASS(obj));
  ilu_OutputObjectID( call, (obj == ILU_NIL) ? ILU_NIL : obj->instanceId, discriminator, putative_class );
  return( ilu_TRUE );
}
 
/* Main invariant holds */
ilu_cardinal _ILU_C_SizeOfObject (ilu_Call call, ILU_C_Object *obj, 
				  ilu_Class putative_class, ilu_boolean discriminator)
{
  ilu_cardinal sz;

  if ( obj == ILU_NIL && putative_class->cl_singleton == ILU_NIL )
    return( 0 );
  if (obj != ILU_NIL)
    ilu_EnterServer (obj->server, MOST_SPECIFIC_ILU_CLASS(obj));
  sz = ilu_SizeOfObjectID( call, (obj == ILU_NIL) ? ILU_NIL : obj->instanceId, discriminator, putative_class );
  if (obj != ILU_NIL)
    ilu_ExitServer (obj->server, MOST_SPECIFIC_ILU_CLASS(obj));
  return( sz );
}
 
/***********************************************************************/
/***********************************************************************/
/***********************************************************************/
/***********************************************************************/

/*		Server code					       */

/***********************************************************************/
/***********************************************************************/
/***********************************************************************/
/***********************************************************************/

#include <stdio.h>      /* I/O defs (including popen and pclose) */
#include <string.h>     /* strdup() */
#include <sys/types.h>
#include <errno.h>
#include <sys/errno.h>

static void     _ILU_C_ReadConnectionRequests(void *arg);
static void     _ILU_C_ReadConnectionRequest(int, ilu_private);
static void     _ILU_C_ReadServiceRequest(int, ilu_private);
static ilu_boolean 
_ILU_C_FinalServiceRequest(int fd, ilu_private arg,
			   ilu_boolean fd_significant);

extern ilu_Exception	ex_ilu_ProtocolError;

static ilu_boolean threaded = ilu_FALSE;
static ilu_boolean threadedSet = ilu_FALSE;

static void     (*Fork) (void (*proc) (void *arg), void *arg) = 0;

static char *DefaultProtocol = "sunrpc_";
static char *DefaultTransport = "tcp_localhost_0";

/* Main invariant holds */
ilu_boolean     ILU_C_SetFork(void (*fork) (void (*proc) (void *arg),
				            void *arg))
{
  if (threadedSet)
    return ilu_FALSE;
  threadedSet = ilu_TRUE;
  Fork = fork;
  threaded = ilu_TRUE;
  _ILU_C_InitializeDefaultServer();
  return (ilu_TRUE);
}

/*
 * Create a true server then open an ILU port so we've got
 * something to export objects on. This procedure should be called
 * only once.
 */
/* Main invariant holds */
ilu_Server
ILU_C_InitializeServer(char *serverID, ilu_ObjectTable obj_tab,
		       char *protocol, char *transport,
		       ilu_boolean setdefaultport)
{
  ilu_Server      server;
  ilu_Port        port;
  ilu_boolean     closed = 0;
  int             fd;
  char           *sid = (serverID == ILU_NIL) ? ilu_InventID() : serverID;
  char           *pinfo = (protocol == ILU_NIL) ? DefaultProtocol : protocol;
  char           *tinfo = (transport == ILU_NIL) ? DefaultTransport
					      : transport;
  threadedSet = ilu_TRUE;
  server = ilu_CreateTrueServer(sid, obj_tab);
  if (server == ILU_NIL) {
    fprintf(stderr,
	    "ilu.c:  Couldn't create true server from (\"%s\", %p).\n",
	    sid, obj_tab);
    exit(1);
  }
  if (protocol != ILU_NIL OR transport != ILU_NIL OR setdefaultport) {
    port = ilu_CreatePort(server, pinfo, tinfo);
    if (port == ILU_NIL) {
      fprintf(stderr,
	      "ilu.c:  Couldn't open Port to export object with protocol = \"%s\" and transport = \"%s\".\n",
	      pinfo, tinfo);
      ilu_BankServer(server);
      return (ILU_NIL);
    }
    if (setdefaultport)
      ilu_SetServerDefaultPort(server, port);
    if (threaded) {
      Fork(_ILU_C_ReadConnectionRequests, port);
    } else {
      fd = ilu_FileDescriptorOfMooringOfPort(port, &closed);
      if (NOT ilu_RegisterInputSource(fd, _ILU_C_ReadConnectionRequest,
				      port)) {
	fprintf(stderr,
	 "ilu.c:  Can't register input source for new server.\n");
	ilu_BankServer(server);
	return (ILU_NIL);
      }
    }
  }
  return (server);
}

/*
 * Create a true server then open an ILU port so we've got
 * something to export objects on.  This function should be called
 * at most once.
 */
/* Main invariant holds */
static ilu_Server
_ILU_C_InitializeDefaultServer()
{
  ilu_Server s;

  if ((s = ILU_C_InitializeServer(ILU_NIL, ILU_NIL, ILU_NIL, ILU_NIL, ilu_TRUE)) == ILU_NIL)
    {
      fprintf(stderr, "ilu.c:  Error:  Couldn't create default server.\n");
      exit(1);
    }
  ilu_SetDefaultServer(s);
  return (s);
}

/* L1, L2, Main unconstrained */
ilu_Method _ILU_C_MethodOfCall (ilu_Call call)
{
    return(ilu_MethodOfCall(call));
}

/* Main invariant holds */
static void _ILU_C_RunConnection(void *arg)
{
  ilu_Connection  conn = (ilu_Connection) arg;
  ilu_boolean     closed = ilu_FALSE;
  while (NOT closed) {
    closed = NOT ilu_BlockingWaitForInputOnConnection(conn, ILU_NIL);
    if (closed)
      break;
    closed = _ILU_C_FinalServiceRequest(0, conn, ilu_FALSE);
  }
  ilu_CloseConnection(conn);
}

extern void _ilu_WaitForInputOnFD (int, ilu_boolean *, void *);

/* Main invariant holds */
static void _ILU_C_ReadConnectionRequests(void *arg)
{
  ilu_Port        p = (ilu_Port) arg;
  ilu_boolean     sure, closed = 0;
  ilu_Connection  conn;
  int             fd;

  fd = ilu_FileDescriptorOfMooringOfPort(p, &closed);
  while (NOT closed) {
    if (ILU_C_Logging)
      printf("Mooring %p :  ", p);
    _ilu_WaitForInputOnFD(fd, &sure, ILU_NIL);
    if ((conn = ilu_HandleNewConnection(p, &closed)) == ILU_NIL) {
      if (ILU_C_Logging)
	printf("Unable to accept connection.\n");
      else
	fprintf(stderr,
		"ilu.c unable to accept connection.\n");
      return;
    } else {
      /* fork thread to run new connection */
      Fork(_ILU_C_RunConnection, conn);
    }
  }
}

/*L1, L2, Main unconstrained*/
static ilu_boolean EnableRequests(int fd, ilu_Connection conn)
{
  ilu_Server      s = ilu_ServerOfConnection(conn);
  ilu_string      sid = ilu_IDOfServer(s);
  ilu_boolean     ans;
  ans = ilu_RegisterInputSource(fd, _ILU_C_ReadServiceRequest, conn);
  if (!ans) {
    if (ILU_C_Logging)
      printf("Can't register input FD %d of conn %p server %s!\n",
	     fd, conn, sid);
    else
      fprintf(stderr,
      "ilu.c:  Can't register input FD %d of conn %p server %s!\n",
	      fd, conn, sid);
  } else if (ILU_C_Logging)
    printf("Connection %p server %s handling input on FD %d\n",
	   conn, sid, fd);
  return ans;
}

/*L1, L2, Main unconstrained*/
static ilu_boolean DisableRequests(int fd, ilu_Connection conn)
{
  ilu_Server      s = ilu_ServerOfConnection(conn);
  ilu_string      sid = ilu_IDOfServer(s);
  ilu_boolean     ans;
  ans = ilu_UnregisterInputSource(fd);
  if (!ans) {
    if (ILU_C_Logging)
      printf("Bug detected when unregistering input source FD %d of conn %p server %s!\n",
	     fd, conn, sid);
    else
      fprintf(stderr,
	      "ilu.c:  bug detected when unregistering input source FD %d for conn %p server %s!\n",
	      fd, conn, sid);
  } else if (ILU_C_Logging)
    printf("Connection %p server %s ignoring input on FD %d\n",
	   conn, sid, fd);

  return ans;
}

/* Main invariant holds */
static void
_ILU_C_ReadConnectionRequest(int fd,
			     ilu_private arg)
{
  ilu_Port        p;
  int             fd2;
  ilu_boolean     closed = 0;
  ilu_Connection  conn;

  p = (ilu_Port) arg;
  if (ILU_C_Logging)
    printf("Mooring %p (fd %d):  ", p, fd);
  if ((conn = ilu_HandleNewConnection(p, &closed)) == ILU_NIL) {
    if (ILU_C_Logging)
      printf("Unable to accept connection.\n");
    else
      fprintf(stderr, "ilu.c was unable to accept connection.\n");
    return;
  } else if ((fd2 = ilu_FileDescriptorOfConnection(conn)) < 0) {
    if (ILU_C_Logging)
      printf("Unable to get file descriptor of new conn %p!\n", conn);
    else
      fprintf(stderr,
	    "ilu.c:  Unable to get file desc of connection %p.\n",
	      conn);
    return;
  } else
    EnableRequests(fd2, conn);
}

/* Returns TRUE iff connection closed. */
/*Main Invariant holds
  before: L2 disjoint {arg's callmu, iomu}
  after:  L2     >=   {conn's callmu, iomu} if result==ilu_good_request,
  after:  L2 disjoint {conn's callmu, iomu} if result!=ilu_good_request */
/*If fd_significant: _ILU_C_ReadServiceRequest registered for fd*/
static          ilu_boolean
_ILU_C_FinalServiceRequest(int fd,
			   ilu_private arg,
			   ilu_boolean fd_significant)
{
  ilu_Call        call;
  ilu_Class       class;
  ilu_Connection  conn;
  ilu_Method      method;
  ilu_cardinal    SN;
  ilu_ReceiveRequestStatus stat;

  conn = (ilu_Connection) arg;
  if (ILU_C_Logging) {
    printf("Activity on connection %p", conn);
    if (fd_significant)
      printf(", fd %d", fd);
    printf(":  ");
  }
  if (fd_significant)
    DisableRequests(fd, conn);
  stat = ilu_ReceiveRequest(conn, &call, &class, &method, &SN);
  if (ILU_C_Logging) {
    int             e;

    if (stat == ilu_read_failed)
      e = errno;
    printf("%s",
	   (stat == ilu_good_request) ? "request"
	   : ((stat == ilu_no_request) ? "ghost input -- OK"
	      : ((stat == ilu_builtin_meth)
		 ? "builtin method, already handled -- OK"
		 : ((stat == ilu_conn_closed)
		    ? "connection closed -- OK"
		    : ((stat == ilu_read_failed)
		       ? "read failed -- Error"
		       : ((stat == ilu_not_request)
		       ? "received message not a request -- Error"
			  : ((stat == ilu_interp_failed)
		    ? "kernel request interpreter failed -- Error"
	     : "unknown ReceiveRequest error -- Bad Error")))))));
    if (stat == ilu_good_request)
      printf(", SN %lu, method \"%s\" of class \"%s\"",
	     SN, method->me_name, class->cl_name);
    else if (stat == ilu_read_failed)
      printf(", errno is %d", e);

    printf(".\n");
  }
  switch (stat) {
  case ilu_good_request:
    /* L2 >= {conn's callmu, iomu} */
    (*method->me_stubproc) (call);
    /* L2 disjoint {conn's callmu, iomu} */
    /* fd_significant == threaded */
    /* !threaded <=> _ILU_C_ReadServiceRequest registered */
    return (ilu_FALSE);
  case ilu_conn_closed:
    ilu_CloseConnection(conn);
    return (ilu_TRUE);
  case ilu_no_request:
  case ilu_builtin_meth:
  case ilu_read_failed:
  case ilu_not_request:
  case ilu_interp_failed:
    if (fd_significant)
      EnableRequests(fd, conn);
    return (ilu_FALSE);
  default:
    fprintf(stderr,
	    "\n!_ILU_C_FinalServiceRequest: Unexpected stat %d from ilu_ReceiveRequest!\n",
	    stat);
  }
  return (ilu_FALSE);
}

/* Returns TRUE iff connection closed. */
/*Main Invariant holds
  before: L2 disjoint {arg's callmu, iomu}
  after:  L2     >=   {conn's callmu, iomu} if result==ilu_good_request,
  after:  L2 disjoint {conn's callmu, iomu} if result!=ilu_good_request */
static void 
_ILU_C_ReadServiceRequest(int fd,
			  ilu_private arg)
{
  (void) _ILU_C_FinalServiceRequest(fd, arg, ilu_TRUE);
  return;
}


/*Main invariant holds
 *before: L2     >=  {call's conn's callmu, iomu}
 *after:  L2 not >= {call's conn's         iomu},
 *after:  L2     >= {call's conn's callmu} iff protocol not concurrent*/
/*If not threaded:
  before: _ILU_C_ReadServiceRequest not registered;
  after:  _ILU_C_ReadServiceRequest registered iff protocol concurrent.*/
ilu_boolean _ILU_C_FinishParameters (ilu_Call call, ILU_C_Object *obj)
{
  ilu_Connection  conn;
  conn = ilu_ConnectionOfCall(call);
  ilu_RequestRead(call);
  if ((!threaded) && ilu_ThreadPerRequest(conn))
    EnableRequests(ilu_FileDescriptorOfConnection(conn),
		   conn);
  if (obj == ILU_NIL) {
    _ILU_C_BeginException(call, 0,
			  ilu_ProtocolException_GarbageArguments);
    _ILU_C_FinishException(call);
    return (ilu_FALSE);
  } else
    return (ilu_TRUE);
}

/*L1_sup < cmu*/
/*before: L2 not >=   {call's conn's iomu},
	  L2     >=   {call's conn's callmu} iff protocol not concurrent;
   after: L2     >=   {call's conn's callmu, iomu} if result is true,
	  L2 disjoint {call's conn's callmu, iomu} if result is false. */
/*If not threaded:
  before: _ILU_C_ReadServiceRequest registered iff protocol concurrent;
  after:  _ILU_C_ReadServiceRequest not registered.*/
ilu_boolean _ILU_C_BeginReply(ilu_Call call,
			      ilu_boolean exceptions,
			      ilu_cardinal argSize)
{
  ilu_Connection  conn;
  conn = ilu_ConnectionOfCall(call);
  if ((!threaded) && ilu_ThreadPerRequest(conn))
    DisableRequests(ilu_FileDescriptorOfConnection(conn),
		    conn);
  return ilu_BeginReply(call, exceptions, argSize);
}


/*L1_sup < cmu*/
/*before: L2 not >=   {call's conn's iomu},
	  L2     >=   {call's conn's callmu} iff protocol not concurrent;
   after: L2     >=   {call's conn's callmu, iomu} if result is true,
	  L2 disjoint {call's conn's callmu, iomu} if result is false. */
/*If not threaded:
  before: _ILU_C_ReadServiceRequest registered iff protocol concurrent;
  after:  _ILU_C_ReadServiceRequest not registered.*/
ilu_boolean _ILU_C_BeginException(ilu_Call call,
				  ilu_cardinal evalue,
				  ilu_cardinal argSize)
{
  ilu_Connection  conn;
  conn = ilu_ConnectionOfCall(call);
  if ((!threaded) && ilu_ThreadPerRequest(conn))
    DisableRequests(ilu_FileDescriptorOfConnection(conn),
		    conn);
  return ilu_BeginException(call, evalue, argSize);
}

/*Main Invariant holds*/
/*L2    >=    {call's conn's callmu, iomu} before,
 *L2 disjoint {call's conn's callmu, iomu} after*/
/*If not threaded:
  before: _ILU_C_ReadServiceRequest not registered;
  after:  _ILU_C_ReadServiceRequest registered.*/
ilu_boolean _ILU_C_FinishReply(ilu_Call call)
{
  ilu_Connection  conn;
  ilu_boolean     ans;

  conn = ilu_ConnectionOfCall(call);
  ans = ilu_FinishReply(call);
  if (!threaded)
    EnableRequests(ilu_FileDescriptorOfConnection(conn),
		   conn);
  return ans;
}

/*Main Invariant holds*/
/*L2    >=    {call's conn's callmu, iomu} before,
 *L2 disjoint {call's conn's callmu, iomu} after*/
/*If not threaded:
  before: _ILU_C_ReadServiceRequest not registered;
  after:  _ILU_C_ReadServiceRequest registered.*/
ilu_boolean _ILU_C_FinishException(ilu_Call call)
{
  ilu_Connection  conn;
  ilu_boolean     ans;
  conn = ilu_ConnectionOfCall(call);
  ans = ilu_FinishException(call);
  if (!threaded)
    EnableRequests(ilu_FileDescriptorOfConnection(conn),
		   conn);
  return ans;
}

/*Main Invariant holds*/
/*before: L2 not >=   {call's conn's iomu},
	  L2     >=   {call's conn's callmu} iff protocol not concurrent;
  after:  L2 disjoint {call's conn's callmu, iomu}*/
/*If not threaded:
  before: _ILU_C_ReadServiceRequest registered iff protocol concurrent;
  after:  _ILU_C_ReadServiceRequest registered.*/
void _ILU_C_NoReply(ilu_Call call)
{
  ilu_Connection  conn;

  conn = ilu_ConnectionOfCall(call);
  ilu_NoReply(call);
  if ((!threaded) && (!ilu_ThreadPerRequest(conn)))
    EnableRequests(ilu_FileDescriptorOfConnection(conn),
		   conn);
  return;
}

/*L1_sup < cmu*/
/*before: L2 not >=   {call's conn's iomu},
	  L2     >=   {call's conn's callmu} iff protocol not concurrent;
 *after: L2 disjoint {call's conn's callmu, iomu} */
/*If not threaded:
  before: _ILU_C_ReadServiceRequest registered iff protocol concurrent;
  after:  _ILU_C_ReadServiceRequest registered.*/
void _ILU_C_UnexpectedErrorReply (ilu_Call call)
{
  if (_ILU_C_BeginException (call, 0, ilu_ProtocolException_Unknown))
    _ILU_C_FinishException (call);
}

/* Main Invariant holds */
void ILU_C_Run( )
{
    int		stop = 0;

    threadedSet = ilu_TRUE;
    ilu_RunMainLoop( &stop );
}

/* unrestricted */
unsigned int _ILU_C_SafeStrlen (ilu_CString str)
{
  if (str == ILU_NIL)
    return 0;
  else
    return (strlen((char *) str));
}

unsigned int _ILU_C_SafeWStrlen (ilu_character *str)
{
  register ilu_character *p = str;

  if (str == ILU_NIL)
    return 0;

  while (*p++ != 0)
    ;
  return (p-str);
}

/*============================================================*/
/*============================================================*/
/*                 Sequence operations                        */
/*============================================================*/
/*============================================================*/

#define Alloc( n, t )   ( t * ) ilu_must_malloc( n * sizeof( t ))
#define Realloc( p, n, t ) ( t * ) ilu_realloc( p, n * sizeof( t ))

void _ILU_C_AppendGeneric (ILU_C_Sequence h, char *p, int sz)
{
  char *ptr;

  /*
   ** place the item pointed to by p
   ** at the end of the sequence
   */

  if (h->_length >= h->_maximum)
    {
      h->_maximum = (h->_maximum + SEQUENCE_INCREMENT);
      if (h->_buffer != ILU_NIL)
	h->_buffer = ilu_realloc(h->_buffer, h->_maximum * sz);
      else
	h->_buffer = ilu_must_malloc(h->_maximum*sz);
    }
  ptr = h->_buffer + (h->_length * sz);
  memcpy (ptr, p, sz);
  h->_length += 1;
}

void _ILU_C_EveryElement(ILU_C_Sequence h, void (*proc)(void *,void *), int sz, void *data)
{
  int i;
  char *p;

  if (!h || h->_length <= 0)
    return;
  for(p = h->_buffer, i = 0; i < h->_length; i++, p += sz)
    (*proc)((void *) p, data);
}

void _ILU_C_PopGeneric (ILU_C_Sequence h, char *p, int sz)
{
    char	*ptr;

    /*
    ** return the top element
    ** in the sequence in p then
    ** remove it from the list.
    */

    if ( !h || h->_length <= 0 )
	return;
    memcpy( p, h->_buffer, sz );
    h->_length--;
    ptr = h->_buffer + sz;
    memmove (h->_buffer, ptr, h->_length * sz);
}

void _ILU_C_PushGeneric (ILU_C_Sequence h, char *p, int sz)
{
    int		l = h->_length;
    int		n;
    char	*ptr;

    /*
    ** place the item pointed to by p
    ** at the beginning of the sequence
    */

    h->_length++;
    n = h->_length * sz;
    if ( h->_length > h->_maximum ) {
        if ( h->_buffer )
    	    h->_buffer = Realloc( h->_buffer, n, char );
        else
    	    h->_buffer = Alloc( n, char );
	h->_maximum = h->_length;
    }
    ptr = h->_buffer + sz;

    memmove (ptr, h->_buffer, l * sz);
    memcpy (h->_buffer, p, sz);
}


/* L1, L2, Main unconstrained (rely on correct calls) */

static _ILU_C_MethodBlock GcCallbackMethods = { ILU_NIL, { 0 } };

static _ILU_C_MethodBlock *GcCallback__TrueTypeVector[] = {
	&GcCallbackMethods,
	ILU_NIL
};

static int CallbackInited = 0;
static ilu_Server cbs = ILU_NIL;
static ILU_C_OBJECT cbo = ILU_NIL;

/*Main invariant holds*/
void _ILU_C_EnsureGcClient(void)
{
  if (CallbackInited)
    return;
  GcCallbackMethods.c = ilu_GetGcCallbackClass();
  cbs = ILU_C_InitializeServer(ILU_NIL, ILU_NIL, ILU_NIL, ILU_NIL, ilu_TRUE);
  cbo = _ILU_C_CreateTrueObject(GcCallback__TrueTypeVector,
				"the-gc-callback", cbs, ILU_NIL);
  if (cbo != ILU_NIL) {
    ilu_EnterServer(cbs, MOST_SPECIFIC_ILU_CLASS(cbo));
    ilu_SetGcClient(cbo->instanceId);
    ilu_ExitServer(cbs, MOST_SPECIFIC_ILU_CLASS(cbo));
  }
  CallbackInited = 1;
  return;
}
