/*
Copyright (c) 1991-1995 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: authunix.c,v 1.6 1995/02/23 06:42:48 janssen Exp $ */

#include "os/posix.h"		/* for getgid(), etc. */

#include "iluntrnl.h"		/* for _ilu_Strdup() and _ilu_Hostname() */
#include "authschm.h"
#include "iluaunix.h"		/* for ilu_AS_SunRPCUnix_UserID */

static ILU_ERRS((BadCredential)) AuthUnixGetUserID (ilu_Credential, void **);

static ILU_ERRS((CantCreateCredential, BadUserID))
     AuthUnixCreateCredential (void *, ilu_Credential *);

struct _ilu_AuthenticationScheme_s _ilu_AuthenticationScheme_SunRPCUnix =
{ "sunrpc-unix", AuthUnixGetUserID, AuthUnixCreateCredential };

static ILU_ERRS((BadCredential)) AuthUnixGetUserID (ilu_Credential cred, void **u)
{
  ilu_AS_SunRPCUnix_UserID id;
  ilu_AS_SunRPCUnix_UserID oldID = (ilu_AS_SunRPCUnix_UserID) (cred->cr_data);
  int i;
  ilu_Error err;

  if (cred == NIL || (cred->cr_data == NIL) ||
      *((ilu_cardinal *) (cred->cr_data)) != ILU_AS_SUNRPCUNIX_MAGIC)
    {
      return ILU_ERR_CONS1(BadCredential, &err, c, cred, err);
    }
  if (*u == NIL)
    *u = (void *) ilu_must_malloc(sizeof(struct _ilu_AS_SunRPCUnix_UserID_s));
  id = (ilu_AS_SunRPCUnix_UserID) *u;
  id->au_magic = ILU_AS_SUNRPCUNIX_MAGIC;
  id->au_hostname = _ilu_Strdup(((ilu_AS_SunRPCUnix_UserID) (cred->cr_data))->au_hostname);
  id->au_UID = oldID->au_UID;
  id->au_GID = oldID->au_GID;
  id->au_ngids = oldID->au_ngids;
  id->au_gids = (gid_t *) ilu_must_malloc(sizeof(gid_t) * id->au_ngids);
  for (i = 0;  i < id->au_ngids;  i++)
    id->au_gids[i] = oldID->au_gids[i];
  return NIL;
}

static ILU_ERRS((CantCreateCredential, BadUserID))
     AuthUnixCreateCredential (void *userID, ilu_Credential *cred)
{
  ilu_Credential newcred;
  ilu_Error err = NULL;
  char *c;
  ilu_cardinal clen;
  ilu_AS_SunRPCUnix_UserID id = NIL;

  if (userID == NIL)
    {
      char *hostname;
      static gid_t gids[16];	/* max allowed in Sun RPC RFC */
      int ngids, i;

      hostname = _ilu_Hostname();
      ngids = getgroups(sizeof(gids)/sizeof(gid_t), gids);

      clen = sizeof(struct _ilu_AS_SunRPCUnix_UserID_s) + (ngids * sizeof(gid_t)) + strlen(hostname) + 1;
      c = (char *) ilu_must_malloc(clen);
      id = (ilu_AS_SunRPCUnix_UserID) c;
      id->au_magic = ILU_AS_SUNRPCUNIX_MAGIC;
      id->au_hostname = (c + sizeof(struct _ilu_AS_SunRPCUnix_UserID_s) + (ngids * (sizeof(gid_t))));
      strcpy (id->au_hostname, hostname);
      id->au_UID = geteuid();
      id->au_GID = getgid();
      id->au_ngids = ngids;
      id->au_gids = (gid_t *) (c + sizeof(struct _ilu_AS_SunRPCUnix_UserID_s));
      for (i = 0;  i < ngids;  i++)
	id->au_gids[i] = gids[i];
    }
  else
    {
      ilu_AS_SunRPCUnix_UserID oldID = (ilu_AS_SunRPCUnix_UserID) userID;
      int i;

      if (oldID->au_magic != ILU_AS_SUNRPCUNIX_MAGIC)
	{
	  return ILU_ERR_CONS2(BadUserID, &err, id, userID, as, &_ilu_AuthenticationScheme_SunRPCUnix, err);
	};
      clen = sizeof(struct _ilu_AS_SunRPCUnix_UserID_s) + (oldID->au_ngids * sizeof(gid_t)) + strlen(oldID->au_hostname) + 1;
      c = (char *) ilu_must_malloc(clen);
      id = (ilu_AS_SunRPCUnix_UserID) c;
      id->au_magic = ILU_AS_SUNRPCUNIX_MAGIC;
      id->au_hostname = (c + sizeof(struct _ilu_AS_SunRPCUnix_UserID_s) + (oldID->au_ngids * (sizeof(gid_t))));
      strcpy (id->au_hostname, oldID->au_hostname);
      id->au_UID = oldID->au_UID;
      id->au_GID = oldID->au_GID;
      id->au_ngids = oldID->au_ngids;
      id->au_gids = (gid_t *) (c + sizeof(struct _ilu_AS_SunRPCUnix_UserID_s));
      for (i = 0;  i < id->au_ngids;  i++)
	id->au_gids[i] = oldID->au_gids[i];
    }
  if (*cred == NIL)
    *cred = (ilu_Credential) ilu_must_malloc(sizeof(struct _ilu_Credential_s));
  newcred = *cred;
  newcred->cr_auth_scheme = &_ilu_AuthenticationScheme_SunRPCUnix;
  newcred->cr_data = (void *) c;
  newcred->cr_len = clen;
  return (err);
}



