/*
Copyright (c) 1991-1994 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: bsdutils.c,v 1.8 1995/01/24 04:15:21 janssen Exp $ */
/* Last edited by Mike Spreitzer January 5, 1995 8:27 pm PST */

#include "iluntrnl.h"
#include "oscalls.h"

#if defined(_IS_POSIX) && !defined(_IS_BSD)
#include <sys/utsname.h>	/* for struct utsname in _ilu_Hostname */
#endif

ilu_string _ilu_Hostname(void)
{
#ifdef _IS_BSD

  static ilu_string name = NIL;

  if (name == NIL)
    {
      char hostname[1000];

      _ilu_Assert(gethostname (hostname, sizeof(hostname)) == 0,
		  "gethostname failed");
      name = _ilu_Strdup(hostname);
    }
  return (name);

#elif defined(_IS_POSIX)

  ilu_string name = NIL;
  struct utsname names;

  if (OS_UNAME(&names) == 0)
    name = _ilu_Strdup(names.nodename);
  return (name);

#endif
}

#ifdef _IS_BSD

#include <sys/types.h>	/* for time() */
#include <sys/time.h>	/* for time(), gettimeofday */

/* ================ BSD UNIX ilu_FineTime ================ */

/*L1, L2, Main unconstrained*/

ilu_cardinal ilu_FineTimeRate = 1000000;

/*
"ilu_FineTime_Now" and "ilu_CoarseTime_Now" should be implemented in the same
module, because {ilu_CoarseTime_Now(), 0} needs to be the corresponding
ilu_FineTime, and both could use a different origin than 1970.0.
*/

ilu_FineTime ilu_FineTime_Now()
{
  struct timeval tv;
  ilu_FineTime ans;
  ASSERT(gettimeofday(&tv, NIL) == 0, buf,
	 (buf, "UNIX time.c:gettimeofday failed, errno=%d", errno));
  ans.ft_s = tv.tv_sec;
  ans.ft_t = tv.tv_usec;
  return ans;
}

ilu_integer ilu_CoarseTime_Now()
{
  return time(NIL);
}

/* ================= BSD UNIX InventID ======================== */

static int need_seed = 1;

/*L1_sup < smu*/
ilu_string ilu_InventID ()
{
  char buf[1000];
  
  _ilu_AcquireMutex(ilu_smu);

  if (need_seed) {
    srand (OS_GETPID());
    need_seed = 0;
  }
  sprintf(buf, "%s.%lx.%lx.%lx", _ilu_Hostname(),
	  (long unsigned) OS_GETPID(), time(NIL),
	  (long unsigned) rand());
  _ilu_ReleaseMutex(ilu_smu);
  return _ilu_Strdup(buf);
}

#endif /* _IS_BSD */

