/*
Copyright (c) 1991-1995 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: iluntrnl.h,v 1.72 1995/02/23 23:01:08 spreitze Exp $ */
/* Last edited by Mike Spreitzer February 23, 1995 2:25 pm PST */

#ifndef _ILU_INTERNALS_
#define _ILU_INTERNALS_

#include <iluxport.h>

#include <stdio.h>
#include <string.h>

#define NIL ILU_NIL
#define NULLFN 0
#define NULLCH 0

#if (defined ( __STDC__ ) || defined ( __cplusplus ))
/* So I can write ansi prototypes and still use Sun cc */
#define	_FN_DECL(type_and_name, args) type_and_name args
#else
#define _FN_DECL(type_and_name, args) type_and_name()
#endif

#define OPTIONAL(x)	x
#define PASS(x)		x
#define RETAIN(x)	x
#define GLOBAL(x)	x

/* ==================== internal type info ==================== */

typedef struct hashTable *ilu_HashTable;

#define MAX_CCTINFO_LEN 32

enum ilu_PipeDirections { ilu_PipeSink, ilu_PipeSource };

typedef enum ilu_PacketTypes {
  ilu_PacketType_Request = 0,
  ilu_PacketType_Reply = 1
} ilu_PacketType;

enum ilu_InterpretRequestStatus {
  ilu_InterpretRequest_Failed = 1,	/* couldn't interpret packet */
  ilu_InterpretRequest_Succeeded = 2,	/* successful, go ahead and
					   invoke app's method */
  ilu_InterpretRequest_Handled = 3};	/* successful, invoke built-in
					   method*/

/*L2 >= {call's connection's callmu, iomu}.
  kobj == NIL => L1, Main unconstrained.
  kobj != NIL => all the following:
  before: Inside(s, cl);
  after:				 L1 disjoint {cmu, s};
  after: cl collectible		      => L1  not >=  {gcmu};
  after: cl collectible & s surrogate => Main Invariant holds;
  where s = kobj's server and cl = kobj's type.
  (We don't really need to hold cmu for surrogate or non-collectible
   objects, but this is convenient because ilu_Enter/ExitServer can
   be used.)*/
ilu_Error _ilu_OutputObjectID (ilu_Call call, ilu_Object kobj,
			       ilu_boolean discriminator_p,
			       ilu_Class static_type);

/*kobj!=NIL => L1 >= {obj's server};
  kobj==NIL => L1 unconstrained.
  L2, Main unconstrained.*/
ilu_cardinal _ilu_SizeOfObjectID (ilu_Call call, ilu_Object kobj,
				  ilu_boolean discriminator_p,
				  ilu_Class static_type,
				  ilu_Error *err_return);

/*before: L1 = {},
  after:  *kobj!=NIL => Inside(*kobj's server, static_type);
  after:  *kobj==NIL => L1 = {};
  forall conn: (L2 >= {conn.iomu}) => (L2 >= {conn.callmu});
  L2 >= {call's connection's callmu, iomu};
  Main otherwise unconstrained*/
ilu_Error _ilu_InputObjectID (ilu_Call call, ilu_Object *kobj,
			      ilu_boolean discriminator_p,
			      ilu_Class static_type);

struct _ilu_Protocol_s {
  /* A protocol is never changed or freed once created, so these fields
   * are readonly.  The locking comments refer to invoc'ns of the methods.
   */

  ilu_boolean		pr_concurrent_requests;
  	/* Does this protocol support concurrent requests on
  	   one connection? */

  /*L2 >= {call's connection's iomu}*/
  /*L1, Main unconstrained*/

  /*L1_sup < prmu*/
  _FN_DECL(ilu_boolean (*pr_begin_request), (ilu_Call call,
					  ilu_Class intro_type,
	/* should be the class that originally defined the method */
					  ilu_Method method,
					  ilu_cardinal arg_size));

  /*Main Invariant holds*/
  _FN_DECL(ilu_boolean (*pr_finish_request), (ilu_Call call));
  /* These bracket the sending of the call parameters,
     including discriminator.  This will be followed
     by some number (0 if !call->ca_connection->co_transport->tr_class->tc_timesout)
     of calls on resend_request, and then exactly one call on
     drop_request. */

  /*Main Invariant holds*/
  _FN_DECL(ilu_boolean (*pr_resend_request), (ilu_Call call));
  /* For use with datagram transports. */

  _FN_DECL(void (*pr_drop_request), (ilu_Call call));
  /* resend_request won't be called again for this call. */

  _FN_DECL(ilu_boolean (*pr_begin_reply),
	   (ilu_Call call, ilu_boolean exceptions_possible,
	    ilu_cardinal reply_size));

  /*Main Invariant holds*/
  _FN_DECL(ilu_boolean (*pr_finish_reply), (ilu_Call call));
  /* These bracket the sending of the results.  This proc calls
     _ilu_CacheCall if transport timesout. */

  _FN_DECL(ilu_boolean (*pr_begin_exception),
	  (ilu_Call call, ilu_cardinal exception_code,
	   ilu_cardinal reply_size));

  /*Main Invariant holds*/
  _FN_DECL(ilu_boolean (*pr_finish_exception), ( ilu_Call call ));
  /* Also calls _ilu_CacheCall if transport timesout. */

  /* some functions concerned with string binding handles */
  /*L2, Main unconstrained*/
  /*L1_sup < prmu (prmu protects class<->prog&ver registry)*/
  _FN_DECL(ilu_string (*pr_form_handle),
			( ilu_Object obj));

  /* lower-level functions used by the language runtime */
  /*Main Invariant holds*/
  /*L2 >= {conn's iomu}*/

  _FN_DECL(ilu_bytes (*pr_read_packet), ( ilu_Transport bs,
	ilu_cardinal *packet_length, ilu_PacketType *packet_type,
	ilu_cardinal *serial_number  ));
  /* *packet_type normally becomes a member of ilu_PacketType;
     *serial_number gets a number identifying the call (eg, XID).
     Result is the whole protocol-level packet, including
     *packet_type and *serial_number.
     Caller owns result. */

  _FN_DECL(ilu_boolean (*pr_interpret_pipe_request),
	   (ilu_Call call, ilu_opaque packet,
	    ilu_cardinal packet_length));
  /* We don't really believe in pipes yet. */

  _FN_DECL(ilu_boolean (*pr_interpret_request),
	   (ilu_Call call, ilu_opaque packet,
	    ilu_cardinal packet_length));
  /* Fills in the intro_type and method fields of call;
     serialNumber, server, connection fields are already set.
     Stores in the call's connection's co_protocol_data slot
     some private data used for unmarshalling arguments.
     */

  /*L1, Main unconstrained*/

  _FN_DECL(void (*pr_request_read), (ilu_Call call));
  /* Called after arguments have been unmarshalled.
     Cleans up after use of call's connection's co_protocol_data slot.
     Note that call's connection may already be closed.
     */

  _FN_DECL(ilu_ProtocolException (*pr_interpret_reply),
	(ilu_Call call, ilu_opaque packet,
	 ilu_cardinal packet_length,
	 ilu_cardinal *exception_code));

  _FN_DECL(void (*pr_finish_call), (ilu_Call call));
  /* Called at the end of process of call.
     Cleans up after use of call's connection's co_protocol_data slot.
     Note that call's connection may already be closed.
     */

  /*Main Invariant holds*/
  _FN_DECL(void (*pr_flush), ( ilu_Call call ));

  /*L2, Main unconstrained*/

  _FN_DECL(ilu_refany (*pr_create_data_block), (void));

  _FN_DECL(void (*pr_free_data_block), ( ilu_refany block ));

  /* [Un]Marshalling routines */
  /*L2 >= {call's connection's iomu} for input & output,
    L2 unconstrained for sizing*/
  /*L1, Main unconstrained*/
  /*(When we cease buffering entire messages: Main Invariant holds)*/
  
  _FN_DECL(ilu_boolean (*pr_output_optional),   (ilu_Call call,
					      ilu_boolean i));
  _FN_DECL(ilu_boolean (*pr_input_optional),    (ilu_Call call,
					      ilu_boolean *i));
  _FN_DECL(ilu_cardinal (*pr_size_of_optional), (ilu_Call call,
					      ilu_boolean i));

  _FN_DECL(ilu_boolean (*pr_output_integer),   (ilu_Call call,
					     ilu_integer i));
  _FN_DECL(ilu_boolean (*pr_input_integer),    (ilu_Call call,
					     ilu_integer *i));
  _FN_DECL(ilu_cardinal (*pr_size_of_integer), (ilu_Call call,
					     ilu_integer i));

  _FN_DECL(ilu_boolean (*pr_output_cardinal),   (ilu_Call call,
					      ilu_cardinal i));
  _FN_DECL(ilu_boolean (*pr_input_cardinal),    (ilu_Call call,
					      ilu_cardinal *i));
  _FN_DECL(ilu_cardinal (*pr_size_of_cardinal), (ilu_Call call,
					      ilu_cardinal i));

  _FN_DECL(ilu_boolean (*pr_output_shortinteger),
			(ilu_Call call, ilu_shortinteger i));
  _FN_DECL(ilu_boolean (*pr_input_shortinteger),
			(ilu_Call call, ilu_shortinteger *i));
  _FN_DECL(ilu_cardinal (*pr_size_of_shortinteger),
			(ilu_Call call, ilu_shortinteger i));

  _FN_DECL(ilu_boolean (*pr_output_shortcardinal),
			(ilu_Call call, ilu_shortcardinal i));
  _FN_DECL(ilu_boolean (*pr_input_shortcardinal),
			(ilu_Call call, ilu_shortcardinal *i));
  _FN_DECL(ilu_cardinal (*pr_size_of_shortcardinal),
			(ilu_Call call, ilu_shortcardinal i));

  _FN_DECL(ilu_boolean (*pr_output_real),
			( ilu_Call call, ilu_real i ));
  _FN_DECL(ilu_boolean (*pr_input_real),
			( ilu_Call call, ilu_real *i ));
  _FN_DECL(ilu_cardinal (*pr_size_of_real),
			( ilu_Call call, ilu_real i ));

  _FN_DECL(ilu_boolean (*pr_output_shortreal),
			( ilu_Call call, ilu_shortreal i ));
  _FN_DECL(ilu_boolean (*pr_input_shortreal),
			( ilu_Call call, ilu_shortreal *i ));
  _FN_DECL(ilu_cardinal (*pr_size_of_shortreal),
			( ilu_Call call, ilu_shortreal i ));

  _FN_DECL(ilu_boolean (*pr_output_longinteger),
			(ilu_Call call, ilu_longinteger i));
  _FN_DECL(ilu_boolean (*pr_input_longinteger),
			(ilu_Call call, ilu_longinteger *i));
  _FN_DECL(ilu_cardinal (*pr_size_of_longinteger),
			(ilu_Call call, ilu_longinteger i));

  _FN_DECL(ilu_boolean (*pr_output_longcardinal),
			(ilu_Call call, ilu_longcardinal i));
  _FN_DECL(ilu_boolean (*pr_input_longcardinal),
			(ilu_Call call, ilu_longcardinal *i));
  _FN_DECL(ilu_cardinal (*pr_size_of_longcardinal),
			(ilu_Call call, ilu_longcardinal i));

  _FN_DECL(ilu_boolean (*pr_output_longreal),
			( ilu_Call call, ilu_longreal i ));
  _FN_DECL(ilu_boolean (*pr_input_longreal),
			( ilu_Call call, ilu_longreal *i ));
  _FN_DECL(ilu_cardinal (*pr_size_of_longreal),
			( ilu_Call call, ilu_longreal i ));

  _FN_DECL(ilu_boolean (*pr_output_byte),
			( ilu_Call call, ilu_byte i ));
  _FN_DECL(ilu_boolean (*pr_input_byte),
			( ilu_Call call, ilu_byte *i ));
  _FN_DECL(ilu_cardinal (*pr_size_of_byte),
			( ilu_Call call, ilu_byte i ));

  _FN_DECL(ilu_boolean (*pr_output_boolean),
			( ilu_Call call, ilu_boolean i ));
  _FN_DECL(ilu_boolean (*pr_input_boolean),
			( ilu_Call call, ilu_boolean *i ));
  _FN_DECL(ilu_cardinal (*pr_size_of_boolean),
			( ilu_Call call, ilu_boolean i ));

  /* this is a Unicode character */
  _FN_DECL(ilu_boolean (*pr_output_character),
			( ilu_Call call, ilu_character i ));
  _FN_DECL(ilu_boolean (*pr_input_character),
			( ilu_Call call, ilu_character *i ));
  _FN_DECL(ilu_cardinal (*pr_size_of_character),
			( ilu_Call call, ilu_character i ));

  /* this is an ISO Latin-1 character using the ISO 8859-1 encoding */
  _FN_DECL(ilu_boolean (*pr_output_shortchar),
			( ilu_Call call, ilu_shortcharacter i ));
  _FN_DECL(ilu_boolean (*pr_input_shortchar),
			( ilu_Call call, ilu_shortcharacter *i ));
  _FN_DECL(ilu_cardinal (*pr_size_of_shortchar),
			( ilu_Call call, ilu_shortcharacter i ));

  _FN_DECL(ilu_boolean (*pr_output_enum_code),
			( ilu_Call call, ilu_shortcardinal i  ));
  _FN_DECL(ilu_boolean (*pr_input_enum_code),
			( ilu_Call call, ilu_shortcardinal *i  ));
  _FN_DECL(ilu_cardinal (*pr_size_of_enum_code),
			( ilu_Call call, ilu_shortcardinal i  ));

  /* a string is a sequence of shortcharacter */
  _FN_DECL(ilu_boolean (*pr_output_string),
	(ilu_Call call, ilu_string s, ilu_cardinal length,
	 ilu_cardinal limit));
  _FN_DECL(ilu_boolean (*pr_input_string),
	(ilu_Call call, ilu_string *s, ilu_cardinal *length,
	 ilu_cardinal limit));
  _FN_DECL(ilu_cardinal (*pr_size_of_string),
	(ilu_Call call, ilu_string s, ilu_cardinal length,
	 ilu_cardinal limit));

  /* a stringvec is a vector of shortcharacter --
   * not null-terminated */
  _FN_DECL(ilu_boolean (*pr_output_stringvec),
	(ilu_Call call, ilu_string s, ilu_cardinal length));
  _FN_DECL(ilu_boolean (*pr_input_stringvec),
	(ilu_Call call, ilu_string *s, ilu_cardinal length));
  _FN_DECL(ilu_cardinal (*pr_size_of_stringvec),
	(ilu_Call call, ilu_string s, ilu_cardinal length));

  /* a wstring is a sequence of character */
  _FN_DECL(ilu_boolean (*pr_output_wstring),
	(ilu_Call call, ilu_wstring s, ilu_cardinal length,
	 ilu_cardinal limit));
  _FN_DECL(ilu_boolean (*pr_input_wstring),
	(ilu_Call call, ilu_wstring *s, ilu_cardinal *length,
	 ilu_cardinal limit));
  _FN_DECL(ilu_cardinal (*pr_size_of_wstring),
	(ilu_Call call, ilu_wstring s, ilu_cardinal length,
	 ilu_cardinal limit));

  /* a wstringvec is a vector of character --
   * not null-terminated */
  _FN_DECL(ilu_boolean (*pr_output_wstringvec),
	(ilu_Call call, ilu_wstring s, ilu_cardinal length));
  _FN_DECL(ilu_boolean (*pr_input_wstringvec),
	(ilu_Call call, ilu_wstring *s, ilu_cardinal length));
  _FN_DECL(ilu_cardinal (*pr_size_of_wstringvec),
	(ilu_Call call, ilu_wstring s, ilu_cardinal length));

  /* opaque is a vector of bytes */
  _FN_DECL(ilu_boolean (*pr_output_opaque),
	(ilu_Call call, ilu_opaque s, ilu_cardinal length));
  _FN_DECL(ilu_boolean (*pr_input_opaque),
	(ilu_Call call, ilu_opaque *s, ilu_cardinal length));
  _FN_DECL(ilu_cardinal (*pr_size_of_opaque),
	(ilu_Call call, ilu_opaque s, ilu_cardinal length));

  /* bytes is a sequence of bytes */
  _FN_DECL(ilu_boolean (*pr_output_bytes),
	(ilu_Call call, ilu_bytes s, ilu_cardinal length,
	 ilu_cardinal limit));
  _FN_DECL(ilu_boolean (*pr_input_bytes),
	(ilu_Call call, ilu_bytes *s, ilu_cardinal *length,
	 ilu_cardinal limit));
  _FN_DECL(ilu_cardinal (*pr_size_of_bytes),
	( ilu_Call call, ilu_bytes s, ilu_cardinal length,
	  ilu_cardinal limit));

  /* object references vary from protocol to protocol */
  
  /*L2 >= {call's connection's callmu, iomu}
    h == NIL => L1, Main unconstrained.
    h != NIL => all the following:
    before: Inside(s, cl);
    after:				   L1 disjoint {cmu, s};
    after: cl collectible		=> L1  not >=  {gcmu};
    after: cl collectible & s surrogate => Main Invariant holds;
    where s = h's server and cl = h's type.
    (We don't really need to hold cmu for surrogate or non-collectible
     objects, but this is convenient because ilu_Enter/ExitServer can
     be used.)*/
  _FN_DECL(ilu_Error (*pr_output_object_id),
	   (ilu_Call call, ilu_Object kobj, ilu_boolean discriminator_p,
	    ilu_Class static_type));

  /*before: L1 = {},
    after:  *kobj!=NIL => Inside(*kobj's server, static_type);
    after:  *kobj==NIL => L1 = {};
    forall conn: (L2 >= {conn.iomu}) => (L2 >= {conn.callmu});
    L2 >= {call's connection's callmu, iomu};
    Main otherwise unconstrained*/
  _FN_DECL(ilu_Error (*pr_input_object_id),
	   (ilu_Call call, ilu_Object *kobj, ilu_boolean discriminator_p,
	    ilu_Class static_type));

  /*kobj!=NIL => L1 >= {obj's server};
    kobj==NIL => L1 unconstrained.*/
  _FN_DECL(ilu_cardinal (*pr_size_of_object_id),
	   (ilu_Call call, ilu_Object kobj, ilu_boolean discriminator_p,
	    ilu_Class static_type, OPTIONAL(ilu_Error *) err_return));

  _FN_DECL(ilu_boolean (*pr_output_pipe),
				(ilu_Call call, ilu_Pipe i));
  _FN_DECL(ilu_boolean (*pr_input_pipe),
				(ilu_Call call, ilu_Pipe *i));
  _FN_DECL(ilu_cardinal (*pr_size_of_pipe),
				(ilu_Call call, ilu_Pipe i));

  _FN_DECL(ilu_boolean (*pr_output_sequence),
	   (ilu_Call call, ilu_cardinal length, ilu_cardinal limit));
  _FN_DECL(ilu_boolean (*pr_output_sequence_mark),
	   (ilu_Call call, ilu_cardinal extent));
	/* called every 2^16-1 items */
  _FN_DECL(ilu_boolean (*pr_input_sequence),
			(ilu_Call call, ilu_cardinal *length,
			 ilu_cardinal limit));
  _FN_DECL(ilu_boolean (*pr_input_sequence_mark),
	   (ilu_Call call, ilu_cardinal extent));
	/* called every 2^16-1 items */
  _FN_DECL(ilu_boolean (*pr_end_sequence),
			(ilu_Call call));
  _FN_DECL(ilu_cardinal (*pr_size_of_sequence),
			( ilu_Call call, ilu_cardinal length,
			  ilu_cardinal limit));

  _FN_DECL(ilu_boolean (*pr_output_array),
			( ilu_Call call, ilu_cardinal length ));
  _FN_DECL(ilu_boolean (*pr_input_array),
			( ilu_Call call ));
  _FN_DECL(ilu_boolean (*pr_end_array),
			( ilu_Call call ));
  _FN_DECL(ilu_cardinal (*pr_size_of_array),
			( ilu_Call call, ilu_cardinal length ));

  _FN_DECL(ilu_boolean (*pr_output_record),
				( ilu_Call call ));
  _FN_DECL(ilu_boolean (*pr_input_record),
				( ilu_Call call ));
  _FN_DECL(ilu_boolean (*pr_end_record),
				( ilu_Call call ));
  _FN_DECL(ilu_cardinal (*pr_size_of_record),
				( ilu_Call call ));

  _FN_DECL(ilu_boolean (*pr_output_union),
			( ilu_Call call, ilu_shortcardinal discrim ));
  _FN_DECL(ilu_boolean (*pr_input_union),
			( ilu_Call call, ilu_shortcardinal *typeIndex ));
  _FN_DECL(ilu_boolean (*pr_end_union),
			( ilu_Call call ));
  _FN_DECL(ilu_cardinal (*pr_size_of_union),
			( ilu_Call call, ilu_shortcardinal typeIndex ));

  _FN_DECL(ilu_boolean (*pr_skip_bytes),
			( ilu_Call call, ilu_cardinal numBytes  ));

};

struct _ilu_TransportClass_s {
  /* A transport class is never changed or freed once created, so
   * these fields are readonly.  The locking comments refer to
   * invocationns of the methods. */

  ilu_cardinal		tc_pFD;	/* # of FDs consumed by a port */
  ilu_cardinal		tc_cFD;	/* # of FDs consumed by a conection */
  ilu_boolean		tc_timesout; /* Retransmissions required? */
  
  /*L1, L2, Main unconstrained*/

  _FN_DECL(ilu_refany (*tc_interpret_info), (ilu_string tinfo));
  /* Interpret transport info, producing private data used in either
     a Mooring or a Transport.  The string arg is owned by the caller.
     The result is owned by the caller (who will put it in a Mooring or
     Transport). */

  _FN_DECL(ilu_string (*tc_form_info), (ilu_refany parameterblock));
  /* Produce a fully-detailed transport info, from the private data of
     either a Mooring or a Transport.  Caller owns arg & result. */

  /*Main Invariant holds*/
  _FN_DECL(ilu_boolean (*tc_connect), (ilu_Transport self));

  /*Main Invariant holds*/
  _FN_DECL(ilu_boolean (*tc_wait_for_input), (ilu_Transport bs,
					   ilu_boolean *sure,
					   ilu_FineTime *limit));
  /* Returns FALSE if error occurred, TRUE if input maybe available.
     Returns TRUE and *sure = TRUE if input or EOF certainly available. */

  /*L1_sup < trmu*/
  _FN_DECL(ilu_Mooring (*tc_create_mooring), (ilu_private mooringInfo));
  
  _FN_DECL(void (*tc_close_mooring), (ilu_Mooring m));

  /*L2 >= {conn's iomu}*/

  _FN_DECL(void (*tc_close), (ilu_Transport bs));

  _FN_DECL(ilu_integer (*tc_n_data_present), (ilu_Transport bs));
  /* positive result <=> some data available;
     negative result <=> transport broken;
     zero result <=> no more data available yet */

  /*Main Invariant holds*/

  _FN_DECL(ILU_ERRS((BadDataStructure, InvalidBufferSize, WriteError))
	   (*tc_write_bytes),
	   (ilu_Transport bs,
	    ilu_bytes message, ilu_cardinal messageLength));
  /* Caller retains ownership of message */

  _FN_DECL(ilu_bytes(*tc_get_output_buffer),
	   (ilu_Transport bs, ilu_cardinal bufferlength,
       ILU_ERRS((BadDataStructure, InvalidBufferSize, WriteError))
	    * err));
  /* used to avoid copying.  bufferlength must be less than 17. */

  _FN_DECL(ilu_boolean (*tc_read_message), (ilu_Transport bs,
	ilu_bytes *message, ilu_cardinal *messageLength ));
  /* Caller owns result. */
  
  _FN_DECL(ILU_ERRS((BadDataStructure, WriteError)) (*tc_flush_output),
	   (ilu_Transport bs));

};

struct _ilu_Transport_s {
  /*L1, L2, Main unconstrained*/

  ilu_TransportClass tr_class;	/* methods and type */
  int tr_fd;			/* every Transport has
				   a UNIX file descriptor */
  ilu_FineTime	tr_to1;		/* Initial timeout */
  ilu_FineTime	tr_toN;		/* Max timeout */
  ilu_FineTime	tr_tto;		/* Total timeout */

  ilu_refany	tr_packet;	/* for transport classes with timeouts:
				   last packet sent. */
  ilu_cardinal	tr_size;	/* size of last packet sent. */

  ilu_refany	tr_data;	/* For use by class; produced
				   by class->tc_interpret_info */
};

typedef struct {
  ilu_Connection next, prev;	/* links in doubly-linked list */
} ilu_ConnLinks;

struct _ilu_Connection_s {
  /*L1 >= {server}; L2, Main unconstrained*/

  ilu_Call co_mucall;		/* the one holding my call mutex */
  ilu_boolean co_ioing;		/* implements iomu */
  
  /*L1 unconstrained*/
  
  ilu_Protocol co_protocol;
  ilu_string co_tinfo;		/* never NIL; owned by conn; complete,
				   canonical, and < MAX_CCTINFO_LEN long
				   for incoming connections */
  ilu_string co_pinfo;		/* never NIL; conn owns; not complete */
  ilu_Transport co_transport;	/* actual byte stream */
  ilu_Port co_port;		/* NIL on client side,
				   points to port on server side */
  ilu_private co_auth_info;	/* information about the identity
				   of the caller */
  ilu_Server co_server;		/* never NIL */

  /*L2 >= {iomu}*/

  ilu_refany co_protocol_data;	/* NIL when closed */

  /* The following three fields are meaningful if the protocol is
     concurrent and the connection is outgoing. */
  ilu_private	*co_replies;		/* queue of unprocessed replies */
  
  /* The following two fields are meaningful if furthermore
     the runtime can do condition variables; NIL otherwise. */
  /*L1 >= {server}; L2 unconstrained*/

  ilu_Call co_reader;		/* the one reading input, queuing it */
  ilu_private co_cc;		/* cond var for changes under server */
  
  ilu_boolean co_closed;
  ilu_cardinal co_next_sn;	/* meaningful for outgoing connections;
				   the next request msg serial number */
  ilu_integer co_nOuts;		/* For concurrent protocols, the
				   number of calls between holding callmu
				   for request and for reply; else 0. */

  /*L1 >= {server} for [ilu_psl], which are for the doubly-linked list of
				the port's or server's connections;
   *L1 >= {cmu} for [ilu_lru],	which are for the global doubly-linked
				LRU list of idle connections. */
  ilu_ConnLinks co_links[2];	/* NIL-terminated, not circular */
  
};

typedef enum {ilu_psl, ilu_lru} ilu_ConnLinkKind;

struct _ilu_Call_s {
  /*L1, L2, Main unconstrained*/
  
  ilu_cardinal	ca_SN;		/* serial number of request */
  ilu_Server	ca_server;		/* of call */
  ilu_Class	ca_intro_type;	/* Significant in incoming calls */
  ilu_Method	ca_method;		/* ID number of method */
  ilu_Connection ca_connection;	/* connection which points
				   to (server or client) */
  ilu_refany	ca_private;	/* slot for lang-specific
				   implementation's use */
  ilu_Passport	ca_caller;	/* slot for caller's identity passport */
  ilu_Passport	ca_callee;	/* slot for callee's identity (on reply) */
};

struct _ilu_Pipe_s {
  /*L1, L2 not designed yet*/
  ilu_Mutex pi_lock;
  ilu_Mooring pi_mooring;		/* non-NIL for local pipes */
  ilu_Call pi_call;		/* wrapper around connection */
  ilu_Protocol pi_protocol;	/* protocol for pipe object */
  ilu_string pi_contact_info;	/* contact info for transport
				   connection */
  ilu_boolean pi_sink_p;		/* pipe end is a sink */
  ilu_boolean pi_valid;		/* pipe can be read/written */
  ilu_string pi_type;		/* name of type of pipe */
};

struct _ilu_Mooring_s {		/* transport-level port */
  /*L1, L2, Main unconstrained*/
  
  int mo_fd;			/* has an FD, like a Transport */
  
  /*For calling L1_sup < trmu; L2, Main unconstrained*/
  _FN_DECL(ilu_Transport (*mo_accept_connection), ( ilu_Mooring self ));
  
  ilu_TransportClass mo_transportClass;
				/* kind of transport returned by
				   accept_connection */
  ilu_private mo_data;		/* For use by transportClass; produced
				   by transportClass->tc_interpret_info */
};

typedef struct {
  char		cc_tinfo[MAX_CCTINFO_LEN]; /* Address of requestor */
  ilu_cardinal	cc_sn;
  ilu_Class	cc_intro_type;
  ilu_Method	cc_meth;
  ilu_bytes	cc_reply;
  ilu_cardinal	cc_reply_len;
} ilu_CachedCall;

struct _ilu_Port_s {		/* where a server listens
				   for connection requests */
  /*L1, L2, Main unconstrained*/
  
  ilu_Server po_server;		/* The server exported via this port. */
  ilu_string po_pinfo;		/* given at create time;
				   owned by port */
  ilu_Protocol po_protocol;	/* protocol being used */
  ilu_string po_tinfo;		/* transport info for this port;
				   owned by the port */
  ilu_Mooring po_mooring;		/* byte stream socket this port
				   listens on */
  
  /*L1 >= {server}*/
  
  ilu_boolean po_closed;
  ilu_Port po_next;		/* ..in chain of server's ports */
  ilu_ConnLinks po_connHead;	/* head of chain of conns from me */
  ilu_CachedCall *po_call_cache; /* non-NIL when transport times out */
  int po_call_cache_size;	/* non-0 when transport timesout */
  int po_call_cache_finger;	/* cache is a FIFO; next goes here */
};

struct _ilu_Server_s {
  /*L1, L2, Main unconstrained*/
  
  ilu_Mutex sr_lock;
  ilu_boolean sr_true;		/* is this a true server? */
  ilu_string sr_id;		/* part of an oid */
  
  /*L1 >= {self}*/
  
  /* The next four fields are meaningful for outgoing connections. */
  
  ilu_string sr_tinfo;		/* transport info; owned by server */
  ilu_TransportClass sr_trcls;	/* from tinfo */
  ilu_string sr_pinfo;		/* protocol info; owned by server */
  ilu_Protocol sr_protocol;	/* from pfino */
  
  ilu_boolean sr_closing;	/* ilu_BankServer sets this TRUE */
  ilu_boolean sr_cfails;	/* connect info suspect */
  ilu_ConnLinks sr_connHead;	/* meaningful for surrogates:
				   list of open connections */
  ilu_Port sr_ports;		/* head of chain of ports usable for
				   contacting this true server. */
  ilu_HashTable sr_objs;	/* maps ih -> ilu_Object,
				   including singletons;
				   goes NIL when closing &&
				   all objects deleted */
  ilu_HashTable sr_singles;	/* maps singleton ilu_Class -> ilu_Object;
				   goes NIL when closing &&
				   all objects deleted */
  ilu_ObjectTable sr_objtab;	/* significant in a true server; if not
				   NIL, the app's chance to instantiate
				   an object not found in objs */
  ilu_Port sr_default_port;	/* default port; could be NIL */
};

struct _ilu_Object_s {
  /*L1, L2, Main unconstrained*/
  
  ilu_string ob_oid;		/* ih@server-id; never NIL */
  ilu_string ob_ih;		/* id wrt server; never NIL */
  ilu_Server ob_server;		/* never NIL */
  ilu_integer ob_timeout;	/* num. sec.s after which to GC */
  ilu_Class  ob_class;
    /* Can only be NIL when doing _ilu_FindClassViaRPC.
       No locking requirements because ilu_ObjectOfSBH doesn't insert
       an object in its server's hash table 'till after class is set. */
  ilu_string ob_mstid;		/* unique_id of true object's most
				   specific type (may be NIL) */
  /*L1 >= {server}*/

  ilu_string ob_sbh;		/* really a cache, lazily eval'ed;
				   may be NIL at any time */
  ilu_private ob_lspo;		/* slot in which to store LS pointer */
  ilu_integer ob_holds;		/* Count of uses in root method stubs
				   and gc.c. */
  int ob_intNoted;			/* LSR thinks this about whether
				   kernel is very interested in me */

  /* The following fields are meaningful only for coll. surrogate objs */
  
  int ob_notifying;		/* we're calling true server now */
  int ob_known;			/* when last !notifying, did true
				   server know about this surrogate? */
  
  /* The following fields are meaningful only for collectible true objs */
  
  /*L1 >= {gcmu}*/
  
  ilu_Alarmette_s ob_gco;		/* mxamu==gcmu */
  ilu_integer ob_lastRemote;	/* when we last knew the object to be
				   loose in the network; not meaningful
				   while gclist non-empty */
  void * ob_gclist;		/* set of gc.c:counted_client*;
				   NIL may represent the empty set */
};
/*
The string fields are owned by the object.

The kernel is very interested in this object if either:
* gclist not empty (for collectible true objects);
* gclist empty and now < lastRemote+timeout (for coll. true objects);
* notifying && !known (for collectible surrogates);
* holds != 0.

As long as either:
* gclist not empty (for collectible true objects);
* gclist empty and now < lastRemote+timeout (for coll. true objects);
* notifying (for collectible surrogates);
* holds != 0;
* lspo != NIL;
the kernel is "interested" in this object, the object is in the server's hash table, and the object is not freed.

The kernel needs to make a GC callback when
	known != (lspo != NIL || holds != 0).
*/


/* ================ Old ilu.h above ================ */
/* ================ Old iluntrnl.h below ================ */

#define NOT		!
#define AND		&&
#define OR		||

#define TRUE		1
#define FALSE		0

#include <stdlib.h>
#include <errno.h>

#include "iludebug.h"

#define FREETOKEN(s) {if((s)!=NIL)ilu_free(s);}

#define PROTOCOL_TYPE(call)	(call->ca_connection->co_protocol->pr_type)

#define OBJECT_HASHTABLESIZE	113		/* number of different object slots */
#define SERVER_HASHTABLESIZE	53		/* number of different server slots */
#define CLASS_HASHTABLESIZE	53		/* number of different classes */

#define INITIAL_GCLIST_SIZE		5
#define INITIAL_GCDOBJECTSET_SIZE	200
#define INITIAL_GCCLIENTS_SIZE		20

#define GC_CALLBACK_CLASS_ID		"ilu.GCCallback"

/* ================ Reply list for putting replies on ================ */

typedef struct reply_s {
  /*L1, L2, Main unconstrained*/
  
  unsigned char * rp_packet;
  ilu_cardinal rp_len;
  ilu_cardinal rp_SN;
  
  /*L2 >= {connection's iomu}*/
  struct reply_s *rp_next;
} Reply;

/* ================ Internal Consistency Checking ================ */

void _ilu_Assert(int t, ilu_string id);
/* Code in the kernel calls this at internal consistency checks.  The first argument should be a C boolean that's true (i.e., an int that's not 0).  The second argument is some string that distinguishes the call point from every other point that calls _ilu_Assert; storage is owned by the caller.  This procedure returns iff t.  When we get our error system, this procedure will also return if !t and the caller should try to communicate the brokenness to the application. */

#define ASSERT(flg,buf,spargs) {if (!(flg)) {char buf[1000]; sprintf spargs; _ilu_Assert(0, buf);}}


/* ================ Locking routines ================ */

/*
We use simple mutual exclusion (ie, semaphores with only two states).  When we get an error system, we may have a way of expressing the interruption of an Acquire operation.
*/

/*L1, L2, Main unconstrained*/

ilu_Mutex _ilu_CreateMutex( ilu_string d1, ilu_string d2 );
/* The concatenation of d1 & d2 describes the mutex;
 * storage for them owned by caller. */

/*L1_sup < m before, L1_sup = m after*/
void _ilu_AcquireMutex( ilu_Mutex m );
/* Blocks until acquisition succeeds. */

/*L1 >= {m}*/
void _ilu_HoldMutex(ilu_Mutex m);
/* Checks that the caller holds the given mutex. */

/*L1 >= {m} before, L1 not >= {m} after*/
void _ilu_ReleaseMutex( ilu_Mutex m );
/* Releases held lock. */

void _ilu_DestroyMutex( ilu_Mutex m );
/* Note that this might enable a pending AcquireMutex operation. */

ilu_boolean _ilu_CanCondition(void);

ilu_Condition
_ilu_CreateCondition(ilu_string d1, ilu_string d2,
		     ILU_ERRS((CantCondition)) * err);

ILU_ERRS((CantCondition)) _ilu_NotifyCondition(ilu_Condition c);

ILU_ERRS((CantCondition)) _ilu_DestroyCondition(ilu_Condition c);

/* L1_sup = m */
ILU_ERRS((CantCondition)) _ilu_WaitCondition(ilu_Condition c,
					     ilu_Mutex m);


/* ================ Synthesized Locking ================ */

/*L1 >= {conn's server}*/
/*L2 as implied by name*/
/*Main unconstrained*/

ILU_ERRS((WrongLocks)) _ilu_AcquireConnIO(ilu_Connection conn);
/* WrongLocks(io) not reliably detected. */

/*L1 >= {cmu, conn's server}*/
void            _ilu_ReleaseConnIO(ilu_Connection conn);
/* Release I/O mutex.
   But first closes the connection if its server is closing. */

/*L1 >= {conn's server};
  conn->co_nOuts==0 => L1 >= {cmu}*/

ILU_ERRS((WrongLocks)) _ilu_AcquireConnCall(ilu_Connection conn,
					    ilu_Call call);
void            _ilu_ReleaseConnCall(ilu_Connection conn, ilu_Call call);
/* These also shift the connection into and out of the idle list. */
/* AcquireConnCall does not reliably detect WrongLocks(call). */

/* ================ FD & Connection Management ================ */

/*L1 >= {cmu}*/
/*L2, Main unconstrained*/

extern ilu_integer ilu_fdbudget;	/* # FDs allowed */
extern ilu_integer ilu_fdsused;		/* # FDs in use */
extern ilu_integer ilu_fdsidle;		/* # FDs idle */

extern ilu_ConnLinks ilu_idleConns;
/* Head of list of idle FD-consuming connections.  A connection is in this list while its callmu isn't held and its nOuts=0.  */

/*L1_sup = cmu*/
/*forall conn: (L2 >= {conn's iomu}) => (L2 >= {conn's callmu})*/
ILU_ERRS((WrongLocks)) _ilu_ReduceFdsTo(ilu_integer goal);
/*
 * Close idle connections until ilu_fdsused+ilu_fdsidle <= goal or
 * we run out of idle connections.  If WrongLocks(io) or
 * WrongLocks(call) is raised, some idle connections may have been
 * closed.
 */


/* ================ iluxport.h counterparts ================ */
/* These procedures are like their similarly-named iluxport.h
 * counterparts, except that they require more mutexes to be held.
 */

/*L2, Main unconstrained*/

/*L1 unconstrained*/
ilu_boolean _ilu_BlockingWaitForInputOnConnection(ilu_Connection conn,
						  ilu_boolean *sure,
						  ilu_FineTime *limit);
/* Like the public version, but also sets *sure to FALSE iff we suspect
   the input may have already been read.  Thus, when *sure is TRUE
   and NDataPresent returns 0 we can decide our peer has abandoned
   the connection (this is difficult because ioctl(,FIONREAD,) doesn't
   do us the favor of returning an error in this case). */

/*L1 >= {port's server}*/

void _ilu_ClearPortFromServer(ilu_Port port, ilu_Server s);
/* Unlink this port from its server, s.
   Called when port is closing and last connection is closed. */

ilu_boolean _ilu_ServerEmptyP (ilu_Server s);
/* True if no objects in server, and no object table defined. */

/*L1_sup = port's server; L1 >= {cmu}*/
void _ilu_ClosePort(ilu_Port port);
/* If port was the server's default, one of the server's other ports
 * becomes the default. */

/*L1 >= {cmu, conn's server}*/
/*L2 >= {conn's iomu}*/
extern void
_ilu_CloseIoingConnection(ilu_Connection conn,
			  ilu_boolean set_cfails);
/*
 * Usable on either incoming or outgoing connections. For a
 * connection of a surrogate server, set_cfails if you suspect the
 * server's connect info is bad.
 */


/* ================ Disorganized ================ */

/*L1 >= {daimu}; L2, Main unconstrained*/

extern struct _ilu_DefaultAlarm_struct _ilu_gcoDefaultAlarm_s,
                _ilu_gccDefaultAlarm_s, _ilu_iotDefaultAlarm_s,
                _ilu_grDefaultAlarm_s;
/*
 * The default implementations of the four alarms used by the
 * kernel.  Note that this header file doesn't define struct
 * _ilu_DefaultAlarm_struct; those details can be private to the
 * default alarm implementation (apparently ANSI C doesn't mind
 * declarations of partially typed variables, at least as long as
 * only the address of such a variable is used in code to which the
 * type is only partial).
 */

/*L1, L2, Main unconstrained*/

extern ilu_MainLoop _ilu_DefaultMainLoop;
/* The one that's used unless otherwise specified. */

void _ilu_AutoSetDebugLevel (void);

ilu_cardinal _ilu_SafeStrlen (ilu_string s);

void _ilu_FreeToken (ilu_refany token);

ilu_string _ilu_Strdup (const ilu_string);

ilu_string _ilu_Hostname (void);
     /* returns pointer to static string giving name of host machine */

ilu_integer _ilu_atoi (ilu_string buf, /*OPTIONAL*/ ilu_string * nextp);
     /* returns int read from buf, and if nextp is non-NIL,
	returns pointer to next non-number char in buf.
	Handles explicit bases like 0xff3, or 0303, or 0d129,
	or 0b11001101010, and uses strtol(). */

int _ilu_casefree_cmp (const ilu_string, const ilu_string);
     /* returns 0 if s1 == s2, -1 if s1 < s2, 1 if s1 > s2 */

ilu_string _ilu_Strcat3(const ilu_string s1, const ilu_string s2,
			const ilu_string s3);
/* Returns s1+s2+s3, in fresh storage; NIL may be used to input
 * an empty string.  Result will be NIL only if malloc fails. */

ilu_string _ilu_Strcat5(const ilu_string s1, const ilu_string s2,
			const ilu_string s3, const ilu_string s4,
			const ilu_string s5);
/* Like _ilu_Strcat3, only more parts. */

/*L1 >= {trmu}*/
ilu_string _ilu_tcp_CurrentHostInetName (void);

/*L1_sup < prmu, L1_sup < trmu*/
ilu_boolean _ilu_ParseConnectInfo (ilu_string handle,
				   ilu_string *protocolInfo,
				   ilu_string *transportInfo);
/* Returns TRUE iff parse successful.  Stores through non-NIL pointer
 * arguments; storage ownership of strings is returned to caller. */

/*L1, L2, Main unconstrained*/

/*L1_sup < smu*/
ilu_Server
_ilu_FindServer(ilu_string serverid, ilu_boolean add,
		ilu_string cinfo,
		ILU_ERRS((BadContactInfo, BadTransportInfo,
			  BadProtocolInfo)) * err);
/* Looks for the server with the given id.  If add and not found,
 * adds a new surrogate server of the given name and contact info.
 * Caller owns id and cinfo. */

/*L2    >=    {conn's callmu, iomu} before,
  L2 disjoint {conn's callmu, iomu} after*/
void _ilu_HandlePing (ilu_Call call);
/* Handles the built-in Ping method. */

/*L1 >= {cmu, server}*/
ILU_ERRS((BadDataStructure))
_ilu_ServerRemoveObject(ilu_Server s, ilu_Object obj);


/*Inside(server, result's type)*/
ilu_Object _ilu_FindObjectInServer(ilu_string ih, ilu_Server s);

/*L1 >= {s}*/

ilu_boolean _ilu_Addable(ilu_Server s, ilu_Class t, ilu_Object *h);

void _ilu_AddSingleton(ilu_Server s, ilu_Class t, ilu_Object o);

ILU_DECL_ERR(TInfoTooLong) {
	char *tinfo;		/* Owned by transport, not err */
} ILU_END_DECL_ERR;

ILU_ERRS((TInfoTooLong)) _ilu_CacheCall(ilu_Call call, ilu_bytes reply,
					ilu_integer len);
/* Called from a Protocol's finish_reply and _exception methods to
   cache the reply. Caller relinquishes ownership of reply. */

/*L1_sup = s; L1 >= {cmu}*/
/*L2, Main unconstrained*/

ilu_Connection _ilu_CreateConnection (ilu_Transport bs, ilu_string tinfo,
				      ilu_Protocol pr, ilu_string pinfo,
				      ilu_Port port, ilu_Server s);
/* Used internally to create both incoming and outgoing connections.
 * bs, tinfo, pr, pinfo, and s aren't NIL;
 * port is NIL iff s is a surrogate.
 * String args owned by caller.
 * Result's call and I/O mutexes are not held by any thread.
 * Caller is responsible for staying within FD budget. */

/*L1 >= {conn's server if k=psl; cmu if k=lru}*/

void _ilu_LinkConnection(ilu_ConnLinks *head, ilu_Connection conn,
			 ilu_ConnLinkKind k);

void _ilu_UnlinkConnection(ilu_ConnLinks *head, ilu_Connection conn,
			   ilu_ConnLinkKind k);

/*L1 >= {s}*/
void _ilu_ClearConnectionFromServer (ilu_Connection c, ilu_Server s);

/*L2 >= {conn.iomu}*/
/*L1, Main unconstrained*/

ilu_integer _ilu_NDataPresent(ilu_Connection conn);
/* positive<=>some, 0<=>none, negative<=>error */

/*Main Invariant holds*/
/*L2 >= {conn's iomu}*/
ilu_bytes _ilu_ReadPacket(ilu_Connection conn,
			  ilu_cardinal *packetLength,
			  ilu_PacketType *packetType,
			  ilu_cardinal *packetSN);
/* Read a call or reply.  Returns the entire protocol-level
 * packet, including *type and *serialNumber.
 * Caller owns result.
 */

/*L2, Main unconstrained*/

extern ilu_refany _ilu_ioTimeoutAlarm;

extern ilu_refany _ilu_grAlarm;	/* for use in ilu_GetReply */

/*L1_sup < prmu*/
ilu_string _ilu_ContactInfoOfConnection (ilu_Connection conn,
					 ilu_Object obj);
/* Caller owns result, which will not be NIL. */

/*L1_sup < trmu*/

ilu_TransportClass _ilu_GetTransportClassByName (ilu_string name);
/* name, owned by caller, may be a full tinfo or just the name. */

/*Main Invariant holds*/
ilu_Transport _ilu_GetTransportFromClass(ilu_TransportClass tc,
					 ilu_string tinfo);
/* Caller owns tinfo */

extern ilu_TransportClass _ilu_tcp_TransportClass(void);
extern ilu_TransportClass _ilu_udp_TransportClass(void);
extern ilu_TransportClass _ilu_xnsspp_TransportClass(void);
extern ilu_TransportClass _ilu_mem_TransportClass(void);
/* TransportClass "instantiators"; see transport.c */

/*L1_sup < prmu*/

ilu_Protocol _ilu_GetProtocolFromInfo (ilu_string pinfo);
/* Caller owns the string arg. */

/* Caller owns the string arg. */

extern ilu_Protocol _ilu_sunrpc_Protocol(void);
extern ilu_Protocol _ilu_courier_Protocol(void);
/* protocol "instantiators" (see protocol.c) */

/*Main Invariant holds*/

_FN_DECL(void _ilu_WaitForInputOnFD , (int fd, ilu_boolean *sure,
				       ilu_FineTime *limit));
_FN_DECL(void _ilu_WaitForOutputOnFD, (int fd, ilu_boolean *sure,
				       ilu_FineTime *limit));
/* These two procedures return when either (1) the appropriate kind of I/O can be done on the given file descriptor without blocking, (2) an exceptional condition exists on the FD, or (3) *limit has arrived.  NIL may be passed for limit, in which case *limit is effectively +infinity.
  In a multi-threaded program, these procedures block only the calling thread.  *sure is set, but not read.  When *sure is set true, (1) or (2) is known to hold upon return.  When *sure is set false, either (3) holds or (1) or (2) might hold.
  In a single-threaded program, these procedures process input on other FDs and the alarm while waiting.  This processing may lead to a nested call on one of these procedures, with the same or a different FD.  When I/O is finally enabled on an FD, all nested calls waiting on the same FD are unblocked; *sure is set TRUE for the innermost call, FALSE for the others.  Thus, after one of these procedures returns, the caller has to call NDataPresent (or something analogous) to tell if there's really input/output waiting.  Because ioctl(,FIONREAD,) will report success & 0 bytes available, instead of a connection-closed error, when the peer closes a TCP connection, we need the sure parameter to be able to reliably detect this situation. */

/*L2 >= {fd's connection's callmu, iomu}*/

_FN_DECL(int _ilu_Read, (int fd, ilu_byte *buf, int nbytes));
_FN_DECL(int _ilu_Write, (int fd, ilu_byte *buf, int nbytes));
/* These are like UNIX read and write, except that they iterate over multiple system calls, blocking only the calling thread, until the entire transfer has been completed or has failed.  In single-threaded runtimes, they assume there'll be no nested attempt to do the same kind of I/O on the same FD. */

/*Inside(obj's server, obj's type)*/

ILU_ERRS((BadDataStructure, KernelBroken, GcRegFailed))
_ilu_DeltaHolds(ilu_Object obj, ilu_integer dholds);

ILU_ERRS((BadDataStructure, KernelBroken)) _ilu_VIUpdate(ilu_Object obj);
/* Call this after holds or gclist's emptiness changes,
   or lastRemote+timeout passes,
   and you're ready to have the server invariant restored.
   L1 mutexes are exited and re-entered inside this procedure! */


/****************************** from type.c ********************/

/*L1, L2, Main unconstrained */

extern const ilu_Class _ilu_rootClass;
/* The one with all the methods every object supports */

extern ilu_Method _ilu_GetTypesMethod;
extern ilu_Method _ilu_RegisterGCInterestMethod;
extern ilu_Method _ilu_UnregisterGCInterestMethod;
extern ilu_Method _ilu_PingMethod;
/* The methods of _ilu_rootClass */

/*L1_sup < otmu*/
ilu_Class _ilu_FindMSKA(ilu_string tid);
/* Returns the one most specific known ancestor of the type identified
   by the given string; returns NIL if that's not well-defined, or we
   don't yet know about the ancestry of the given type. */

/*L1_sup < otmu*/
void _ilu_EnumerateClasses (void (*proc) (ilu_Class, ilu_refany rock), ilu_refany rock);
/* Calls "proc" on every registered class, passing the class and "rock".
   The order in which the classes are supplied is not specified. */

/*Main Invariant holds; L2 otherwise unconstrained*/

ilu_Class _ilu_FindClassViaRPC (ilu_Object o);
/* o->ob_class is temporarily set to some known type of o */

/*L2    >=    {conn's callmu, iomu} before,
  L2 disjoint {conn's callmu, iomu} after*/
void _ilu_HandleGetTypes (ilu_Call call);

/****************************** Hash Table ops ********************/

#include "iluhash.h"

/********************* from pipe.c ********************/

/*L1, L2 not designed*/

ilu_string _ilu_Pipe_ContactInfo (ilu_Pipe p);

/********************* from gc.c ********************/

/*L1, L2, Main unconstrained*/

extern const ilu_Class _ilu_GcCallbackClass;

/*L1 >= {gcmu}*/

_FN_DECL(void _ilu_StartGCingTrueObj, ( ilu_Object obj ));

_FN_DECL(void _ilu_StopGCingTrueObj, ( ilu_Object obj ));

/*L1 >= {gcmu, cmu, obj's server}*/
ILU_ERRS((BadDataStructure, KernelBroken))
_ilu_TouchedObj(ilu_Object obj);
/* Applicable to collectible true objects;
   call this when lastRemote or gclist's emptiness changes. */

extern ilu_refany _ilu_gcoAlarm;
extern ilu_refany _ilu_gccAlarm;

/*Main Invariant holds*/
/*L2    >=    {conn's callmu, iomu} before,
  L2 disjoint {conn's callmu, iomu} after*/

_FN_DECL(void _ilu_HandleGCInterestDeregistration, ( ilu_Call call ));
_FN_DECL(void _ilu_HandleGCInterestRegistration, ( ilu_Call call ));
/* Server stubs for two built-in methods. */

/*Main Invariant holds; L2 otherwise unconstrained*/

ILU_ERRS((GcRegFailed)) _ilu_RegisterGCInterest(ilu_Object obj);
ILU_ERRS((GcRegFailed)) _ilu_UnregisterGCInterest(ilu_Object obj);
/* Notify the true server of surrogate obj's (non-)existance. */

#endif /* ndef _ILU_INTERNALS_ */
