/*
Copyright (c) 1991, 1992, 1993 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: memory.c,v 1.6 1994/12/05 21:43:13 spreitze Exp $ */
/* Last edited by Mike Spreitzer December 5, 1994 1:37 pm PST */

#include "iluntrnl.h"
#include "oscalls.h"	/* for OS_SLEEP */

/*L1, L2, Main unconstrained*/

typedef struct ilu_FreerCons_s FreerCons, *FreerList;

struct ilu_FreerCons_s {
	FreerList next;
	void (*free)(ilu_cardinal size);
};

static FreerList freers = NIL;

ilu_boolean
ilu_AddFreer(void (*free) (ilu_cardinal size))
{
  FreerList       fl;
  DEBUG(MALLOC_DEBUG, (stderr, "ilu_AddFreer(%p)\n", free));
  fl = (FreerList) ilu_malloc(sizeof(FreerCons));
  if (fl == NIL)
    return ilu_FALSE;
  fl->free = free;
  fl->next = freers;
  freers = fl;
  return ilu_TRUE;
}

void           *
ilu_malloc(ilu_cardinal size)
{
  void           *ans;
  FreerList       fl;
  ans = malloc(size);
  DEBUG(MALLOC_DEBUG,
	(stderr, "ilu_malloc: malloc(%lu=0x%lx) => %p\n",
	 size, size, ans));
  if (ans != NIL)
    return ans;
  for (fl = freers; fl != NIL; fl = fl->next) {
    DEBUG(MALLOC_DEBUG,
	  (stderr, "ilu_malloc: trying freer %p\n", fl->free));
    fl->free(size);
    ans = malloc(size);
    DEBUG(MALLOC_DEBUG,
	  (stderr, "ilu_malloc: re-malloc(%lu=0x%lx) => %p\n",
	   size, size, ans));
    if (ans != NIL)
      return ans;
  }
  DEBUG(MALLOC_DEBUG,
	(stderr, "ilu_malloc(%lu=0x%lx) fails.\n", size, size));
  return NIL;
}

void           *
ilu_realloc(void *p, ilu_cardinal size)
{
  void           *ans;
  FreerList       fl;
  ans = realloc(p, size);
  DEBUG(MALLOC_DEBUG,
	(stderr, "ilu_realloc: realloc(%p, %lu=0x%lx) => %p\n",
	 p, size, size, ans));
  if (ans != NIL)
    return ans;
  for (fl = freers; fl != NIL; fl = fl->next) {
    DEBUG(MALLOC_DEBUG,
	  (stderr, "ilu_realloc: trying freer %p\n", fl->free));
    fl->free(size);
    ans = realloc(p, size);
    DEBUG(MALLOC_DEBUG,
	  (stderr, "ilu_realloc: realloc(%p, %lu=0x%lx) => %p\n",
	   p, size, size, ans));
    if (ans != NIL)
      return ans;
  }
  DEBUG(MALLOC_DEBUG,
   (stderr, "ilu_realloc(%p, %lu=0x%lx) fails.\n", p, size, size));
  return NIL;
}

void ilu_free(void *p)
{
  DEBUG(MALLOC_DEBUG, (stderr, "ilu_free(%p)\n", p));
  free(p);
}

static int theMFA = -1;

void ilu_SetMemFaultAction(int mfa)
{
  DEBUG(MALLOC_DEBUG,
	(stderr, "ilu_SetMemFaultAction: from %d to %d.\n", theMFA, mfa));
  theMFA = mfa;
  return;
}

void           *
ilu_must_malloc(ilu_cardinal size)
{
  void           *ans = ilu_malloc(size);
  if (ans != NIL)
    return ans;
  else {
    DEBUG(MALLOC_DEBUG,
	(stderr, "ilu_must_malloc(%lu=0x%lx) failed!  theMFA=%u\n",
	 size, size, theMFA));
    if (theMFA > 0) {
      exit(theMFA);
      return NIL;
    } else if (theMFA == -1) {
      while (1)
	OS_SLEEP(10);
    } else {
      (*(int *) NIL) = 1;	/* This had better terminate the
				 * prog */
      exit(32767);		/* ... just in case it doesn't */
      return NIL;
    }
  }
}
