/*
Copyright (c) 1991, 1992, 1993 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: simpbind.c,v 1.14 1995/02/24 00:08:38 spreitze Exp $ */
/* Last edited by Mike Spreitzer February 23, 1995 4:07 pm PST */

#ifdef MACOS
#pragma segment ilu
#endif

#include "iluntrnl.h"
#include "object.h"

#include "oscalls.h"

#define OPTIONAL(x)	x

char *ilu_BindingDirectory = ILU_BINDING_DIRECTORY;

/*before: Inside(s, cl)
  after:				 L1 disjoint {cmu, s};
  after: cl collectible		      => L1  not >=  {gcmu};
  after: cl collectible & s surrogate => Main Invariant holds;
  where s = obj's server and cl = obj's type.
  (We don't really need to hold cmu for surrogate or non-collectible
   objects, but this is convenient because ilu_Enter/ExitServer can
   be used.)*/
PASS(OPTIONAL(char *)) ilu_PublishObject (ilu_Object obj)
{
  char           *oid;
  char           *sbh;
  char            buf[1000];
  FILE           *f;
  char           *proof = NIL;

  if ((sbh = ilu_SBHOfObject(obj)) == NIL) {
    DEBUG(OBJECT_DEBUG,
	  (stderr,
	   "ilu_PublishObject:  object %p not exported.\n",
	   obj));
    return (NIL);
  } else if (!ilu_ParseStringBindingHandle(sbh, &oid, NIL)) {
    DEBUG(OBJECT_DEBUG,
	  (stderr,
      "ilu_PublishObject:  can't parse sbh <%s> for object %p.\n",
	   sbh, obj));
  } else {
    sprintf(buf, "%s/%s", ilu_BindingDirectory, oid);
    ilu_free(oid);
    if (OS_ACCESS(buf, OS_ACCESS_R_OK) == 0
	&& OS_ACCESS(buf, OS_ACCESS_W_OK) != 0) {
      DEBUG(EXPORT_DEBUG,
	    (stderr,
	     "ilu_PublishObject:  OID %s already published.\n",
	     buf));
    } else if ((f = fopen(buf, "w+")) == NIL) {
      DEBUG(EXPORT_DEBUG,
	    (stderr,
	     "ilu_PublishObject:  Can't open file %s.\n",
	     buf));
    } else {
      proof = _ilu_Strdup(ilu_InventID());
      fprintf(f, "%s\n%s\n%s\n", proof, sbh, ilu_MstidOfObject(obj));
      fclose(f);
      OS_CHMOD(buf,
	       (OS_CHMOD_S_IWUSR | OS_CHMOD_S_IRUSR
		| OS_CHMOD_S_IWGRP | OS_CHMOD_S_IRGRP
		| OS_CHMOD_S_IWOTH | OS_CHMOD_S_IROTH));
    }
  }
  ilu_ExitServer(object_server(obj), object_class(obj));
  return (proof);
}

/*before: Inside(s, cl)
  after:				 L1 disjoint {cmu, s};
  after: cl collectible		      => L1  not >=  {gcmu};
  after: cl collectible & s surrogate => Main Invariant holds;
  where s = obj's server and cl = obj's type.
  (We don't really need to hold cmu for surrogate or non-collectible
   objects, but this is convenient because ilu_Enter/ExitServer can
   be used.)*/
ilu_boolean
ilu_WithdrawObject(ilu_Object obj, PASS(char *) ownership_proof)
{
  char           *oid;
  char           *sbh;
  char            buf[1000];
  FILE           *f;
  ilu_boolean     result = ilu_FALSE;

  if ((sbh = ilu_SBHOfObject(obj)) == NIL) {
    DEBUG(OBJECT_DEBUG,
	  (stderr,
	   "ilu_WithdrawObject:  object %p not exported.\n", obj));
  } else if (!ilu_ParseStringBindingHandle(sbh, &oid, NIL)) {
    DEBUG(OBJECT_DEBUG,
	  (stderr,
     "ilu_WithdrawObject:  can't parse sbh <%s> for object %p.\n",
	   sbh, obj));
  } else {
    sprintf(buf, "%s/%s", ilu_BindingDirectory, oid);
    if (OS_ACCESS(buf, OS_ACCESS_R_OK) == 0
	&& OS_ACCESS(buf, OS_ACCESS_W_OK) != 0) {
      DEBUG(EXPORT_DEBUG,
	    (stderr,
	     "ilu_WithdrawObject:  Can't access registry file %s.\n",
	     buf));
    } else if ((f = fopen(buf, "r+")) == NIL) {
      DEBUG(EXPORT_DEBUG,
	    (stderr,
	     "ilu_WithdrawObject:  Can't open registry file %s.\n",
	     buf));
    } else {
      char            buf2[1000];

      if (fgets(buf2, sizeof(buf2), f) == NIL) {
	DEBUG(EXPORT_DEBUG,
	      (stderr,
	       "ilu_WithdrawObject:  Bad registry file %s.\n",
	       buf));
	fclose(f);
      } else if (strncmp(ownership_proof, buf2,
			 strlen(ownership_proof)) != 0) {
	DEBUG(EXPORT_DEBUG,
	      (stderr,
	       "ilu_WithdrawObject:  no match on ownership proofs with proof %s.\n",
	       ownership_proof));
	fclose(f);
      } else {
	OS_UNLINK(buf);
	fclose(f);
	result = ilu_TRUE;
      }
    }
  }
  ilu_free(ownership_proof);
  ilu_ExitServer(object_server(obj), object_class(obj));
  return result;
}

/*before: L1 = {};
  after:  result!=NIL => Inside(result's server, pclass);
  after:  result==NIL => L1 = {};
  forall conn: (L2 >= {conn.iomu}) => (L2 >= {conn.callmu});
  Main otherwise unconstrained*/
OPTIONAL(ilu_Object) ilu_LookupObject(char *oid, ilu_Class pclass)
{
  ilu_Object      ans;
  ilu_boolean     change;
  change = ilu_ReLookupObject(oid, pclass, &ans);
  return ans;
}

/*before: L1 = {};
  after:  *po != NIL => Inside(result's server, pclass);
  after:  *po == NIL => L1 = {};
  forall conn: (L2 >= {conn.iomu}) => (L2 >= {conn.callmu});
  Main otherwise unconstrained*/
ilu_boolean
ilu_ReLookupObject(char *oid,
		   ilu_Class pclass,
		   OPTIONAL(ilu_Object) * po)
{
  FILE           *f;
  char            buf[1000], proof[1000], sbh[1000], mstid[1000];
  ilu_Object      obj;
  ilu_ConsiderSbhResult csr;
  ILU_ERRS((BadSBH, BadURL, BadContactInfo,
	    BadTransportInfo, BadProtocolInfo)) err;
  ilu_boolean     ans;
  ilu_Server      s2;
  *po = NIL;
  sprintf(buf, "%s/%s", ilu_BindingDirectory, oid);
  if (OS_ACCESS(buf, OS_ACCESS_R_OK) != 0) {
    DEBUG(INCOMING_DEBUG,
	(stderr, "ilu_LookupObject:  No such object %s.\n", buf));
    return (ilu_FALSE);
  }
  if ((f = fopen(buf, "r")) == NIL) {
    DEBUG(INCOMING_DEBUG,
	  (stderr,
	   "ilu_LookupObject:  Can't open registry file %s.\n",
	   buf));
    return (ilu_FALSE);
  }
  if (fgets(proof, sizeof(proof), f) == NIL
      || fgets(sbh, sizeof(sbh), f) == NIL
      || fgets(mstid, sizeof(mstid), f) == NIL) {
    DEBUG(INCOMING_DEBUG,
	  (stderr, "ilu_LookupObject:  Bad registry file %s.\n",
	   buf));
    fclose(f);
    return (ilu_FALSE);
  }
  fclose(f);

  proof[strlen(proof) - 1] = '\0';
  sbh[strlen(sbh) - 1] = '\0';
  mstid[strlen(mstid) - 1] = '\0';

  DEBUG(INCOMING_DEBUG,
	(stderr,
   "ilu_LookupObject:  found SBH=<%s>, mstid=<%s>, pclass=<%s>\n",
	 sbh, mstid, pclass->cl_name));

  csr = ilu_ConsiderSBH(sbh, &s2, &err);
  switch (csr) {
  case ilucsr_err:
    ilu_FreeError(err);
  case ilucsr_noProblem:
  case ilucsr_isTrue:
  case ilucsr_noNews:
    ans = ilu_FALSE;
    break;
  case ilucsr_notReified:
  case ilucsr_changed:
    ans = ilu_TRUE;
    break;
  default:
    _ilu_Assert(0, "Lookup: unexpected csr");
  }

  if ((obj = ilu_ObjectOfSBH(sbh, mstid, pclass)) == NIL) {
    DEBUG(INCOMING_DEBUG,
	  (stderr, "ilu_LookupObject:  Bad object info.\n"));
    return (ilu_FALSE);
  }
  *po = obj;
  if (ilu_TrueInstanceP(obj)) {
    DEBUG(INCOMING_DEBUG,
	  (stderr, "ilu_LookupObject:  Local object.\n"));
    return (ans);
  } else {
    ilu_boolean     status;
    ilu_Server      s = object_server(obj);
    ILU_ERRS((BadDataStructure, KernelBroken, GcRegFailed)) err;
    err = ilu_DeltaHolds(obj, 1);
    ILU_ERR_SWITCH(err) {
      ILU_SUCCESS_CASE {
	ilu_ExitServer(s, pclass);
	status = ilu_PingObject(obj);
	ilu_EnterServer(s, pclass);
      }
      ILU_ERR_CASE(GcRegFailed, v)
	status = ilu_FALSE;
      ILU_ERR_CASE2(BadDataStructure, KernelBroken) {
	ilu_ExitServer(s, pclass);
	*po = NIL;
	return (ans);
      }
    } ILU_ERR_ENDSWITCH;
    if (status) {
      ilu_DHolds(obj, -1);
      return (ans);
    } else {
      int             wasonly, nobj, nconn, bankit = 0;
      wasonly = obj->ob_lspo == NIL
	&& _ilu_hash_PairsInTable(s->sr_objs) == 1;
      DEBUG(INCOMING_DEBUG,
	    (stderr,
	     "ilu_LookupObject:  Bad ping of object %s\n",
	     sbh));
      ilu_DeltaHolds(obj, -1);
      if (wasonly) {
	nobj = ilu_NumObjsInServer(s);
	nconn = ilu_NumIoingConnsOfServer(s);
	if (bankit = (nobj == 0 && nconn == 0))
	  ilu_PreBankServer(s);
      }
      ilu_ExitServer(s, pclass);
      if (bankit)
	ilu_BankServer(s);
      *po = NIL;
      return (ans);
    }
  }
}
