/*
Copyright (c) 1991, 1992, 1993 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: transprt.c,v 1.29 1995/02/12 01:46:10 janssen Exp $ */
/* Last edited by Mike Spreitzer November 30, 1994 4:02 pm PST */

#define _POSIX_SOURCE


#include "iluntrnl.h"
#include "transprt.h"

struct transports_s {
  /*For access: L1, L2, Main unconstrained.
    For calling: L1_sup < trmu; L2, Main unconstrained.*/

  ilu_string name;
  ilu_TransportClass (*instantiator)(void);
};

/*L1, L2, Main unconstrained*/

static struct transports_s transports[MAX_TRANSPORTS] = {
#ifdef UDPSOCKET_TRANSPORT
  { "udp", _ilu_udp_TransportClass },
#endif /* UDPSOCKET_TRANSPORT */
#ifdef TCPIP_TRANSPORT
  { "tcp", _ilu_tcp_TransportClass },
#endif /* TCPIP_TRANSPORT */
#ifdef XNSSPP_TRANSPORT
  { "xnsspp", _ilu_xnsspp_TransportClass },
#endif /* XNSSPP_TRANSPORT */
#ifdef INMEMORY_TRANSPORT
  { "memory", _ilu_mem_TransportClass },
#endif /* INMEMORY_TRANSPORT */
  { NIL, NULLFN } };

ILU_ERRS((TransportAlreadyRegistered, MaxCountExceeded))
ilu_RegisterTransport(char *name,
		      ilu_TransportClass(*new_transport) (void),
		      ilu_boolean override)
{
  int             i;
  ilu_Error       e;

  for (i = 0; i < MAX_TRANSPORTS && transports[i].name != NIL; i++) {
    if (strcmp(transports[i].name, name) == 0) {
      if (override)
	transports[i].instantiator = new_transport;
      else
	return ILU_ERR_CONS3(TransportAlreadyRegistered, &e,
			     name, name,
			     old_transport, transports[i].instantiator,
			     new_transport, new_transport, e);
    }
  }
  if (i < MAX_TRANSPORTS && transports[i].name == NIL) {
    transports[i].name = name;
    transports[i].instantiator = new_transport;
    if ((i + 1) < MAX_TRANSPORTS)
      transports[i + 1].name = NIL;
    return ILU_NO_ERR;
  } else
    return ILU_ERR_CONS1(MaxCountExceeded, &e,
			 max_count, MAX_TRANSPORTS, e);
}

static struct transports_s *FindTransport (ilu_string pinfo)
{
  char buf[1000];
  ilu_integer i;

  if (pinfo != NIL)
    {
      ilu_string p;

      p = strchr(pinfo, '_');
      if (p == NIL)
	strcpy (buf, pinfo);
      else
	{
	  strncpy (buf, pinfo, p - pinfo);
	  buf[p - pinfo] = '\0';
	}
      for (i = 0;  transports[i].name != NIL;  i += 1)
	if (_ilu_casefree_cmp(buf, transports[i].name) == 0)
	  return(&transports[i]);
    }
  return (NIL);
}

static ilu_Transport _ilu_TransportCreate(ilu_TransportClass class,
					  ilu_integer fd, ilu_refany data)
{
  ilu_Transport new = (ilu_Transport)
		      ilu_malloc(sizeof(struct _ilu_Transport_s));
  static double to1 =  0.7;
  static double toN =  5.0;
  static double tto = 20.0;
  new->tr_class = class;
  new->tr_fd = fd;
  new->tr_to1 = ilu_FineTime_FromDouble(to1);
  new->tr_toN = ilu_FineTime_FromDouble(toN);
  new->tr_tto = ilu_FineTime_FromDouble(tto);
  new->tr_packet = NIL;
  new->tr_data = data;
  return (new);
}

static void _ilu_TransportFree (ilu_Transport t)
{
  ilu_free(t);
}

/*L1_sup < trmu*/
ilu_TransportClass _ilu_GetTransportClassByName (ilu_string name)
{
  struct transports_s *p;

  if ((p = FindTransport(name)) == NIL)
    return (NIL);
  else
    return ((*(p->instantiator))());
}

/*Main Invariant holds*/
ilu_Transport _ilu_GetTransportFromClass(ilu_TransportClass tc,
					 ilu_string pinfo)
{
  ilu_Transport t;
  t = _ilu_TransportCreate (tc, 0, NIL);
  t->tr_data = transport_interpret_info(t, pinfo);
  if (transport_connect(t))
      return (t);
  else
    {
      _ilu_TransportFree(t);
      return (NIL);
    }
}

ILU_DEF_ERR(WriteError, "(transport) error on write") {
  /* nothing to do */
}

ILU_QUADEF(WriteError);

ILU_DEF_ERR(InvalidBufferSize, "Asked for invalid number of bytes from buffer") {
  /* nothing to do */
}

ILU_QUADEF(InvalidBufferSize);
