/*
Copyright (c) 1991, 1992, 1993 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: udp.c,v 1.27 1994/12/06 23:52:36 spreitze Exp $ */
/* Last edited by Mike Spreitzer December 6, 1994 3:51 pm PST */

#ifdef _IS_BSD	/* we believe FIONREAD to be a BSD-ism.  Is this true? */
#define BSD_COMP
#ifndef FIONREAD
#include <sys/ioctl.h>	/* for FIONREAD */
#endif
#endif

#include "iluntrnl.h"

#include "transprt.h"
#include "mooring.h"

#include "socket.h"

#include "oscalls.h"

#define max_udp_recv 8200

/*L1_sup < trmu; L2, Main unconstrained*/
ilu_TransportClass _ilu_udp_TransportClass(void);

typedef struct udpparms {
  /*L1, L2, Main unconstrained*/
  
  ilu_string hostname;
  struct sockaddr_in addr;	/* For a Transport, the other side;
				   For a Mooring, this side */
  ilu_boolean incoming;		/* for a Transport: into server,
				   or going out from client? */
  /*L1 >= {connection's iomu}*/

  ilu_bytes ibuf;		/* for incoming; the packet */
  ilu_integer len;		/* for incoming; the pkt len */

  ilu_bytes obuf;		/* for outgoing; the output buffer */
  ilu_cardinal size;		/* buffer size */
  ilu_bytes next;		/* next pointer */
  ilu_cardinal left;		/* n bytes left in buffer */

} *UDPParms;
/* What goes in the data field of a UDP ilu_Transport or ilu_Mooring.
   The FD (socket) of an outgoing UDP Transport is `bound' to its
   destination; the FD of an incoming UDP Mooring or Transport is
   not bound. */

/*L1, L2, Main unconstrained*/

#define UDP_HOSTNAME(a) (((UDPParms)(a))->hostname)
#define UDP_PORT(a) (((UDPParms)(a))->addr.sin_port)

#define MAXDUMP		10000

/* for output buffers */
#define INITBUFSIZE	8192
#define INCRBUFSIZE	256

/**********************************************************************
***********************************************************************
***********************************************************************
***** First, the methods for the UDP Transport ************************
***********************************************************************
***********************************************************************
**********************************************************************/

/*L2, Main unconstrained*/

/*L1 unconstrained*/

#ifdef _IS_POSIX
#include <sys/types.h>
#include <fcntl.h>
static int set_non_blocking (int fd)
{
  int flags;

  if ((flags = fcntl(fd, F_GETFL, 0)) < 0 OR
      fcntl(fd, F_SETFL, flags | O_NONBLOCK) < 0)
    return (errno);
  else
    return 0;
}
#elif _IS_BSD
static int set_non_blocking (int fd)
{
  int one = 1;

  return (ioctl(fd, FIONBIO, (int) &one));
}
#else	/* not POSIX or BSD */
"Don't know how to do non-blocking I/O for anything but POSIX and BSD just now"
#endif

static ilu_private InterpretInfo (ilu_string info)
{
  char hostname[1000];
  ilu_cardinal port;
  struct hostent *hp;

  if ((sscanf (info, "udp_%[^_]_%lu", hostname, &port)) == 2)
    {
      UDPParms new = (UDPParms) ilu_malloc(sizeof(struct udpparms));
      memset((ilu_string ) &new->addr, 0, sizeof(new->addr));
      new->addr.sin_family = AF_INET;
      new->addr.sin_port = htons((ilu_shortcardinal) port);

      new->addr.sin_addr.s_addr = inet_addr(hostname);
      if (   new->addr.sin_addr.s_addr == -1
          || new->addr.sin_addr.s_addr == 0)
	if ((hp = gethostbyname(hostname)) != NIL)
	  memcpy ((ilu_string) &new->addr.sin_addr, hp->h_addr, hp->h_length);

      if (   new->addr.sin_addr.s_addr == -1
          || new->addr.sin_addr.s_addr == 0)
	{
	  DEBUG((CONNECTION_DEBUG | UDP_DEBUG),
		(stderr, "udp InterpretInfo:  Invalid host name (%s).\n",
		 hostname));
	  ilu_free(new);
	  return (NIL);
	}

      new->hostname = _ilu_Strdup(hostname);
      new->incoming = FALSE;
      new->ibuf = NIL;
      new->len = 0;

      new->obuf = NIL;
      return ((ilu_private) new);
    }
  else
    return (NIL);
}

static void DumpPacket (ilu_byte *packet, ilu_shortcardinal length)
{
  ilu_cardinal dumplength;
  int n;
  ilu_byte c;
  register ilu_integer i, j;

  if (length > MAXDUMP)
    {
      fprintf (stderr, "Request to dump packet of %u bytes.  Only %u bytes being dumped.\n",
	       length, MAXDUMP);
      dumplength = MAXDUMP;
    }
  else
    dumplength = length;
  if (packet == NIL)
    {
      fprintf (stderr, "Attempt to dump NIL packet.\n");
      return;
    }
  fprintf (stderr, "DumpPacket of packet %p, length %u bytes, dumping %lu bytes:\n",
	   packet, length, dumplength);
  for (i = 0;  i < dumplength;  i += 16)
    {
      fprintf (stderr, "%6lu:  ", i);
      for (j = 0;  j < 16 AND (i + j) < dumplength;  j += 1)
	fprintf (stderr, "%02x%s ", packet[i + j],
		 ((j % 4) == 3) ? " " : "");
      n = (((16-j)/4)*(13)+((16-j)%4)*3)+1; /* padding before ascii */
      fprintf (stderr, "%*.*s", n, n, "");
      for (j = 0;  j < 16 AND (i + j) < dumplength;  j += 1)
	{
	  c = packet[i + j];
	  fprintf (stderr, "%c", ((c >= ' ') && (c <= '~')) ? (char) c
							    : '.');
	}
      fprintf (stderr, "\n");
    }
}

#ifdef _IS_BSD	/* we believe FIONREAD to be a BSD-ism.  Is this true? */

/*L2 >= {conn's iomu}*/
/*L1, Main unconstrained*/
static ilu_integer NDataPresent (ilu_Transport t)
{
  UDPParms parms = (UDPParms) transport_data(t);
  long count;
  int status;
  if (parms->incoming)
    return 1;
  status = ioctl (t->tr_fd, FIONREAD, (int) &count);
  /* we cast the third arg to ioctl 'cause it's declared as int. */
  if (status==0 && count==0) {
      int errnum = 0;
      SOCKET_SIZE_TYPE errsize = sizeof(errnum);
      status = getsockopt(t->tr_fd, SOL_SOCKET, SO_ERROR,
			     (char *) &errnum, &errsize);
      if (status < 0 || errnum != 0)
           return -1;
      else return 0;
    }
  return ( (status != 0) ? -1 : (count > 0) ? 1 : 0 );
}

#else
"Don't know how to do FIONREAD on anything but BSD"
#endif

/*Main Invariant holds*/
/*L2 >= {conn's iomu}*/

static ilu_boolean WaitForInput (ilu_Transport t, ilu_boolean *sure,
				 ilu_FineTime *limit)
{
  UDPParms parms = (UDPParms) transport_data(t);
  if (parms->incoming) {
      *sure = TRUE;
      return (parms->ibuf != NIL);
    }
  else {
      _ilu_WaitForInputOnFD(t->tr_fd, sure, limit);
      return (TRUE);
    }
}

/*Main Invariant holds*/
/*L2 >= {conn's iomu}*/

static          ilu_bytes
_udp_AcquireBuffer(ilu_Transport self, ilu_cardinal size,
		   ILU_ERRS((InvalidBufferSize)) * err)
{
  UDPParms parms = (UDPParms) self->tr_data;
  
  DEBUG(UDP_DEBUG, (stderr, "_udp_AcquireBuffer:  Acquiring %lu bytes in the UDP buffer (%p, %lu, %lu left)\n",
		    size, parms->obuf, parms->size, parms->left));

  *err = ILU_NO_ERR;
  if (parms->obuf == NIL)
    {
      parms->size = MAX(size, INITBUFSIZE);
      parms->obuf = (ilu_bytes) ilu_malloc(parms->size);
      if (parms->obuf == NIL) {
	ILU_BIND_ERR(InvalidBufferSize, err, ev) {
	  ev->maxsize = 1;
	  ev->minsize = 0;
	  ev->requested_size = parms->size;
	}
	ILU_END_BIND_ERR;
	return NIL;
      }
      parms->size = INITBUFSIZE;
      parms->left = parms->size;
      parms->next = parms->obuf;
      return (_udp_AcquireBuffer(self, size, err));
    }
  else if (size > parms->left)
    {
      DEBUG(UDP_DEBUG,
	    (stderr,
       "_udp_AcquireBuffer:  Adding %lu bytes to current buffer\n",
	     MAX(size, INCRBUFSIZE)));
      parms->size += MAX(size, INCRBUFSIZE);
      parms->obuf = (ilu_bytes) realloc(parms->obuf, parms->size);
      if (parms->obuf == NIL)
	return ILU_ERR_CONS3(InvalidBufferSize, err,
			     maxsize, 1,
			     minsize, 0,
			     requested_size, parms->size, NIL);
      parms->next = parms->obuf + (parms->size - MAX(size, INCRBUFSIZE))
	                        - parms->left;
      parms->left = parms->left + MAX(size, INCRBUFSIZE);
      return (_udp_AcquireBuffer(self, size, err));
    }
  else	/* have space */
    {
      ilu_bytes ptr = parms->next;
      parms->next += size;
      parms->left -= size;
      return (ptr);
    }
}

static
ILU_ERRS((InvalidBufferSize)) _udp_WriteBytes(ilu_Transport self,
					      ilu_bytes b,
					      ilu_cardinal len)
{
  ilu_Error       err = ILU_NO_ERR;
  ilu_bytes       ptr;

  if ((ptr = _udp_AcquireBuffer(self, len, &err)) != NIL) {
    memcpy((void *) ptr, (void *) b, len);
    return (NIL);
  } else
    return err;
}

static 
ILU_ERRS((BadDataStructure, WriteError))
_udp_FlushOutput(ilu_Transport conn)
{
  ilu_Error status = NIL;
  ilu_integer sent;
  UDPParms parms = (UDPParms) conn->tr_data;
  ilu_bytes buffer;
  ilu_cardinal count;

  if (conn == NIL) {
    BIND_BadDataStructure(&status, conn, "ilu_Transport (udp)",
			  "not supposed to be NIL");
    return status;
  }

  parms = (UDPParms) conn->tr_data;

  if (parms->obuf == NIL)
    return (NIL);

  buffer = parms->obuf;
  count = parms->size - parms->left;
  parms->obuf = NIL;
  conn->tr_packet = buffer;
  conn->tr_size = count;

  DEBUG((CONNECTION_DEBUG | UDP_DEBUG),
	(stderr,
	 "_udp_FlushOutput:  writing %lu bytes from %p to fd %d.\n",
	 count, buffer, conn->tr_fd));

  if ((_ilu_DebugLevel & PACKET_DEBUG) != 0)
      DumpPacket(buffer, count);

  if (parms->incoming)
       sent = sendto(conn->tr_fd, (char *) buffer, count, 0,
		     (struct sockaddr *) &parms->addr,
		     sizeof(parms->addr));
       /* Even though the SunOS man page says sendto can be used on
          connected sockets, in fact that raises errno 56 */
  else sent = send(conn->tr_fd, (char *) buffer, count, 0);
  _ilu_Assert(sent <= (ilu_integer) count,
              "udp SendMessage: excess bytes sent!");
  if (sent < count) {
    int             the_errno = errno;
    ILU_BIND_ERR(WriteError, &status, ev) {
      ev->transport = conn;
      ev->return_status = sent;
      ev->buffer = buffer;
      ev->size = count;
      ev->ep_errno = the_errno;
      ev->errstr = strerror(the_errno);

      fprintf(stderr, "%s %d of %lu bytes signalled error \"%s\".\n",
	    "_udp_FlushOutput:  output to fd", conn->tr_fd, count,
	      (sent < 0) ? strerror(the_errno)
	      : ((sent == 0) ? "Connection lost"
		 : "not enough bytes written"));
    } ILU_END_BIND_ERR;
  }
  return (status);
}

/*L1, L2, Main unconstrained*/

/*Main Invariant holds*/
static ilu_boolean Connect (ilu_Transport self)
{
  UDPParms parms = (UDPParms) transport_data(self);
  int fd;
  ilu_boolean status = TRUE;

  _ilu_AutoSetDebugLevel();

  if (self == NIL)
    return (FALSE);

  DEBUG((CONNECTION_DEBUG | UDP_DEBUG),
	(stderr, "udp.c:Connect:  connecting to host %s, port %d...\n",
	 UDP_HOSTNAME(transport_data(self)),
	 UDP_PORT(transport_data(self))));

  if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    {
      DEBUG((CONNECTION_DEBUG | UDP_DEBUG),
	    (stderr, "udp.c:Connect:  socket call failed:  %s.\n",
	     strerror(errno)));
      status = FALSE;
    }
  else
    {
#ifdef SO_REUSEADDR
      (void) setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
			   (ilu_string ) NIL, 0);
#endif /* SO_REUSEADDR */

#ifdef SO_USELOOPBACK
      (void) setsockopt(fd, SOL_SOCKET, SO_USELOOPBACK,
			   (ilu_string ) NIL, 0);
#endif /* SO_USELOOPBACK */
    

      if (set_non_blocking(fd) != 0)
      /* we cast the third arg to ioctl 'cause it's declared as int. */
        {
        DEBUG((CONNECTION_DEBUG | UDP_DEBUG),
		(stderr, "%s (descriptor %d, host %s) non-blocking.\n",
		 "udp.c:Connect:  Failed to set socket",
		 fd, parms->hostname));;
	  status = FALSE;
	}
      else
	{
	  if (connect(fd, (struct sockaddr *) &parms->addr,
	                 sizeof(parms->addr)) == 0)
	    {
	      DEBUG((CONNECTION_DEBUG | UDP_DEBUG),
		    (stderr, "udp.c:Connect:  connected to %s:%u\n",
		     inet_ntoa(parms->addr.sin_addr),
		     ntohs(parms->addr.sin_port)));
	      self->tr_fd = fd;
	      status = TRUE;
	    }
	  else if (errno == EINPROGRESS) {
	    struct sockaddr peername;
	    SOCKET_SIZE_TYPE pnlen = sizeof(peername);
	    ilu_boolean sure;
	    _ilu_WaitForOutputOnFD(fd, &sure, NIL);
	    if (getpeername(fd, &peername, &pnlen) == 0) {
	      DEBUG((CONNECTION_DEBUG | UDP_DEBUG),
		    (stderr,
		     "udp.c:Connect:  eventually connected to %s[%u]\n",
		     inet_ntoa(parms->addr.sin_addr),
		     ntohs(parms->addr.sin_port)));
	      self->tr_fd = fd;
	      status = TRUE;
	      }
	    else {
	      DEBUG((CONNECTION_DEBUG | UDP_DEBUG),
		    (stderr, "%s %s:%u failed (%s).\n",
		     "udp.c:Connect: connect to",
		     inet_ntoa(parms->addr.sin_addr),
		     ntohs(parms->addr.sin_port),
		     "no meaningful errno available" ));
	      close(fd);
	      status = FALSE;
	      }
	    }
	  else
	    {
	      DEBUG((CONNECTION_DEBUG | UDP_DEBUG),
		    (stderr, "%s %s[%u] failed, error %d (%s)\n",
		     "udp.c:Connect: connect to",
		     inet_ntoa(parms->addr.sin_addr),
		     ntohs(parms->addr.sin_port), errno,
		     strerror(errno)));
	      close(fd);
	      status = FALSE;
	    }
	}
    }
  return (status);
}

/*L2 >= {conn's iomu}*/

static void Close (ilu_Transport self)
{
  UDPParms parms = (UDPParms) transport_data(self);
  if (self == NIL)
    return;
  if (parms->incoming && (parms->ibuf != NIL))
    ilu_free(parms->ibuf);
  if (!parms->incoming && (parms->obuf != NIL))
    ilu_free(parms->obuf);
  if (!parms->incoming)
      close (self->tr_fd);
  return;
}

/*Main Invariant holds*/
static ilu_boolean ReadMessage (ilu_Transport self,
				ilu_bytes *buffer,
				ilu_cardinal *size)
{
  UDPParms parms = (UDPParms) transport_data(self);
  ilu_bytes packet;
  int len;

  if (self == NIL)
    return(FALSE);

  if (! parms->incoming) {
      packet = (unsigned char *) ilu_malloc(max_udp_recv);
      _ilu_Assert(packet != NIL, "udp.c:ReadMessage: ilu_malloc failed");
      len = recv(self->tr_fd, (char *) packet, max_udp_recv, 0);
      if (len == 0) {
          DEBUG(UDP_DEBUG, (stderr, "udp.c:ReadMessage: got length 0"));
          ilu_free(packet);
          return FALSE;
        }
      if (len < 0) {
          DEBUG(UDP_DEBUG,
                (stderr,
                 "udp.c:ReadMessage: recv(%s:%d) failed, ernno=%d(%s).\n",
                 UDP_HOSTNAME(transport_data(self)),
                 UDP_PORT(transport_data(self)),
                 errno, strerror(errno) ));
          ilu_free(packet);
          return FALSE;
        }

      *buffer = packet;
      *size = len;
    }
  else {
      _ilu_Assert(parms->ibuf != NIL,
                  "udp ReadMessage: incoming buf is NIL");
      *buffer = parms->ibuf;
      *size = parms->len;
      parms->ibuf = NIL;
      parms->len = 0;
    }
  if (_ilu_DebugLevel & PACKET_DEBUG != 0)
    DumpPacket(*buffer, *size);

  return (TRUE);
}

/**********************************************************************
***********************************************************************
***********************************************************************
**** Now the methods for the UDP Mooring ******************************
***********************************************************************
***********************************************************************
**********************************************************************/

/*L1, L2, Main unconstrained*/

static double _udp_Timeout_to1 = 0.7;
static double _udp_Timeout_toN = 5.0;
static double _udp_Timeout_tto = 20.0;

void _ilu_udp_SetTimeouts (double to1, double toN, double tto)
{
  _udp_Timeout_to1 = to1;
  _udp_Timeout_toN = toN;
  _udp_Timeout_tto = tto;
}

static ilu_string FormHandle (ilu_refany parameterblock)
{
  UDPParms parms = (UDPParms) parameterblock;
  char buf[11];
  if (parms == NIL)
    return (NIL);
  sprintf (buf, "%u", UDP_PORT(parms));
  return (_ilu_Strcat5("udp", "_", UDP_HOSTNAME(parms), "_", buf));
}

/*L1_sup < trmu*/
static ilu_Transport AcceptClient (ilu_Mooring self)
{
  int buflen = max_udp_recv;
  struct sockaddr_in from;
  SOCKET_SIZE_TYPE fromlen = sizeof(from);
  UDPParms mparms = self->mo_data;
  UDPParms cparms = (UDPParms) ilu_malloc(sizeof(struct udpparms));
  ilu_Transport new = NIL;

  _ilu_AutoSetDebugLevel();

  cparms->ibuf = (unsigned char *) ilu_malloc(buflen);
  _ilu_Assert(cparms->ibuf != NIL, "udp_AcceptClient: ilu_malloc failed");
  
  cparms->len = recvfrom(self->mo_fd, (char *) cparms->ibuf, buflen, 0,
                         (struct sockaddr *) &from, &fromlen);
  
  if (cparms->len < 0) {
      DEBUG((CONNECTION_DEBUG | UDP_DEBUG),
            (stderr,
             "_udp_AcceptClient: recvfrom port %u failed, errno=%d.\n",
             ntohs(mparms->addr.sin_port), errno));
      ilu_free(cparms->ibuf);
      ilu_free(cparms);
      return (NIL);
    }

  cparms->hostname = _ilu_Strdup(inet_ntoa(from.sin_addr));
  cparms->addr = from;
  cparms->incoming = TRUE;
  new = (ilu_Transport) ilu_malloc (sizeof(struct _ilu_Transport_s));
  new->tr_class = _ilu_udp_TransportClass();
  new->tr_fd = self->mo_fd;
  new->tr_to1 = ilu_FineTime_FromDouble(_udp_Timeout_to1);
  new->tr_toN = ilu_FineTime_FromDouble(_udp_Timeout_toN);
  new->tr_tto = ilu_FineTime_FromDouble(_udp_Timeout_tto);
  new->tr_data = cparms;

  DEBUG((CONNECTION_DEBUG | UDP_DEBUG),
	(stderr,
	 "udp AcceptClient: request of len %lu from %s:%u on port %u.\n",
	 cparms->len, cparms->hostname, ntohs(cparms->addr.sin_port),
	 ntohs(mparms->addr.sin_port)));

  return (new);
}

/*L1_sup < trmu*/
static ilu_Mooring CreateMooring (ilu_private mooringInfo)
{
  UDPParms parms = (UDPParms) mooringInfo;
  struct sockaddr_in sin = ((UDPParms) mooringInfo)->addr;
  int skt;
  SOCKET_SIZE_TYPE namelen;
  ilu_Mooring self;

  _ilu_AutoSetDebugLevel();

  if ((skt = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    DEBUG((EXPORT_DEBUG | UDP_DEBUG),
	  (stderr, "_udp_CreateMooring: socket failed:  %s.\n",
	   strerror(errno)));
    return (NIL);
  }
    
  sin.sin_addr.s_addr = INADDR_ANY;	/* me, me! */

  if (bind(skt, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
    DEBUG((EXPORT_DEBUG | UDP_DEBUG),
	  (stderr, "_udp_CreateMooring: bind to port %u failed:  %s.\n",
	   ntohs(parms->addr.sin_port), strerror(errno)));
    close (skt);
    return (NIL);
  }

  if (sin.sin_port == 0) {
    namelen = sizeof(sin);
    if (getsockname(skt, (struct sockaddr *) &sin, &namelen) < 0)
      {
	DEBUG((EXPORT_DEBUG | UDP_DEBUG),
	      (stderr, "_udp_CreateMooring: getsockname failed:  %s.\n",
	       strerror(errno)));
	close (skt);
	return (NIL);
      }
    parms->addr = sin;
  }
  
  DEBUG((EXPORT_DEBUG | UDP_DEBUG),
        (stderr, "udp CreateMooring: new mooring on FD %d, port %u.\n",
         skt, ntohs(sin.sin_port)));

  self = (ilu_Mooring) ilu_malloc (sizeof(struct _ilu_Mooring_s));

  self->mo_fd = skt;
  self->mo_accept_connection = AcceptClient;
  if (UDP_HOSTNAME(parms) != NIL)
    ilu_free(UDP_HOSTNAME(parms));
  _ilu_AcquireMutex(ilu_trmu);
  UDP_HOSTNAME(parms) = _ilu_Strdup(_ilu_Hostname());
  _ilu_ReleaseMutex(ilu_trmu);
  self->mo_transportClass = _ilu_udp_TransportClass();
  self->mo_data = (ilu_private) parms;

  return (self);
}

static void CloseMooring(ilu_Mooring m)
{
  int res;
  res = close(m->mo_fd);
  ASSERT(res==0, buf,
	 (buf, "udp_CloseMooring: res=%d, errno=%s", res,
	  strerror(errno) ));
}

/*L1_sup < trmu*/
ilu_TransportClass _ilu_udp_TransportClass (void)
{
  static ilu_TransportClass m = NIL;
  _ilu_AcquireMutex(ilu_trmu);
  if (m == NIL)
    {
      m = (ilu_TransportClass)
	  ilu_malloc(sizeof(struct _ilu_TransportClass_s));
      m->tc_pFD		= 1;
      m->tc_cFD		= 1;	/* Ack!  Half untrue! */
      m->tc_timesout	= TRUE;
      m->tc_interpret_info	= InterpretInfo;
      m->tc_form_info	= FormHandle;
      m->tc_connect	= Connect;
      m->tc_wait_for_input	= WaitForInput;
      m->tc_create_mooring	= CreateMooring;
      m->tc_close_mooring	= CloseMooring;
      m->tc_close		= Close;
      m->tc_n_data_present	= NDataPresent;
      m->tc_write_bytes		= _udp_WriteBytes;
      m->tc_get_output_buffer	= _udp_AcquireBuffer;
      m->tc_read_message	= ReadMessage;
      m->tc_flush_output	= _udp_FlushOutput;
    }
  _ilu_ReleaseMutex(ilu_trmu);
  return (m);
}

