/*
Copyright (c) 1991, 1992, 1993 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

$Id: ilu-franz-skin.c,v 1.8 1995/02/21 01:16:07 janssen Exp $
*/

#define _BSD_SOURCE		/* to allow SGI to process select() */

#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>

#include <iluxport.h>

/* set up technology for blocking on read and write appropriately */

static int (*read_fn) (int fd, unsigned long msec) = NULL;
static int (*write_fn) (int fd, unsigned long msec) = NULL;

#ifdef ALLEGRO_4_2
#define FixnumToInt(v)	((v)>>2)
#else
#define FixnumToInt(v)	((v)>>3)
#endif /* ALLEGRO_4_2 */

#define MILLISECONDS(x)	(((x)==NULL)?0:((x)->ft_s * 1000 + (((x)->ft_t * 1000) / ilu_FineTimeRate)))

int ilufranz_OutputPossibleP (int fd)
{
  static fd_set writefds;
  int width, stat;
  static struct timeval to = { 0, 0 };

  width = fd + 1;
  FD_ZERO(&writefds);
  FD_SET(fd, &writefds);
  stat = select (width, NULL, &writefds, NULL, &to);
  if (stat > 0 && FD_ISSET(fd, &writefds))
    return 1;
  else
    return 0;
}

static void internal_read_wait (int fd, ilu_boolean *sure, ilu_FineTime *limit)
{
  if (read_fn != NULL)
    *sure = FixnumToInt((*read_fn)(fd, (limit == NULL) ? 0 : MILLISECONDS(limit)));
  else
    *sure = 0;
}

static void internal_write_wait (int fd, ilu_boolean *sure, ilu_FineTime *limit)
{
  if (ilufranz_OutputPossibleP(fd))
    {
      *sure = 1;
    }
  else
    {
      if (write_fn != NULL)
	*sure = FixnumToInt((*write_fn)(fd, (limit == NULL) ? 0 : MILLISECONDS(limit)));
      else
	*sure = 0;
    }
}

void ilufranz_SetWaitTech (int (*read_wait) (int fd, unsigned long timetowait),
			   int (*write_wait) (int fd, unsigned long timetowait))
{
  static ilu_WaitTech wait_tech;

  read_fn = read_wait;
  write_fn = write_wait;
  wait_tech.wt_read_wait = internal_read_wait;
  wait_tech.wt_write_wait = internal_write_wait;
  ilu_SetWaitTech (&wait_tech);
}

/* these functions are used for locking and condition variables */

typedef long (*franzCreateFn) (ilu_string, ilu_string);

static franzCreateFn internal_mcreate;

static ilu_private franz_mcreate (ilu_string d1, ilu_string d2)
{
  long retval;
  retval = (*internal_mcreate) (d1, d2);
  return ((ilu_private) FixnumToInt(retval));
}

static franzCreateFn internal_ccreate;

static ilu_private franz_ccreate (ilu_string d1, ilu_string d2)
{
  long retval;
  retval = (*internal_ccreate) (d1, d2);
  return ((ilu_private) FixnumToInt(retval));
}

void ilufranz_SetLockTech (long (*mcreate)(ilu_string d1, ilu_string d2),
			   void (*macquire) (ilu_private m),
			   void (*mhold) (ilu_private m),
			   void (*mrelease) (ilu_private m),
			   void (*mdestroy) (ilu_private m),
			   long (*ccreate)(ilu_string d1, ilu_string d2),
			   void (*cnotify) (ilu_private c),
			   void (*cdestroy) (ilu_private c),
			   void (*cwait) (ilu_private c, ilu_private m))
{
  static ilu_LockTech franzLockTech;

  internal_mcreate = mcreate;
  internal_ccreate = ccreate;

  franzLockTech.lt_mcreate = franz_mcreate;
  franzLockTech.lt_acquire = macquire;
  franzLockTech.lt_hold = mhold;
  franzLockTech.lt_release = mrelease;
  franzLockTech.lt_mdestroy = mdestroy;
  franzLockTech.lt_ccreate = franz_ccreate;
  franzLockTech.lt_notify = cnotify;
  franzLockTech.lt_cdestroy = cdestroy;
  franzLockTech.lt_wait = cwait;

  ilu_SetLockTech (&franzLockTech);
}

void ilufranz_CallAlarmClosure (void (*proc) (ilu_private rock), ilu_private rock)
{
  (*proc)(rock);
}

static int (*internal_create_alarm)();

static ilu_refany franz_create_alarm ()
{
  return ((ilu_refany) (FixnumToInt((*internal_create_alarm)())));
}

static void (*internal_set_alarm) (unsigned long alarm,
				   long time_sec, unsigned long time_msec,
				   unsigned long p1, unsigned long p2);

static void franz_set_alarm (ilu_refany alarm, ilu_FineTime t,
			     void (*proc)(ilu_private rock), ilu_private rock)
{
/*
  printf ("Setting alarm %u to %u.%u (%24.24s)\n", (unsigned long) alarm, t.ft_s, t.ft_t,
	  ctime(&t.ft_s));
*/
  (*internal_set_alarm) ((unsigned long) alarm,
			 t.ft_s, (t.ft_t * 1000) / ilu_FineTimeRate,
			 (unsigned long) proc, (unsigned long) proc);
}

static ilu_boolean EmptyFn (int fd)
{
  /* empty function used as dummy for ilu_Unregister{Input,Output}Handler (which must be present) */
  return (ilu_FALSE);
}

void ilufranz_SetAlarms (int (*create_alarm) (),
			 void (*set_alarm) (unsigned long alarm, long time_sec, unsigned long time_msec,
					    unsigned long p1, unsigned long p2),
			 void (*unset_alarm) (ilu_refany alarm))
{
  static ilu_MainLoop ml;

  internal_create_alarm = create_alarm;
  internal_set_alarm = set_alarm;

  ml.ml_run = NULL;
  ml.ml_exit = NULL;
  ml.ml_register_input = NULL;
  ml.ml_unregister_input = EmptyFn;
  ml.ml_register_output = NULL;
  ml.ml_unregister_output = EmptyFn;

  ml.ml_create_alarm = franz_create_alarm;
  ml.ml_set_alarm = franz_set_alarm;
  ml.ml_unset_alarm = unset_alarm;

  ilu_SetMainLoop (&ml);
}

static unsigned long (*internal_interest_hook) (unsigned long index, unsigned long interest) = 0;

/*Inside(obj's server, obj's type)*/
static void kernel_interest_hook (ilu_Object kobj, int vi)
{
  unsigned long index = (unsigned long) ilu_GetLanguageSpecificObject (kobj);
  unsigned long newindex;
  ilu_Class cl = ilu_ClassOfObject(kobj);
  
  /* printf("Interest:  %s %4u %s\n", (vi == 0) ? "NO  " : "YES", index, ilu_SBHOfObject(kobj)); */

  if (index != 0 &&
      internal_interest_hook != 0 &&
      ((!ilu_TrueInstanceP(kobj)) || (cl->cl_collectible)) &&
      (((vi == 0) && ((index & 0x1) != 0)) ||
       ((vi != 0) && ((index & 0x1) == 0))))
    {
      newindex = FixnumToInt((*internal_interest_hook) (index, vi));
      if (newindex != index)
	{
/*	  printf("           LSPO changed to %u\n", newindex); */
	  ilu_RegisterLanguageSpecificObject (kobj, (void *) newindex);
	}
    }
}

void ilufranz_SetInterestHook (unsigned long (*hook) (unsigned long, unsigned long))
{
  internal_interest_hook = hook;
  ilu_SetNoter (&kernel_interest_hook);
}
