(* IluSimpleBind.m3 *)
(* Copyright (c) 1994 Xerox Corporation.  All Rights Reserved.

   Unlimited use, reproduction, and distribution of this software is
   permitted.  Any copy of this software must include both the above
   copyright notice of Xerox Corporation and this paragraph.  Any
   distribution of this software must comply with all applicable United
   States export control laws.  This software is made available AS IS,
   and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
   INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
   NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY LIABILITY
   FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS EXPRESSLY
   DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING NEGLIGENCE)
   OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGES. *)
(* $Id: IluSimpleBind.m3,v 1.1 1994/10/12 20:03:00 spreitze Exp spreitze
   $ *)
(* Edited by Mike Spreitzer at October 17, 1994 11:34 am PDT *)

MODULE IluSimpleBind;
FROM IluBasics IMPORT Failed, Failure;
IMPORT Ilu, IluF, IluKernel, IluRuntime, M3asC;

<*PRAGMA lL, Ll, Main*>

EXCEPTION CantHappen;

<*FATAL CantHappen*>

REVEAL Cookie = BRANDED REF RECORD proof: IluKernel.C_String END;

<*Main invariant holds*>

PROCEDURE Publish (obj: Ilu.Object): Cookie RAISES {Failed} =
  VAR ko: IluKernel.Object;
  VAR ans: Cookie;
  BEGIN
    IF obj = NIL THEN RETURN NIL END (*if*);
    ko := IluF.KOfO(obj);
    ans := NEW(Cookie, proof := IluKernel.ilu_PublishObject(ko));
    IF ans.proof = NIL THEN
      RAISE Failed(NEW(Failure, info := "ilu_PublishObject failed"))
    ELSE
      RETURN ans
    END (*if*);
  END Publish;

PROCEDURE Withdraw (obj: Ilu.Object; c: Cookie) RAISES {Failed} =
  VAR ko: IluKernel.Object;
  VAR ans: IluKernel.Bool;
  BEGIN
    IF obj = NIL THEN
      RAISE
        Failed(
          NEW(
            Failure,
            info := "You couldn't possibly Withdraw the NIL object"))
    END (*if*);
    IF c = NIL THEN
      RAISE
        Failed(
          NEW(
            Failure, info := "NIL is not a valid IluSimpleBind.Cookie"))
    END (*if*);
    ko := IluF.KOfO(obj);
    ans := IluKernel.ilu_WithdrawObject(ko, c.proof);
    IF ans = 0 THEN
      RAISE Failed(NEW(Failure, info := "ilu_Withdraw failed"))
    END (*if*);
  END Withdraw;

PROCEDURE Lookup (iid: Ilu.InstanceId; ot: Ilu.ObjectType): Ilu.Object
  RAISES {Failed} =
  VAR ans: Ilu.Object := NIL;
  VAR f: Failure := NIL;
  PROCEDURE WithC (ciid: IluKernel.C_String) RAISES {} =
    VAR ko := IluKernel.ilu_LookupObject(ciid, ot);
    BEGIN
      IF ko = NIL THEN
        ans := NIL
      ELSE
        TRY
          ans := IluRuntime.M3ObjectFromKernObj(ko, ot)
        EXCEPT
          Failed (x) => f := x
        END (*try-except*);
      END (*if*);
    END WithC;
  BEGIN
    IF iid = NIL THEN
      RAISE
        Failed(
          NEW(Failure, info := "NIL is not a valid Ilu.InstanceId"))
    END;
    IF ot = NIL THEN RAISE Failed(NEW(Failure, info := "NIL ot")) END;
    M3asC.TasS(iid, WithC);
    IF f # NIL THEN RAISE Failed(f) END;
    RETURN ans;
  END Lookup;

BEGIN
END IluSimpleBind.
