/*
Copyright (c) 1991, 1992, 1993 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* Last edited by Mike Spreitzer December 6, 1994 3:36 pm PST */

/*
$Id: cplusplus.c,v 1.17 1994/12/07 00:32:34 janssen Exp $
*/
#ifdef MACOS
#pragma segment islcppmain
#endif

#include <stdlib.h>
#include "cplusplus.h"

char *OutputDirectory = ".";
char *ProgramName = NULL;
char *NamesFile = NULL;
boolean GenerateImakefile = FALSE;
boolean GenerateServerSkeleton = FALSE;
boolean GenerateClasses = FALSE;

static void generate_c_boilerplate(FILE *file, Interface parse);

extern boolean ReadSynonyms (string filename);	/* from names.c */

static void GenerateHeaders (Interface parse)
{
  char filename[1000];
  FILE *file;

#ifdef MACOS
  sprintf (filename, "%s.H", cplusplus_interface_name(parse));
#else
  sprintf (filename, "%s/%s.H", OutputDirectory, cplusplus_interface_name(parse));
#endif

  if ((file = fopen (filename, "w")) == NULL)
    {
      fprintf (stderr, "Couldn't open output file %s.\n", filename);
      exit (1);
    }
  printf ("header file interface %s to %s...\n", cplusplus_interface_name(parse), filename);
  generate_c_boilerplate (file, parse);
  generate_headers (parse, file);
  fclose (file);
}

static void GenerateCode (Interface parse)
{
  char filename[1000];
  FILE *file;

#ifdef MACOS
  sprintf (filename, "%s.cp", cplusplus_interface_name(parse));
#else
  sprintf (filename, "%s/%s.cc", OutputDirectory, cplusplus_interface_name(parse));
#endif
  if ((file = fopen (filename, "w")) == NULL)
    {
      fprintf (stderr, "Couldn't open client stubs output file %s.\n", filename);
      exit (1);
    }
  printf ("code for interface %s to %s...\n", cplusplus_interface_name(parse), filename);
  generate_c_boilerplate (file, parse);
  generate_code (parse, file);
  fclose (file);

#ifdef MACOS
  sprintf (filename, "%s-server-stubs.cp", cplusplus_interface_name(parse));
#else
  sprintf (filename, "%s/%s-server-stubs.cc", OutputDirectory, cplusplus_interface_name(parse));
#endif
  if ((file = fopen (filename, "w")) == NULL)
    {
      fprintf (stderr, "Couldn't open server stubs output file %s.\n", filename);
      exit (1);
    }
  printf ("code for server stubs of interface %s to %s...\n", cplusplus_interface_name(parse), filename);
  generate_c_boilerplate (file, parse);
  generate_server_code (parse, file);
  fclose (file);
}

static void generate_imakefile (Interface parse, FILE *file)
{
  struct context_s c;

  c.interface = parse;
  c.file = file;
}

#if (defined(_IS_BSD) || defined(_IS_POSIX))
#include <sys/stat.h>
#ifdef _IS_POSIX
#include <unistd.h>
#endif
#endif

static boolean makedir (char *dirname)
{
#if (defined(_IS_BSD) || defined(_IS_POSIX))

  struct stat s;

  if (stat(dirname, &s) == 0)
    {
      if (S_ISDIR(s.st_mode))
	{

#ifdef _IS_POSIX

	  return (access (dirname, W_OK) == 0);

#else /* not POSIX */

	  return (TRUE);

#endif
	  
	}
      else
	return (FALSE);
    }
  else
    {
      char buf[1000];
      sprintf(buf, "mkdir %s", dirname);
      return((system(buf) >> 8) == 0);
    }

#else

  fprintf (stderr, "%s:  Can't test directory \"%s\".\n",
	   ProgramName, dirname);
  return (TRUE);

#endif
}

static void GenerateStubs (Interface parse)
{
  char filename[1000];
  FILE *file;

  if (! makedir (OutputDirectory))
    {
      fprintf (stderr, "Couldn't create output directory %s.\n", OutputDirectory);
      exit (1);
    }

  if (GenerateClasses)
    {
      GenerateHeaders (parse);
      GenerateCode (parse);
    }

  if (GenerateImakefile)
    {
      sprintf (filename, "%s/Imakefile", OutputDirectory);
      if ((file = fopen (filename, "w")) == NULL)
	{
	  fprintf (stderr, "Couldn't open output file %s.\n", filename);
	  exit (1);
	}
      printf ("ILU-style Imakefile to %s...\n", filename);
      generate_imakefile (parse, file);
      fclose (file);
    }
}

static void PrintUsage (void)
{
  fprintf (stderr, "Usage:  %s [-dir OUTPUTDIRECTORY] [-imakefile] [-serverskeleton] [-classes] [-renames NAMESFILE] ISLFILE [ISLFILE ...]\n", ProgramName);
}

static boolean ParseArgs (int *pac, char ***pav)
{
  int ac = *pac;
  char **av = *pav;

  ac--;
  av++;
  while (*av[0] == '-')
    {
      if (strcmp(*av, "-dir") == 0)
	OutputDirectory = *++av;
      else if (strcmp(*av, "-imakefile") == 0)
	GenerateImakefile = TRUE;
      else if (strcmp(*av, "-serverskeleton") == 0)
	GenerateServerSkeleton = TRUE;
      else if (strcmp(*av, "-renames") == 0)
	NamesFile = *++av;
      else if (strcmp(*av, "-classes") == 0)
	GenerateClasses = TRUE;
      else
	{
	  fprintf (stderr, "%s:  Invalid switch \"%s\".\n", ProgramName, *av);
	  PrintUsage();
	  return (FALSE);
	}
      ac--;
      av++;
    }

  if (!(GenerateServerSkeleton || GenerateImakefile || GenerateClasses))
    {
      GenerateClasses = TRUE;
    }

  *pac = ac;
  *pav = av;
  return (TRUE);
}

void
main (int ac, char **av, char **envp)
{
  list s;
  char **filename;
  extern Interface GetInterface(string, string);

  ProgramName = iluparser_GetProgramName(av[0]);

  if (ac < 2)
    {
      PrintUsage();
      exit(1);
    }

  if (!ParseArgs (&ac, &av))
    exit(1);

  if (NamesFile != NULL)
    {
      if (!ReadSynonyms (NamesFile))
	{
	  fprintf (stderr, "Couldn't read names file %s.\n", NamesFile);
	  exit (1);
	};
    };      

  for (filename = av;  *filename != NULL;  filename += 1)
    {
      if ((s = ParseFile (*filename)) == NULL)
	{
	  fprintf (stderr, "Couldn't find or parse %s.\n", *filename);
	  exit (1);
	}
      list_enumerate (s, (iluparser_EnumProc) GenerateStubs, NULL);
    }
  exit(0);
}

static void generate_c_boilerplate (FILE *file, Interface parse)
{
  static char *prefixesForC[2] = { "/*", " *" };

  iluparser_GenerateBoilerplate(file, parse, ProgramName, prefixesForC);
  fprintf (file, " */\n\n");
}
