/*
 ** Copyright (c) 1991-1994 Xerox Corporation.  All Rights Reserved.
 **
 ** Unlimited use, reproduction, and distribution of this software is
 ** permitted.  Any copy of this software must include both the above
 ** copyright notice of Xerox Corporation and this paragraph.  Any
 ** distribution of this software must comply with all applicable United
 ** States export control laws.  This software is made available AS IS,
 ** and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 ** INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 ** AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 ** PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 ** THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 ** CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 ** XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
/* $Id: server.c,v 1.45 1995/03/03 02:26:51 janssen Exp $ */
/* Last edited by Mike Spreitzer November 21, 1994 11:05 pm PST */

#include "cstubber.h"

static boolean  FirstClass = TRUE;
extern boolean generateAnsi;
extern cardinal MethodRecordID;

static void declareReturnValue (Type type, Context context)
{
  if (type != NULL AND
      type_basic_type(type) != void_Type)
    {
      fprintf (context->file, "  %s _retvalue;\n", c_return_type(type));
    }
}

static void listArgument (refany elt, refany rock)
{
  Argument a = (Argument) elt;
  Context context = (Context) rock;
  enum PrimitiveTypes t = type_basic_type (ur_type(a->type));

  fprintf (context->file, ", %s%s",
	   ((a->direction == In AND
	     (t == union_Type OR t == record_Type OR
	      (t == sequence_Type AND NOT TypeIsString(ur_type(a->type)) AND NOT TypeIsWString(ur_type(a->type))))) OR
	    ((a->direction == Out OR a->direction == InOut) AND
	     t != array_Type)) ? "&" : "",
	   c_argument_name(a));
}

void FreeValue (Type type, char *name, Context context, int indent)
{
  enum PrimitiveTypes t = type_ur_kind(type);
  Type ut = ur_type(type);

  if (HasFreeRoutine(ut))
    fprintf (context->file, "%*.*s%s__Free (%s%s);\n", indent, indent, "",
	     c_type_name(ut),
	     (t == array_Type OR t == optional_Type OR TypeIsString(ut) OR TypeIsWString(ut)) ? "" : "&",
	     name);
  else if (TypeIsString(ut))
    {
      fprintf (context->file, "%*.*silu_CString__Free((ilu_CString) %s);\n", indent, indent, "", name);
    }
  else if (TypeIsWString(ut))
    {
      fprintf (context->file, "%*.*silu_WString__Free((ilu_WString) %s);\n", indent, indent, "", name);
    }
  else if (t == optional_Type)
    {
      char buf[1000];
      sprintf (buf, "*%s", name);
      FreeValue (type_description(ut)->structuredDes.optional, buf, context, indent);
    }
}     

static void freeArgument (refany elt, refany rock)
{
  Argument arg = (Argument) elt;
  Context context = (Context) rock;
  FreeValue (arg->type, c_argument_name(arg), context, 2);
}     

static void freeReturnValue (Type type, Context context)
{
  FreeValue (type, "_retvalue", context, 2);
}

static void sizeArgument (refany elt, refany rock)
{
  Argument a = (Argument) elt;
  Context context = (Context) rock;
  char b[ 1000 ];
  Type ut = ur_type(a->type);
  enum PrimitiveTypes t = type_basic_type (ur_type(a->type));

  if  (a->direction == In)
    return;
  if (t == union_Type OR
      t == record_Type OR
      (t == sequence_Type AND NOT TypeIsString(ut) AND NOT TypeIsWString(ut)))
    {
      b[0] = '&';
      strcpy (b+1, c_argument_name(a));
    }
  else
    strcpy(b, c_argument_name(a));
  fprintf (context->file, "\n        +");
  SizeValue (context, ut, b);
}

static void outputArgument (refany elt, refany rock)
{
  Argument a = (Argument) elt;
  Context context = (Context) rock;
  char b[ 1000 ];
  Type ut = ur_type(a->type);
  enum PrimitiveTypes t = type_basic_type (ur_type(a->type));

  if  (a->direction == In)
    return;
  if (t == union_Type OR
      t == record_Type OR
      (t == sequence_Type AND NOT TypeIsString(ut) AND NOT TypeIsWString(ut)))
    {
      b[0] = '&';
      strcpy (b+1, c_argument_name(a));
    }
  else
    strcpy(b, c_argument_name(a));
  MarshallValue (context, ut, b, 6);
}

static void inputArgument (refany elt, refany rock)
{
  Argument arg = (Argument) elt;
  Context context = (Context) rock;
  char buf[1000];

  if (arg->direction == Out)
    return;
  sprintf (buf, "%s%s", TypeIsPointer(ur_type(arg->type)) ? "" : "&",
	   c_argument_name(arg));
  UnmarshallValue (context, ur_type(arg->type), arg->def, buf, 2, TRUE, FALSE);
}

static void declareArgument (refany elt, refany rock)
{
  Argument arg = (Argument) elt;
  Context context = (Context) rock;
  fprintf(context->file, "  %s %s;\n", c_type_name(arg->type),
  c_argument_name(arg));
}

static void serverMethods (Procedure  m, Context  context)
{
  fprintf(context->file, "    server_%s_%s,\n", c_type_name(context->class), c_simple_name(m->name));
}

static void serverProto (Procedure  m, Context  context)
{
  fprintf(context->file, "extern void server_%s_%s ();\n", c_type_name(context->class), c_simple_name(m->name));
}

static void generateCalleeStub (Procedure  m, Context  context)
{
  enum PrimitiveTypes t = type_basic_type(ur_type(m->returnType));

  if (methodInList(c_simple_name(m->name))) {
    MethodRecordID++;
    return;
  }
  addMethodToList(c_simple_name(m->name));

  fprintf(context->file, "static void _%s_%s_stub(ilu_Call _call)\n{\n",
	  c_type_name(context->class), c_simple_name(m->name));
  fprintf(context->file, "  %s _h;\n", c_type_name(context->class));
  fprintf(context->file, "  ILU_C_ENVIRONMENT _status;\n");
  list_enumerate(m->arguments, declareArgument, context);
  declareReturnValue(m->returnType, context);
  fprintf(context->file, "\n");

  fprintf(context->file, "  _h = (%s) ",
	  c_type_name(context->class));
  if (SINGLETON(context->class))
    fprintf(context->file, "_ILU_C_GetServerSingleton (_call);\n");
  else
    fprintf(context->file,
	    "_ILU_C_InputObject (_call, &_%s__ILUClassRecord, ilu_TRUE);\n",
	    c_type_name(context->class));
  list_enumerate(m->arguments, inputArgument, context);
  fprintf(context->file, "\n");
  fprintf(context->file,
	  "  if (_ILU_C_FinishParameters (_call, _h)) {\n");
  fprintf(context->file, "    _status.returnCode = ILU_NIL;\n");
  fprintf(context->file, "    _status.callerPassport = ilu_CallerPassportOfCall(_call);\n    ");
  if (t != void_Type)
    fprintf(context->file, "_retvalue = ");
  fprintf(context->file, "%s_%s (_h, &_status",
	  c_type_name(context->class), c_simple_name(m->name));
  list_enumerate(m->arguments, listArgument, context);
  fprintf(context->file, ");\n\n");
  if (m->asynch) {
    fprintf(context->file, "    /* asynchronous method -- no reply */\n");
    fprintf(context->file, "  }\n");
    fprintf(context->file, "  _ILU_C_NoReply(_call);\n");
  } else {
    char            buf[1000];
    sprintf(buf, "%s_retvalue",
	    (t == record_Type
	     || NonStringSequence(ur_type(m->returnType))
	     || t == union_Type)
	    ? "&" : "");
    fprintf(context->file, "    if (_status.returnCode == ILU_NIL) { /* check for errors */\n");
    fprintf(context->file, "      _ILU_C_BeginReply (_call, %s, (",
       (list_size(m->exceptions) > 0) ? "ilu_TRUE" : "ilu_FALSE");
    if (t == void_Type)
      fprintf(context->file, "0");
    else
      SizeValue(context, ur_type(m->returnType), buf);
    list_enumerate(m->arguments, sizeArgument, context);
    fprintf(context->file, "));\n");
    if (type_basic_type(ur_type(m->returnType)) != void_Type)
      MarshallValue(context, m->returnType, buf, 6);
    list_enumerate(m->arguments, outputArgument, context);
    fprintf(context->file, "      _ILU_C_FinishReply (_call);\n");
    fprintf(context->file, "    }\n");
    fprintf(context->file, "    else {\n");
    if (list_size(m->exceptions) > 0) {
      fprintf(context->file,
	      "      _%s_SendException (_call, &_status);\n",
	      c_interface_name(m->interface));
    } else {
      fprintf(context->file, "      fprintf(stderr, \"** Error in stub for %s_%s:  non-NULL returnCode returned, but %s doesn't signal any exceptions!\\n\");\n",
	      c_type_name(context->class),
	      c_simple_name(m->name), c_simple_name(m->name));
      fprintf(context->file, "      _ILU_C_UnexpectedErrorReply(_call);\n");
    }
    fprintf(context->file, "    }\n");
    fprintf(context->file, "  }\n");
  }
  if (t != void_Type)
    freeReturnValue(m->returnType, context);
  list_enumerate(m->arguments, freeArgument, context);
  fprintf(context->file, "}\n\n");
}

static void classTable (Type class, Context context)
{
  Class od;

  if  (class == NULL || 
       type_basic_type (class) != object_Type || 
       (od = class_object (class)) == NULL)
    return;
  if  (list_size (od->methods) > 0)
    list_enumerate (od->methods, (void (*)(refany, refany)) serverMethods, context);
}

static void generateProto (Type  class, Context  context)
{
  Class  od;

  if  (class == NULL || 
       type_basic_type (class) != object_Type || 
       (od = class_object (class)) == NULL)
    return;
  if  (list_size (od->methods) > 0)
    list_enumerate (od->methods, (void (*)(refany, refany)) serverProto, context);
}


static void generateServerMiMethodPtrs (Type class, Context context)
{
  Type ut = ur_type(class);

  if (class_object(ut)->superclasses != NULL)
    list_enumerate (class_object (ut)->superclasses, (void (*)(refany, refany)) generateServerMiMethodPtrs, context);
  if (methodInList(c_type_name(ut)))
    return;
  addMethodToList(c_type_name(ut));
  generateProto(ut, context);
  fprintf (context->file,
	   "static struct _%s__MethodBlock_s _%s_%s__TrueMethodBlock = {\n",
	   c_type_name(ut), c_type_name(context->class), c_type_name(ut));
  fprintf (context->file, "  &_%s__ILUClassRecord", c_type_name(ut));
  if (list_size(class_object(ut)->methods) > 0)
    {
      fprintf (context->file, ",\n  {\n");
      classTable(ut, context);
      fprintf (context->file, "  }\n");
    }
  fprintf (context->file, "};\n");
}

static void generateServerClsPtrs (Type t, Context c)
{
  if (class_object(t)->superclasses != NULL)
    list_enumerate (class_object (t)->superclasses, (void (*)(refany, refany)) generateServerClsPtrs, c);
  fprintf (c->file, "  (_ILU_C_MethodBlock *) &_%s_%s__TrueMethodBlock,\n",
	   c_type_name(c->class), c_type_name(ur_type(t)));
}

static void generateServerClassTable (Type t, Context c)
{
  if (type_basic_type(t) != object_Type)
    return;

  c->class = t;
  clearMethodList ();
  generateServerMiMethodPtrs (t, c);
  fprintf (c->file, "static _ILU_C_MethodBlock *_%s__TrueTypeVector[] = {\n", c_type_name(t));
  fprintf (c->file, "  (_ILU_C_MethodBlock *) &_%s_%s__TrueMethodBlock,\n",
	   c_type_name(c->class), c_type_name(t));
  if (class_object(t)->superclasses != NULL)
    list_enumerate (class_object(t)->superclasses, (void (*)(refany, refany)) generateServerClsPtrs, c);
  fprintf (c->file, "  ILU_NIL\n};\n\n");
}

static void classSetupStubs (Type  class, Context  context)
{
  Class  od;

  if  (class == NULL || 
       type_basic_type (class) != object_Type || 
       (od = class_object (class)) == NULL)
    return;
  if  (list_size (od->methods) > 0)
    list_enumerate (od->methods, (void (*)(refany, refany)) generateCalleeStub, context);
}


static void generateServerCodeForClass (Type class, Context context)
{
  if (type_basic_type(class) != object_Type)
    return;

  context->class = class;
  clearMethodList ();
  classSetupStubs (class, context);
}

static void initializeStubPointer (Procedure  method, Context  context)
{
  fprintf (context->file, "  _%s__ILUClassRecord.cl_methods[%ld].me_stubproc = _%s_%s_stub;\n", c_type_name (context->class), MethodRecordID++, c_type_name (context->class /*method->object*/), c_simple_name (method->name));
}

static void setupServerStubsInMethodTable (Type  type, Context  context)
{
  Class od;
  
  if  (type == NULL || 
       type_basic_type (type) != object_Type || 
       (od = class_object (type)) == NULL)
    return;
  fprintf (context->file, "\n");
  list_enumerate (od->methods, (void (*)(refany, refany)) initializeStubPointer, context);
}

static void setupServerStubsTable (Type  type, Context  context)
{
  MethodRecordID = 0;
  FirstClass = True;
  context->class = type;
  setupServerStubsInMethodTable (type, context);
}

static void generateHandles (Type  type, Context  context)
{
  fprintf (context->file, "ILU_C_OBJECT %s__CreateTrue  (ilu_string instance_handle, ilu_Server server, void * data)\n{\n", c_type_name(type));
  fprintf (context->file, "  return (_ILU_C_CreateTrueObject (_%s__TrueTypeVector, instance_handle, server, data));\n}\n\n", c_type_name (type));
}

static void generateServerRegistrationCode (Interface interface, Context context)
{
  char  *interface_name = (char *) c_interface_name(interface);
  
  list_enumerate (interface->classes, (void (*)(refany, refany)) generateHandles, context);
  fprintf (context->file, "void %s__InitializeServer(void)\n{\n", interface_name);
  fprintf (context->file, "  extern void _%s__GeneralInitialization(void);\n", c_interface_name(interface));
  fprintf (context->file, "  static ilu_boolean initialized = ilu_FALSE;\n  if (initialized) return;\n  initialized = ilu_TRUE;\n\n");
  fprintf (context->file, "  _%s__GeneralInitialization ();\n", interface_name);
  list_enumerate (interface->classes, (void (*)(refany, refany)) setupServerStubsTable, context);
  fprintf (context->file, "}\n\n");
}

void generateServerCode (Interface parse, FILE *file)
{
  struct context_s context;

  context.file = file;
  context.interface = parse;
  fprintf (file, "#include <stdio.h>\n\n#define _%s_TRUE_STUB_CODE_\n\n#include \"%s.h\"\n\n",
	   c_interface_name(parse), c_interface_name(parse));

  list_enumerate (parse->classes, (void (*)(refany, refany)) generateServerClassTable, &context);

  list_enumerate (parse->classes, (void (*)(refany, refany)) generateServerCodeForClass, &context);

  generateServerRegistrationCode (parse, &context);
}
