/*
Copyright (c) 1991, 1992, 1993 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

$Id: iocode.c,v 1.25 1994/11/11 01:13:42 spreitze Exp $
*/
/* Last edited by Mike Spreitzer November 10, 1994 3:36 pm PST */

#include "m3.h"

/*===================================================================
=====================================================================

              IO for non-object types

=====================================================================
=====================================================================*/

/*********************************************************************
**********************************************************************

     Output code

**********************************************************************
**********************************************************************/

static unsigned long DimCount;
static unsigned long DimStop = 0;

static void DimHeader (long int d, Context context)
{
  if (DimStop > 0 AND DimCount >= DimStop)
    return;

  fprintf (context->file, "%*.*sFOR i%u := 0 TO %u DO\n",
	   (int) DimCount * 2 + 4, (int) DimCount * 2 + 4, "", (int) DimCount, d - 1);
  DimCount += 1;
}

static void DimRef (long int d, char *buf)
{
  if (DimStop > 0 AND DimCount >= DimStop)
    return;

  sprintf (buf + strlen(buf), "[i%u]", DimCount);
  DimCount += 1;
}

static void DimFooter (long int d, Context context)
{
  DimCount -= 1;
  if (DimStop > 0 AND DimCount < DimStop)
    return;
  fprintf (context->file, "%*.*sEND (*for*);\n", (int) DimCount * 2 + 2, (int) DimCount * 2 + 2, "");
}

typedef struct {
  Context         c;
  Type            t;
  int             skipDefault;
}               UnionTemp;

static void OutputRecordArg (Argument arg, Context context)
{
  char buf[1000];

  sprintf (buf, "val.%s", m3_argument_name(arg));
  MarshallValue (arg->type, buf, context, 6);
}

static string DiscMemVal(ConstantValue v, Type u, Type disc)
{
  TypeKind pt = type_ur_kind(disc);
  static char     buf[100];
  if ((pt == shortcardinal_Type || pt == cardinal_Type
       || pt == byte_Type
       || pt == shortcharacter_Type)
      && v->type == integer_Type)
    sprintf(buf, "%u", v->val.i.value);
  else if ((pt == shortinteger_Type || pt == integer_Type)
      && v->type == integer_Type)
    sprintf(buf, "%s16_%lX", (v->val.i.sign < 0) ? "-" : "",
	    v->val.i.value);
  else if (pt == shortcharacter_Type && v->type == shortcharacter_Type)
    sprintf(buf, "\'\\%03o\'", v->val.s[0]);
  else if (pt == boolean_Type && v->type == boolean_Type)
    sprintf(buf, "%s", v->val.b ? "TRUE" : "FALSE");
  else if (pt == enumeration_Type
	   && v->type == shortcharacter_Type)
    sprintf(buf, "%s.%s", trname(disc), v->val.s);
  else {
    fprintf(stderr, "Unable to render a tag of union type %s!\n",
	    trname(u));
    sprintf(buf, "(* can't render *)");
  }
  return buf;
}

static int GuardUnionArm (Argument a, UnionTemp *s)
{
  Type            union_type = s->t;
  TypeDescription td = type_description(union_type);
  Type            disc = td->structuredDes.uniond.discriminator_type;
  if (list_size(a->values) > 0) {
    int             first = 1;
    listElement    *e;
    ConstantValue   v;
    fprintf(s->c->file, "    | ");
    for (e = a->values->head; e != NULL; e = e->next) {
      if (first)
	first = 0;
      else
	fprintf(s->c->file, ", ");
      v = (ConstantValue) e->data;
      fprintf(s->c->file, "%s", DiscMemVal(v, union_type, disc));
    }
    fprintf(s->c->file, " =>\n");
  } else if (s->skipDefault)
    return 1;
  else
    fprintf(s->c->file, "    ELSE\n");
  return 0;
}

static void OutputUnionArm (Argument a, UnionTemp *s)
{
  if (GuardUnionArm(a, s))
    return;
  fprintf(s->c->file, "      TYPECASE val OF\n");
  fprintf(s->c->file, "      | %s (x) =>\n", UnionCaseName(s->t, a));
  MarshallValue(a->type, "x.v", s->c, 8);
  fprintf(s->c->file, "      ELSE\n");
  fprintf(s->c->file, "        RAISE\n");
  fprintf(s->c->file, "          IluBasics.Failed(\n");
  fprintf(s->c->file, "            NEW(IluBasics.Failure,\n");
  fprintf(s->c->file, "                info := \"type of a %s doesn't match its tag\"))\n", M3_TYPE_NAME(s->t));
  fprintf(s->c->file, "      END (* narrow for this arm *)\n");
}

static void ComputeTotalNumberOfElements (unsigned long dim, unsigned long *total)
{
  *total = *total * dim;
}

static void FindZero (long unsigned int val, boolean *z)
{
  if (val == 0)
    *z = TRUE;
}

static void MemToCard(string mem, string card, Type disc, Context context)
{
  TypeKind dpt = type_ur_kind(disc);
  Type ud = ur_type(disc);
  switch (dpt) {
  case boolean_Type:
  case shortcharacter_Type:
    fprintf(context->file, "    %s := ORD(%s);\n", card, mem);
    break;
  case enumeration_Type:
    fprintf(context->file, "    %s := %s%s__EnumValues[ORD(%s)];\n",
	    card, SupportPrefix(ud, context), M3_SHORT_TYPE_NAME(ud),
	    mem);
    break;
  case shortinteger_Type:
    fprintf(context->file, "    %s := %s MOD 65536;\n",
	    card, mem);
    break;
  case integer_Type:
  case cardinal_Type:
    fprintf(context->file, "    CASE %s OF\n", mem);
    if (dpt == integer_Type)
      fprintf(context->file, "    -32768..32767 => %s := %s MOD 65536\n",
	      card, mem);
    else
      fprintf(context->file, "    0..65535 => %s := %s\n",
	      card, mem);
    fprintf(context->file, "    ELSE RAISE IluBasics.Failed(\n");
    fprintf(context->file, "      NEW(IluBasics.Failure, info := ");
    fprintf(context->file, "\"Can't output tag value of \" &\n");
    fprintf(context->file, "          Fmt.%s(%s, 10) ))\n",
	    (dpt == integer_Type) ? "Int" : "Unsigned", mem);
    fprintf(context->file, "    END (*case*);\n");
    break;
  case shortcardinal_Type:
  case character_Type:
  case byte_Type:
    fprintf(context->file, "    %s := %s;\n", card, mem);
    break;
  default:
    fprintf(stderr,
	    "Don't know how to convert union tag of type %s to Ilu.ShortCard!\n",
	    trname(disc));
    fprintf(context->file, "    %s := %s (* error here *);\n", card, mem);
  }
}

static void CardToMem(string subj, Type disc, Context context)
{
  TypeKind dpt = type_ur_kind(disc);
  Type ud = ur_type(disc);
  switch (dpt) {
  case boolean_Type:
    fprintf(context->file, "%s # 0", subj);
    break;
  case shortcharacter_Type:
    fprintf(context->file, "VAL(%s, %s)", subj, trname(disc));
    break;
  case enumeration_Type:
    fprintf(context->file, "%s%s__VAL(%s)", SupportPrefix(ud, context),
	    M3_SHORT_TYPE_NAME(ud), subj);
    break;
  case shortinteger_Type:
  case integer_Type:
    fprintf(context->file,
	    "((%s + 32768) MOD 65536) - 32768",
	    subj);
    break;
  case shortcardinal_Type:
  case cardinal_Type:
  case character_Type:
  case byte_Type:
    fprintf(context->file, "%s", subj);
    break;
  default:
    fprintf(stderr,
	    "Don't know how to convert Ilu.ShortCard to union tag of type %s!\n",
	    trname(disc));
    fprintf(context->file, "%s (* error here *)", subj);
  }
}

static void generate_output_code (Type type, TypeKind t, Context context)
{
  TypeDescription d = type_description(type);
  boolean zeroarray = FALSE;
  int dims = 0;

  if (t == array_Type)
    {
      dims = list_size(d->structuredDes.array.dimensions);
      zeroarray = FALSE;
      list_enumerate(d->structuredDes.array.dimensions, (EnumProc) FindZero, &zeroarray);
    }

  fprintf (context->file, "PROCEDURE Marshall_%s (call_ : IluRuntime.Call; %sREADONLY val : %s)\n",
	   M3_SHORT_TYPE_NAME(type), zeroarray ? "<*UNUSED*> " : "", M3_TYPE_NAME(type));
  fprintf (context->file, "  RAISES {IluBasics.Failed, Thread.Alerted} =\n");

  if (t == array_Type)
    {
      unsigned long size = 1;
      char buf[1000];

      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) ComputeTotalNumberOfElements, &size);
      fprintf (context->file, "  BEGIN\n    IluRuntime.OutputArray(call_, %lu);\n", size);
      if (NOT zeroarray)
	{
	  TypeKind t2 = type_ur_kind(d->structuredDes.array.type);
	  if (t2 == byte_Type OR t2 == shortcharacter_Type)
	    {
	      DimStop = dims - 1;
	      DimCount = 0;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimHeader, context);
	      sprintf (buf, "val");
	      DimCount = 0;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimRef, buf);
	      fprintf (context->file, "%*.*sIluRuntime.Output%s (call_, %s);\n", dims * 2 + 2, dims*2+2, "",
		       (t2 == byte_Type) ? "Opaque" : "StringVec", buf);
	      DimCount = dims;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimFooter, context);
	      DimStop = 0;
	    }
	  else
	    {
	      DimCount = 0;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimHeader, context);
	      sprintf (buf, "val");
	      DimCount = 0;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimRef, buf);
	      fprintf (context->file, "%*.*s", dims * 2 + 2, dims*2+2, "");
	      MarshallValue (d->structuredDes.array.type, buf, context, dims*2+2);
	      DimCount = dims;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimFooter, context);
	    }
	}
      fprintf (context->file, "    IluRuntime.EndArray(call_);\n");
    }
  else if (t == sequence_Type)
    {
      fprintf (context->file, "  VAR len := NUMBER(val^);\n  BEGIN\n");
      fprintf (context->file, "    IluRuntime.OutputSequence(call_, len, %u);\n",
	       d->structuredDes.sequence.limit);
      fprintf (context->file, "    FOR i := 0 TO len - 1 DO\n");
      MarshallValue (d->structuredDes.sequence.type, "val[i]", context, 6);
      fprintf (context->file, "    END (*for*);\n    IluRuntime.EndSequence(call_);\n");
    }
  else if (t == union_Type)
    {
      UnionTemp       s;
      TypeDescription td = type_description(type);
      Type            disc = td->structuredDes.uniond.discriminator_type;

      s.c = context;
      s.t = type;
      s.skipDefault = 1;
      fprintf(context->file, "  VAR discCode: Ilu.ShortCard;\n");
      fprintf(context->file, "  BEGIN\n");
      MemToCard("val.d", "discCode", disc, context);
      fprintf(context->file, "    IluRuntime.OutputUnion(call_, discCode);\n");
      fprintf(context->file, "    CASE val.d OF\n");
      list_enumerate(d->structuredDes.uniond.types,
		     (EnumProc) OutputUnionArm, &s);
      s.skipDefault = 0;
      if (td->structuredDes.uniond.default_arm != NULL)
	OutputUnionArm(td->structuredDes.uniond.default_arm, &s);
      else if (td->structuredDes.uniond.others_allowed) {
	fprintf(context->file, "    ELSE\n");
	fprintf(context->file, "      TYPECASE val OF\n");
	fprintf(context->file, "      | %s_OTHERS =>\n",
		M3_SHORT_TYPE_NAME(type));
	fprintf(context->file, "        EVAL 0\n");
	fprintf(context->file, "      ELSE\n");
	fprintf(context->file, "        RAISE\n");
	fprintf(context->file, "          IluBasics.Failed(\n");
	fprintf(context->file, "            NEW(IluBasics.Failure,\n");
	fprintf(context->file, "                info := \"type of a %s doesn't match its tag\"))\n", M3_TYPE_NAME(type));
	fprintf(context->file, "      END (* narrow for this arm *)\n");
      } else {
	fprintf(context->file, "    ELSE\n");
	fprintf(context->file, "      RAISE\n");
	fprintf(context->file, "        IluBasics.Failed(NEW(IluBasics.Failure, ");
	fprintf(context->file, "info := \"marshalling union with unexpected tag (code = \" & Fmt.Unsigned(discCode, 10) & \")\"))\n");
      }
      fprintf(context->file, "    END (*CASE val.d*);\n");
      fprintf(context->file, "    IluRuntime.EndUnion(call_);\n");
    }
  else if (t == record_Type)
    {
      fprintf (context->file, "  BEGIN\n    IluRuntime.OutputRecord(call_);\n");
      list_enumerate (d->structuredDes.record, (EnumProc) OutputRecordArg, context);
      fprintf (context->file, "    IluRuntime.EndRecord(call_);\n");
    }
  else if (t == object_Type)
    {}	/* done elsewhere */
  else
    {}

  fprintf (context->file, "  END Marshall_%s;\n\n", M3_SHORT_TYPE_NAME(type));
}

/**********************************************************************
***********************************************************************
*
*     Input code
*
***********************************************************************
***********************************************************************/


static void InputRecordField (Argument arg, Context context)
{
  char buf[1000];

  sprintf (buf, "val.%s", m3_argument_name(arg));
  UnmarshallValue (context, arg->type, 0, buf, 6, TRUE);
}

static void InputUnionArm (Argument a, UnionTemp *s)
{
  Type            union_type = s->t;
  TypeDescription td = type_description(union_type);
  Type            disc = td->structuredDes.uniond.discriminator_type;

  if (GuardUnionArm(a, s))
    return;

  if (IsBulkyData(a->type)) {
    fprintf(s->c->file, "        VAR v_ : %s;\n",
	    M3_TYPE_NAME(a->type));
    fprintf(s->c->file, "        BEGIN\n");
    UnmarshallValue(s->c, a->type, 0, "v_", 10, TRUE);
    fprintf(s->c->file, "          val := NEW(%s, d:= disc, v := v_);\n",
	    UnionCaseName(s->t, a));
    fprintf(s->c->file, "        END\n");
  } else {
    fprintf(s->c->file, "        val := NEW(%s, d:= disc, ",
	    UnionCaseName(s->t, a), UnionCaseName(s->t, a));
    UnmarshallValue(s->c, a->type, 0, "v", 0, FALSE);
    fprintf(s->c->file, ")\n");
  }
  return;
}

static void generate_input_code (Type type, TypeKind t, Context context)
{
  char *name = M3_SHORT_TYPE_NAME(type);
  char *tname = M3_TYPE_NAME(type);
  TypeDescription d = type_description(type);
  boolean zeroarray = FALSE;
  int dims = 0;

  if (t == array_Type)
    {
      dims = list_size(d->structuredDes.array.dimensions);
      zeroarray = FALSE;
      list_enumerate(d->structuredDes.array.dimensions, (EnumProc) FindZero, &zeroarray);
    }

  context->class = type;

  if (IsBulkyData(type))
    fprintf (context->file, "PROCEDURE Unmarshall_%s (call_ : IluRuntime.Call; %sVAR val : %s)\n",
	     name, zeroarray ? "<*UNUSED*> " : "", tname);
  else
    fprintf (context->file, "PROCEDURE Unmarshall_%s (call_ : IluRuntime.Call) : %s\n",
	     name, tname);
  fprintf (context->file, "  RAISES {IluBasics.Failed, Thread.Alerted} =\n");
  if (!IsBulkyData(type))
    fprintf (context->file, "  VAR val: %s;\n", tname);

  if (t == array_Type)
    {
      char buf[1000];

      fprintf (context->file, "  BEGIN\n");
      fprintf (context->file, "    IluRuntime.InputArray(call_);\n");
      if (NOT zeroarray)
	{
	  TypeKind t2 = type_ur_kind(d->structuredDes.array.type);
	  if (t2 == byte_Type OR t2 == shortcharacter_Type)
	    {
	      DimStop = dims - 1;
	      DimCount = 0;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimHeader, context);
	      sprintf (buf, "val");
	      DimCount = 0;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimRef, buf);
	      fprintf (context->file, "%*.*sIluRuntime.Input%s (call_, %s);\n", dims * 2 + 2, dims*2+2, "",
		       (t2 == byte_Type) ? "Opaque" : "StringVec", buf);
	      DimCount = dims;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimFooter, context);
	      DimStop = 0;
	    }
	  else
	    {
	      DimCount = 0;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimHeader, context);
	      sprintf (buf, "val");
	      DimCount = 0;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimRef, buf);
	      UnmarshallValue (context, d->structuredDes.array.type, 0, buf, dims*2+2, TRUE);
	      DimCount = dims;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimFooter, context);
	    }
	}
      fprintf (context->file, "    IluRuntime.EndArray(call_);\n");
    }
  else if (t == sequence_Type)
    {
      Type et = d->structuredDes.sequence.type;

      fprintf (context->file, "  VAR len : IluRuntime.Cardinal;\n  BEGIN\n");
      fprintf (context->file, "    len := IluRuntime.InputSequence (call_, %u);\n",
	       d->structuredDes.sequence.limit);
      fprintf (context->file, "    val := NEW(%s, len);\n", tname);
      fprintf (context->file, "    FOR i := 0 TO len - 1 DO\n");
      UnmarshallValue (context, et, type->def, "val[i]", 6, TRUE);
      fprintf (context->file, "    END (* for *);\n");
      fprintf (context->file, "    IluRuntime.EndSequence(call_);\n");
    }
  else if (t == union_Type)
    {
      UnionTemp       s;
      Type            disc = d->structuredDes.uniond.discriminator_type;

      s.c = context;
      s.t = type;
      s.skipDefault = 1;
      fprintf(context->file, "  VAR discCode := IluRuntime.InputUnion(call_);\n");
      fprintf(context->file, "  VAR disc: %s;\n", trname(disc));
      fprintf(context->file, "  BEGIN\n");
      fprintf(context->file, "    disc := ");
      CardToMem("discCode", disc, context);
      fprintf(context->file, ";\n");
      fprintf(context->file, "    CASE disc OF\n");
      list_enumerate(d->structuredDes.uniond.types,
		     (EnumProc) InputUnionArm, &s);
      s.skipDefault = 0;
      if (d->structuredDes.uniond.default_arm != NULL)
	InputUnionArm(d->structuredDes.uniond.default_arm, &s);
      else if (d->structuredDes.uniond.others_allowed) {
	fprintf(context->file, "    ELSE\n");
	fprintf(context->file, "      val := NEW(%s_OTHERS, d:= disc)\n",
		name);
      } else {
	fprintf(context->file, "    ELSE\n");
	fprintf(context->file, "      RAISE\n");
	fprintf(context->file, "        IluBasics.Failed(NEW(IluBasics.Failure, ");
	fprintf(context->file,
		"info := Fmt.F(\"got %%s as tag code for a %s\", Fmt.Int(discCode))))\n",
		tname);
      }
      fprintf(context->file, "    END (*CASE disc*);\n");
      fprintf(context->file, "    IluRuntime.EndUnion(call_);\n");
    }
  else if (t == record_Type)
    {
      fprintf (context->file, "  BEGIN\n");
      fprintf (context->file, "    IluRuntime.InputRecord (call_);\n");
      list_enumerate (d->structuredDes.record, (EnumProc) InputRecordField, context);
      fprintf (context->file, "    IluRuntime.EndRecord(call_);\n");
    }
  else if (t == optional_Type)
    {
      TypeKind t2 = type_ur_kind(d->structuredDes.optional);
      boolean pointer = (t2 == object_Type OR t2 == sequence_Type OR t2 == union_Type);

      fprintf (context->file, "  VAR test : BOOLEAN;\n");
      fprintf (context->file, "  BEGIN\n");
      fprintf (context->file, "    test := IluRuntime.InputOptional(call_);\n");
      fprintf (context->file, "    IF test THEN\n");
      if (!pointer)
	fprintf (context->file, "      val := NEW(%s);\n", tname);
      UnmarshallValue (context, d->structuredDes.optional, d->structuredDes.optional->def,
		       pointer ? "val" : "val^", 6, TRUE);
      fprintf (context->file, "    ELSE\n");
      fprintf (context->file, "      val := NIL;\n");
      fprintf (context->file, "    END (*if*);\n");
    }
  else if (t == enumeration_Type)
    {
      fprintf (context->file, "  VAR code : IluRuntime.EnumOrd;\n");
      fprintf (context->file, "  BEGIN\n");
      fprintf (context->file, "    code := IluRuntime.InputEnum(call_);\n");
      fprintf (context->file, "    val := %s__VAL(code);\n", name);
    }
  else if (t == object_Type)
    {}	/* done elsewhere */
  else
    {}

  if (!IsBulkyData(type))
    fprintf (context->file, "    RETURN val;\n");
  fprintf (context->file, "  END Unmarshall_%s;\n\n", name);
  
  if (t == enumeration_Type) {
    fprintf(context->file, "\n");
    fprintf(context->file, "PROCEDURE %s__VAL(code: IluRuntime.EnumOrd)",
	    name);
    fprintf(context->file, ": %s RAISES {IluBasics.Failed} =\n", name);
    fprintf(context->file, "  BEGIN\n");
    fprintf(context->file, "    FOR i := 0 TO %u DO\n",
	    list_size(d->structuredDes.enumeration) - 1);
    fprintf(context->file, "      IF %s__EnumValues[i] = code THEN\n",
	    name);
    fprintf(context->file, "        RETURN VAL(i, %s);\n", tname);
    fprintf(context->file, "      END (* if *);\n");
    fprintf(context->file, "    END (* for *);\n");
    fprintf(context->file, "    RAISE\n");
    fprintf(context->file, "      IluBasics.Failed(NEW(IluBasics.Failure, info := Fmt.F(\"got %%s as code for an enumeration type %s\", Fmt.Int(code))))\n", tname);
    fprintf(context->file, "    END %s__VAL;\n\n", name);
  }
}

/**********************************************************************
***********************************************************************
*
*     Sizeof code
*
***********************************************************************
***********************************************************************/

static void SizeOfUnion (Argument a, UnionTemp *s)
{
  if (GuardUnionArm(a, s))
    return;
  fprintf(s->c->file, "      TYPECASE val OF\n");
  fprintf(s->c->file, "      | %s (x) =>\n", UnionCaseName(s->t, a));
  fprintf(s->c->file, "        size := size + ");
  SizeValue (a->type, "x.v", s->c);
  fprintf (s->c->file, ";\n");
  fprintf(s->c->file, "      ELSE\n");
  fprintf(s->c->file, "        RAISE\n");
  fprintf(s->c->file, "          IluBasics.Failed(\n");
  fprintf(s->c->file, "            NEW(IluBasics.Failure,\n");
  fprintf(s->c->file, "                info := \"type of a %s doesn't match its tag\"))\n", M3_TYPE_NAME(s->t));
  fprintf(s->c->file, "      END (* narrow for this arm *)\n");
}

static void SizeRecordArg (Argument arg, Context context)
{
  char buf[1000];
  fprintf (context->file, "    INC(size, ");
  sprintf (buf, "val.%s", m3_argument_name(arg));
  SizeValue (arg->type, buf, context);
  fprintf (context->file, ");\n");
}

static void generate_sizeof_code (Type type, TypeKind t, Context context)
{
  TypeDescription d = type_description(type);
  char *name = M3_SHORT_TYPE_NAME(type);
  char *tname = M3_TYPE_NAME(type);
  boolean zeroarray = FALSE;
  int dims = 0;

  if (t == array_Type)
    {
      dims = list_size(d->structuredDes.array.dimensions);
      zeroarray = FALSE;
      list_enumerate(d->structuredDes.array.dimensions, (EnumProc) FindZero, &zeroarray);
    }

  context->class = type;

  fprintf (context->file, "PROCEDURE ComputeMarshalledSize_%s (call_: IluRuntime.Call; %sREADONLY val: %s): IluRuntime.Cardinal\n",
	   name, zeroarray ? "<*UNUSED*> " : "", tname);
  fprintf (context->file, "  RAISES {IluBasics.Failed} =\n  VAR size: IluRuntime.Cardinal;\n");
  if (t == array_Type)
    {
      unsigned long size = 1;
      char buf[1000];

      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) ComputeTotalNumberOfElements, &size);
      fprintf (context->file, "  BEGIN\n");
      fprintf (context->file, "    size := IluRuntime.SizeArray(call_, %lu);\n", size);
      if (NOT zeroarray)
	{
	  TypeKind t2 = type_ur_kind(d->structuredDes.array.type);
	  if (t2 == byte_Type OR t2 == shortcharacter_Type)
	    {
	      DimStop = dims - 1;
	      DimCount = 0;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimHeader, context);
	      sprintf (buf, "val");
	      DimCount = 0;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimRef, buf);
	      fprintf (context->file, "%*.*sINC(size, IluRuntime.Size%s (call_, %s));\n", dims * 2 + 2, dims*2+2, "",
		       (t2 == byte_Type) ? "Opaque" : "StringVec", buf);
	      DimCount = dims;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimFooter, context);
	      DimStop = 0;
	    }
	  else
	    {
	      DimCount = 0;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimHeader, context);
	      sprintf (buf, "val");
	      DimCount = 0;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimRef, buf);
	      fprintf (context->file, "%*.*sINC(size, ", dims * 2 + 4, dims*2+4, "");
	      SizeValue (d->structuredDes.array.type, buf, context);
	      fprintf (context->file, ");\n");
	      DimCount = dims;
	      list_enumerate (d->structuredDes.array.dimensions, (EnumProc) DimFooter, context);
	    }
	}
      fprintf (context->file, "    IluRuntime.EndArray(call_);\n");
    }
  else if (t == sequence_Type)
    {
      fprintf (context->file, "  VAR len := NUMBER(val^);\n");
      fprintf (context->file, "  BEGIN\n");
      fprintf (context->file, "    size := IluRuntime.SizeSequence(call_, len, %u);\n",
	       d->structuredDes.sequence.limit);
      fprintf (context->file, "    FOR i := 0 TO len - 1 DO\n");
      fprintf (context->file, "      INC(size, ");
      SizeValue (d->structuredDes.sequence.type, "val[i]", context);
      fprintf (context->file, ");\n");
      fprintf (context->file, "    END (*for*);\n");
      fprintf (context->file, "    IluRuntime.EndSequence(call_);\n");
    }
  else if (t == union_Type)
    {
      UnionTemp       s;
      Type            disc = d->structuredDes.uniond.discriminator_type;

      s.c = context;
      s.t = type;
      s.skipDefault = 1;
      fprintf(context->file, "  VAR discCode: Ilu.ShortCard;\n");
      fprintf(context->file, "  BEGIN\n");
      MemToCard("val.d", "discCode", disc, context);
      fprintf(context->file, "    size := IluRuntime.SizeUnion(call_, discCode);\n");
      fprintf(context->file, "    CASE val.d OF\n");
      list_enumerate(d->structuredDes.uniond.types,
		     (EnumProc) SizeOfUnion, &s);
      s.skipDefault = 0;
      if (d->structuredDes.uniond.default_arm != NULL)
	SizeOfUnion(d->structuredDes.uniond.default_arm, &s);
      else {
	fprintf(context->file, "    ELSE EVAL 0\n");
	if (!d->structuredDes.uniond.others_allowed)
	  fprintf(context->file, "      (*output routine will raise exn*)\n");
      }
      fprintf(context->file, "    END (*CASE val.d*);\n");
      fprintf(context->file, "    IluRuntime.EndUnion(call_);\n");
    }
  else if (t == record_Type)
    {
      fprintf (context->file, "  BEGIN\n");
      fprintf (context->file, "    size := IluRuntime.SizeRecord (call_);\n");
      list_enumerate (d->structuredDes.record, (EnumProc) SizeRecordArg, context);
      fprintf (context->file, "    IluRuntime.EndRecord(call_);\n");
    }
  else if (t == optional_Type)
    {
      TypeKind t2 = type_ur_kind(d->structuredDes.optional);
      char *s = (t2 == object_Type OR t2 == sequence_Type OR t2 == union_Type) ? "val" : "val^";

      fprintf (context->file, "  BEGIN\n");
      fprintf (context->file, "    size := IluRuntime.SizeOptional (call_, (val # NIL));\n");
      fprintf (context->file, "    IF val # NIL THEN\n      size := size + ");
      SizeValue (d->structuredDes.optional, s, context);
      fprintf (context->file, ";\n    END;\n");
    }
  else if (t == object_Type)
    {}	/* done elsewhere */
  else
    {}

  fprintf (context->file, "    RETURN size;\n  END ComputeMarshalledSize_%s;\n\n", name);
}

void generate_type_io_code (Type type, Context context)
{
  TypeKind t = type_ur_kind(type);
  TypeDescription td = type_description(type);

  if ((TypeIsNonObjectStruct(type)
       || (TypeIsArray(type)
	   && ((list_size(td->structuredDes.array.dimensions) > 1)
	       || ((type_ur_kind(td->structuredDes.array.type)
		    != byte_Type)
		 && (type_ur_kind(td->structuredDes.array.type)
		     != shortcharacter_Type))))
       || (t == sequence_Type
	   && (type_ur_kind(td->structuredDes.sequence.type)
	       != byte_Type)
	   && (type_ur_kind(td->structuredDes.sequence.type)
	       != shortcharacter_Type)))
      && (t != object_Type)
      && (type->importInterfaceName == NULL)
      && (type->supertype == NULL)) {
    generate_output_code(type, t, context);
    generate_input_code(type, t, context);
    generate_sizeof_code(type, t, context);
  }
  if (t == optional_Type
      && (type->importInterfaceName == NULL)
      && (type->supertype == NULL)) {
    generate_input_code(type, t, context);
    generate_sizeof_code(type, t, context);
  }
  if (t == enumeration_Type
      && (type->importInterfaceName == NULL)
      && (type->supertype == NULL)) {
    generate_input_code(type, t, context);
  }
}
