/*
Copyright (c) 1991, 1992, 1993 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

$Id: m3.h,v 1.12 1994/11/11 01:13:42 spreitze Exp $
*/
/* Last edited by Mike Spreitzer November 10, 1994 4:45 pm PST */

#include <stdio.h>
#include <string.h>

#define TRUE  1
#define FALSE 0
typedef int Boolean;

#define AND  &&
#define OR   ||
#define NOT  !

#include <iluptype.h>

#define SEGFAULT	*((char *)1)='a'

typedef void (*EnumProc)(void *, void *);
typedef boolean (*FindProc)(void *, void *);

enum Module { GeneralInterface, InternalInterface, InternalModule, ClientModule, ServerModule };

typedef struct context_s {
  Interface interface;
  FILE *file;
  Type class;
  enum Module module;
} * Context;

extern char *ProgramName;

extern boolean ReadSynonyms (string filename);
extern Class class_object();
extern TypeDescription type_description();
extern void ImportInterface(Imported i, Context context);
extern Type exception_type();
extern string type_name(Type);
extern string m3_return_type();
extern string m3_type_name(Type t, boolean full);
extern string m3_simple_name();
extern string m3_string(string s);
extern string m3_interface_name(Interface i);
extern string interface_name(Interface);
extern list new_list();
extern boolean IsSequenceObject(Type type, Interface i);
extern boolean IsPipeObject(Type type, Interface i);
extern boolean IsCacheableMethod (Procedure m);
extern boolean TypeIsStruct();
extern boolean TypeIsArray(Type type);
extern boolean TypeIsString(Type type);
extern boolean TypeIsNonObjectStruct(Type type);
extern void SizeValue (Type type, string name, Context context);
extern void MarshallValue (Type type, string name, Context context,
				int indent);
extern void UnmarshallValue (Context context, Type type, LineNumber line,
			string name, int indent, boolean terminate);
extern boolean BlockingCall (Procedure proc);

extern void generate_headers (Interface interface, FILE *file);
extern void declare_object_type(Type type, Context c);
extern void generate_support_headers (Interface i, FILE *file);
extern void generate_support_code (Interface i, FILE *file);
extern void generate_client_code (Interface i, FILE *file);
extern void generate_server_code (Interface parse, FILE *file);
extern void generate_type_io_code (Type type, Context context);
extern string iname(string name, Interface interface);
extern string trname(Type t);
extern string m3_constant_name (Constant c, boolean full);
extern string m3_argument_name (Argument a);
extern char *SupportPrefix(Type type, Context context);
extern char * UnionCaseName (Type union_type, Argument arm);
extern char * TwoTypeName (Type t1, Type t2);
extern string m3_exception_name (Exception e, boolean full);
extern void generate_boilerplate (FILE *file, Interface parse);
extern string m3_procedure_name (Procedure p, boolean full);
extern boolean IsBulkyData (Type type);
extern void UnmarkSupertypes(Type);
extern void ListArgument (Argument arg, Context context);
extern void ListExceptionName (Exception e, Context context);

extern boolean matchString (refany s1, refany s2);

#define M3_TYPE_NAME(type)		m3_type_name(type, TRUE)
#define M3_SHORT_TYPE_NAME(type)	m3_type_name(type, FALSE)
#define M3_EXCEPTION_NAME(exception)		m3_exception_name(exception, TRUE)
#define M3_SHORT_EXCEPTION_NAME(exception)	m3_exception_name(exception, FALSE)
#define M3_CONSTANT_NAME(constant)		m3_constant_name(constant, TRUE)
#define M3_SHORT_CONSTANT_NAME(constant)	m3_constant_name(constant, FALSE)
#define M3_METHOD_NAME(method)		m3_procedure_name(method, TRUE)
#define M3_SHORT_METHOD_NAME(method)	m3_procedure_name(method, FALSE)

