/*
Copyright (c) 1991, 1992, 1993 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

$Id: object.c,v 1.14 1994/11/11 01:13:42 spreitze Exp $
*/
/* Last edited by Mike Spreitzer November 10, 1994 3:35 pm PST */

#include "m3.h"

static boolean FirstItem;

static void DeclareMethodArgument (Argument a, Context context)
{
  char           *mode;
  switch (a->direction) {
  case In:
    mode = "";
    break;
  case Out:
  case InOut:
    mode = "VAR ";
    break;
  }
  if (FirstItem)
    FirstItem = FALSE;
  else
    fprintf(context->file, "; ");
  fprintf(context->file, "%s%s: %s", mode, m3_argument_name(a),
	  iname(M3_TYPE_NAME(a->type), a->type->interface));
}

static void ListExceptionIName (Exception e, Context context)
{
  fprintf (context->file, ", %s", iname(M3_EXCEPTION_NAME(e), e->interface));
}

static void DeclareMethod (Procedure m, Context context)
{
  boolean oldFirst;

  fprintf (context->file, "    %s (", M3_SHORT_METHOD_NAME(m));
  oldFirst = FirstItem;  FirstItem = TRUE;
  list_enumerate (m->arguments, (EnumProc) DeclareMethodArgument, context);
  FirstItem = oldFirst;
  fprintf (context->file, ")");
  if (m->returnType != NULL AND type_ur_kind(m->returnType) != void_Type)
    fprintf (context->file, ": %s", iname(M3_TYPE_NAME(m->returnType), m->returnType->interface));
  fprintf (context->file, " RAISES {IluBasics.Failed, Thread.Alerted");
  list_enumerate (m->exceptions, (EnumProc) ListExceptionIName, context);
  fprintf (context->file, "};\n");
}

boolean IsCacheableMethod (Procedure m)
{
  TypeKind t = type_ur_kind(m->returnType);

  return (m->functional && list_size(m->arguments) == 0
	  && (t == object_Type || t == enumeration_Type || t == byte_Type
	      || t == integer_Type || t == shortinteger_Type
	      || t == cardinal_Type || t == shortcardinal_Type
	      || t == real_Type || t == shortreal_Type
	      || t == character_Type || t == shortcharacter_Type));
}

void declare_object_type (Type type, Context c)
{
  Class           od;
  string          name;
  int             isalias = type->supertype != NULL;

  od = class_object(type);
  c->class = type;
  name = M3_SHORT_TYPE_NAME(type);

  if (isalias) {
  fprintf(c->file,
	  "(* The next 3 procs are equivalent to their %s counterparts. *)\n",
	  M3_TYPE_NAME(type->supertype));
  } else {
    fprintf(c->file, "\n\n(* declaration of M3 type \"%s\"",
	    M3_TYPE_NAME(type));
    fprintf(c->file, "from ILU class \"%s:%s\"  *)\n\n",
	    interface_name(type->interface), type_name(type));

    fprintf(c->file, "TYPE %s = %s OBJECT\n  METHODS\n",
	    M3_SHORT_TYPE_NAME(type),
	    (list_size(od->superclasses) > 0)
	    ? iname(M3_TYPE_NAME(list_car(od->superclasses)),
		    ((Type) list_car(od->superclasses))->interface)
	    : "Ilu.Object");

    list_enumerate(od->methods, (EnumProc) DeclareMethod, c);

    fprintf(c->file, "  OVERRIDES\n");
    fprintf(c->file, "    ILU_Get_Type := ILU_Get_Type_%s\n", name);
    fprintf(c->file, "  END;\n\n");
  }

  fprintf(c->file, "PROCEDURE ILU_SBH_To_%s", name);
  fprintf(c->file, "(sbh: TEXT; mostSpecificTypeID: TEXT := NIL): %s\n",
	  name);
  fprintf(c->file, "  RAISES {IluBasics.Failed, Thread.Alerted};\n\n");

  fprintf(c->file, "PROCEDURE ILU_Get_Type_%s", name);
  fprintf(c->file, "(self : Ilu.Object): Ilu.ObjectType;\n\n");

  fprintf(c->file, "PROCEDURE ILU_Qua_%s (x: Ilu.Object): %s;\n\n",
	  name, name);
}
